/* This example mimic PnetCDF APIs to write two 1D array of float and int types.
 * The PnetCDF objects created in this example are:
 * 1. One file (corresponding to Damsel model)
 * 2. One dimension (corresponding to a Damsel container)
 * 3. Two 1D array variables using the dimension from 2
 *    3.1 flt_var1 (corresponding to a Damsel tag of type float)
 *    3.2 int_var1 (corresponding to a Damsel tag of type int)
 * 4. Write data to the file
 * 5. close the file (corresponding to close containers and model)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include <mpi.h>
#include <damsel.h>

/* define macros used in this program */
#define DIM_LEN 15

#define WRITE_MODE    0
#define READ_MODE     1
#define NUM_PER_CHUNK 32

/* C structures used in this program */
typedef struct {
    char              name[128];
    damsel_container  container;
    int               num_tags;
    damsel_handle    *tags;
} container_list;

typedef struct {
    MPI_Comm        comm;
    damsel_model    model;
    int             num_containers;
    container_list *containers;
} file_descriptor;

typedef file_descriptor * file_dscpt;

/* function declaration */
static damsel_err_t mimic_ncmpi_file_create(MPI_Comm comm, char *filename, int cmode, file_dscpt *fd);
static damsel_err_t mimic_ncmpi_def_dim(file_dscpt fd, char *name, int dim_len, damsel_container *dim_id);
static damsel_err_t mimic_ncmpi_def_var(file_dscpt fd, char *name, damsel_data_type type, damsel_container dim_id, void *buf, damsel_handle *var_id);
static damsel_err_t mimic_ncmpi_enddef(file_dscpt fd);
static damsel_err_t mimic_ncmpi_iput(file_dscpt fd, damsel_request_t *req);
static damsel_err_t mimic_ncmpi_wait(file_dscpt fd, damsel_request_t req, damsel_status_t *status);
static damsel_err_t mimic_ncmpi_close(file_dscpt *fd);

#define CHECK_ERR(e) {if ((e.id!=DMSL_OK.id)) { printf("Error at line=%d msg=%s\n", __LINE__,DMSLerror_str(e)); goto fn_exit;}}

/*----< main() >--------------------------------------------------------------*/
int main(int argc, char **argv) {
    int i, rank, nprocs, *int_buf;
    float *flt_buf;
    damsel_err_t err;

#ifndef DAMSEL_SINGLETHREAD
    int res;
    MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
    assert(res == MPI_THREAD_MULTIPLE);
#else
    MPI_Init(&argc, &argv);
#endif

    /* initialize Damsel environment */
    damsel_library lib;
    err = DMSLlib_init(&lib);
    CHECK_ERR(err)

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
    if (argc != 2) {
        if (!rank) printf("Usage: %s filename\n",argv[0]);
        goto fn_exit;
    }

    /* create a new file */
    file_dscpt fd;
    err = mimic_ncmpi_file_create(MPI_COMM_WORLD, argv[1], WRITE_MODE, &fd);
    CHECK_ERR(err)

    /* define a dimension of size DIM_LEN (global size) */
    damsel_container dim_id;
    err = mimic_ncmpi_def_dim(fd, "dim_name", DIM_LEN, &dim_id);
    CHECK_ERR(err)

    /* initialize local array of size local_len */
    int local_len = DIM_LEN / nprocs;
    if (rank < DIM_LEN % nprocs) local_len++;
    flt_buf = (float*) malloc(local_len * sizeof(float));
    int_buf = (int*)   malloc(local_len * sizeof(int));
    for (i=0; i<local_len; i++) {
        flt_buf[i] = 0.01 * rank + rank;
        int_buf[i] = rank;
    }

    /* define a global variable, a 1D array of float type */
    damsel_handle var_id[2] = {123, 456}; /* what value should this be? */
    err = mimic_ncmpi_def_var(fd, "flt_var1", DAMSEL_DATA_TYPE_FLOAT, dim_id,
                              flt_buf, &var_id[0]);
    CHECK_ERR(err)

    /* define another global variable, a 1D array of int type */
    err = mimic_ncmpi_def_var(fd, "int_var1", DAMSEL_DATA_TYPE_INTEGER, dim_id,
                              int_buf, &var_id[1]);
    CHECK_ERR(err)

    /* end of define */
    err = mimic_ncmpi_enddef(fd);
    CHECK_ERR(err)

    /* Now we can tell Damsel to start writing the data */
    damsel_request_t req;
    err = mimic_ncmpi_iput(fd, &req);
    CHECK_ERR(err)

    /* Other computations can be performed here, i.e. between the post of
     * asynchronous transfer and wait, as long as the registered data buffer,
     * buf, is not touched */

    damsel_status_t status;
    err = mimic_ncmpi_wait(fd, req, &status);
    CHECK_ERR(err)

    err = mimic_ncmpi_close(&fd);

    free(flt_buf);
    free(int_buf);
fn_exit:
    /* exit Damsel environment */
    if (lib != DAMSEL_LIBRARY_INVALID) {
        err = DMSLlib_finalize(lib);
        CHECK_ERR(err)
    }

    MPI_Finalize();
    return 0;
}

/*----< mimic_ncmpi_file_create() >-------------------------------------------*/
/* 1. create a file descriptor, to store the following Damsel objects.
 * 2. make a copy of MPI communicator
 * 3. create a Damsel model handle
 * 4. set the access mode for the model
 */
static
damsel_err_t mimic_ncmpi_file_create(MPI_Comm    comm,
                                     char       *filename,
                                     int         cmode, /* create mode */
                                     file_dscpt *fd)    /* OUT */
{
    damsel_err_t err;

    *fd = (file_descriptor*) malloc(sizeof(file_descriptor));
    (*fd)->num_containers = 0;
    (*fd)->containers     = NULL;

    /* make a copy of comm in fd */
    MPI_Comm_dup(comm, &(*fd)->comm);

    /* create a model, a in-memory store */
    err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &(*fd)->model);
    if (err.id != DMSL_OK.id) return err;

    /* make sure the file does not exist, ignore error if any */
    unlink(filename);

    /* set the file access mode */
    damsel_trait_bundle tb;
    if (err.id != DMSL_OK.id) return err;
    err = DMSLtrait_create(&tb);
    if (err.id != DMSL_OK.id) return err;

    if (cmode != WRITE_MODE) return DMSL_ECMODE;

    err = DMSLtrait_put(tb, "mode", "WRITE");
    if (err.id != DMSL_OK.id) return err;

    /* register the filename and access mode to model (in-memory store) */
    err = DMSLmodel_attach((*fd)->model, filename, comm, tb);
    if (err.id != DMSL_OK.id) return err;

    err = DMSLtrait_release(tb);
    if (err.id != DMSL_OK.id) return err;

    return DMSL_OK;
}

/*----< mimic_ncmpi_def_dim() >-----------------------------------------------*/
/* Since a Damsel container can only contain one set of handles, this program
 * uses a Damsel container to represent a dimension.
 * This API does the followings:
 * 1. figure out how many local handles each process rank will create
 * 2. calculate the starting number for each process and use it to set the handle
 *    values
 * 3. create a container with a sequence of handles of "count" elements
 * 4. store the newly created Damsel objects in the file descriptor
 */
static
damsel_err_t mimic_ncmpi_def_dim(file_dscpt        fd,
                                 char             *name,
                                 int               dim_len,
                                 damsel_container *dim_id)
{
    int rank, nprocs, start, count;
    damsel_err_t err;

    MPI_Comm_rank(fd->comm, &rank);
    MPI_Comm_size(fd->comm, &nprocs);

    count = dim_len / nprocs;
    start = count * rank;
    if (rank < dim_len % nprocs) {
        start += rank;
        count++;
    }
    else {
        start += dim_len % nprocs;
    }

    damsel_handle start_handle = start;
    err = DMSLcontainer_create_sequence(fd->model, start_handle, count, 1, dim_id);
    if (err.id != DMSL_OK.id) return err;

    /* add the new container to the list */
    if (fd->num_containers % NUM_PER_CHUNK == 0) {
        int num_alloc = fd->num_containers + NUM_PER_CHUNK;
        fd->containers = (container_list*) realloc(fd->containers,
                                           num_alloc * sizeof(container_list));
    }
    fd->containers[fd->num_containers].container = *dim_id;;
    fd->containers[fd->num_containers].num_tags = 0;
    fd->containers[fd->num_containers].tags     = NULL;
    fd->num_containers++;

    return DMSL_OK;
}

/*----< mimic_ncmpi_def_var() >-----------------------------------------------*/
/* This program considers defining a variable as tagging a container.
 * 1. define a tag
 * 2. map the tag to a container
 * 3. store the newly created tag in the file descriptor
 */
static
damsel_err_t mimic_ncmpi_def_var(file_dscpt        fd,
                                 char             *name,
                                 damsel_data_type  type,
                                 damsel_container  dim_id,
                                 void             *buf,
                                 damsel_handle    *var_id)
{
    int i;
    damsel_err_t err;

    err = DMSLtag_define(fd->model, var_id, type, name);
    if (err.id != DMSL_OK.id) return err;

    /* find the container dim_id from the container_list*/
    for (i=0; i<fd->num_containers; i++)
        if (fd->containers[i].container == dim_id)
            break;

    /* report error if not found */
    if (i == fd->num_containers) return DMSL_EINVALCONTAINER;

    /* Tag the container and register I/O buffer (buf) */
    err = DMSLmodel_map_tag(buf, dim_id, var_id);
    if (err.id != DMSL_OK.id) return err;

    strcpy(fd->containers[i].name, name);

    /* add the new varid (tag handle) to the tag list */
    if (fd->containers[i].num_tags % NUM_PER_CHUNK == 0) {
        int num_alloc = fd->containers[i].num_tags + NUM_PER_CHUNK;
        fd->containers[i].tags = (damsel_handle*) realloc(fd->containers[i].tags,
                                 num_alloc * sizeof(damsel_handle));
    }
    fd->containers[i].tags[fd->containers[i].num_tags] = *var_id;
    fd->containers[i].num_tags++;

    return DMSL_OK;
}

/*----< mimic_ncmpi_enddef() >------------------------------------------------*/
/* In PnetCDF, calling ncmpi_enddef() is when all data definition is completed.
 * In Damsel, this is also the time for creating the file-side handles
 * Since the file descriptor has all the Damsel objects (tags and containers),
 * we will only need the descriptor for the task.
 */
static
damsel_err_t mimic_ncmpi_enddef(file_dscpt fd)
{
    int i, j;
    damsel_err_t err;

    /* invent/map file-side handle for each container */
    for (i=0; i<fd->num_containers; i++) {
        err = DMSLmodel_map_handles_inventing_file_handles(fd->containers[i].container);
        if (err.id != DMSL_OK.id) return err;

        /* invent/map a single file-side handle for each tag */ 
        for (j=0; j<fd->containers[i].num_tags; j++) {
            err = DMSLmodel_map_single_handle_inventing_file_handle(fd->model,
                  fd->containers[i].tags[j]);
            if (err.id != DMSL_OK.id) return err;
        }
    }

    return DMSL_OK;
}

/*----< mimic_ncmpi_iput() >--------------------------------------------------*/
static
damsel_err_t mimic_ncmpi_iput(file_dscpt        fd,
                              damsel_request_t *req)
{
    damsel_err_t err;
    err = DMSLmodel_transfer_async(fd->model, DAMSEL_TRANSFER_TYPE_WRITE, req);
    if (err.id != DMSL_OK.id) return err;

    return DMSL_OK;
}

/*----< mimic_ncmpi_wait() >--------------------------------------------------*/
static
damsel_err_t mimic_ncmpi_wait(file_dscpt        fd,
                              damsel_request_t  req,
                              damsel_status_t  *status)
{
    damsel_err_t err;
    /* wait for the I/O transfer to complete */
    err = DMSLmodel_wait(req, status);
    if (err.id != DMSL_OK.id) return err;

    return DMSL_OK;
}

/*----< mimic_ncmpi_close() >-------------------------------------------------*/
/* Since the file descriptor stores all Damsel objects, we can retrieve them
 * and release the resource used.
 */
static
damsel_err_t mimic_ncmpi_close(file_dscpt *fd)
{
    int i;
    damsel_err_t err;

    /* close all container handles */
    for (i=0; i<(*fd)->num_containers; i++) {
        err = DMSLcontainer_close((*fd)->containers[i].container);
        if (err.id != DMSL_OK.id) return err;
    }
    /* close in-memory store */
    err = DMSLmodel_close((*fd)->model);
    if (err.id != DMSL_OK.id) return err;

    free(*fd);

    return DMSL_OK;
}

/*
Compile:
mpicc -c -g -I/path/to/Damsel/include two_array1d.c

Link:
mpicc -o two_array1d two_array1d.o -L/path/to/Damsel/lib -L/path/to/HDF5/lib -L/path/to/PnetCDF/lib -ldamsel -lhdf5 -lhdf5_hl -lm -lz -lpnetcdf

Run: (on 2 process in parallel)
mpiexec -n 2 two_array1d output/dmsl.h5

Output: (from command h5dump)
h5dump output/dmsl.h5

HDF5 "output/dmsl.h5" {
GROUP "/" {
   DATASET "damsel-collections" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_I32LE "type";
         H5T_STD_U64LE "collection_handle";
         H5T_STD_U64LE "container_handle";
      }
      DATASPACE  SIMPLE { ( 0 ) / ( 0 ) }
      DATA {
      }
   }
   DATASET "damsel-entity-connectivity" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_I32LE "entity_type";
         H5T_STD_I32LE "vertices_per_entity";
         H5T_STD_U64LE "vertex_container_handle";
         H5T_STD_U64LE "entity_container_handle";
      }
      DATASPACE  SIMPLE { ( 0 ) / ( H5S_UNLIMITED ) }
      DATA {
      }
   }
   DATASET "damsel-handle-allocator" {
      DATATYPE  H5T_STD_I64LE
      DATASPACE  SIMPLE { ( 1 ) / ( 1 ) }
      DATA {
      (0): 2
      }
   }
   DATASET "damsel-seq-handle-containers" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_U64LE "handle";
         H5T_STD_U64LE "base";
         H5T_STD_I32LE "quantity";
         H5T_STD_I32LE "step";
      }
      DATASPACE  SIMPLE { ( 6 ) / ( H5S_UNLIMITED ) }
      DATA {
      (0): {
            8589939604,
            8589939596,
            8,
            1
         },
      (1): {
            8589939595,
            8589939594,
            1,
            1
         },
      (2): {
            8589939593,
            8589939592,
            1,
            1
         },
      (3): {
            12884906899,
            12884906892,
            7,
            1
         },
      (4): {
            12884906891,
            12884906890,
            1,
            1
         },
      (5): {
            12884906889,
            12884906888,
            1,
            1
         }
      }
   }
   DATASET "damsel-tag-containers" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_I64LE "tag_datatype";
         H5T_STD_U64LE "tag_handle";
         H5T_STRING {
            STRSIZE 256;
            STRPAD H5T_STR_NULLTERM;
            CSET H5T_CSET_ASCII;
            CTYPE H5T_C_S1;
         } "name";
      }
      DATASPACE  SIMPLE { ( 2 ) / ( H5S_UNLIMITED ) }
      DATA {
      (0): {
            4,
            8589939594,
            "flt_var1"
         },
      (1): {
            2,
            8589939592,
            "int_var1"
         }
      }
   }
   DATASET "damsel-vector-handle-containers" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_U64LE "handle";
         H5T_STD_I32LE "quantity";
         H5T_STD_I32LE "blocknumber";
         
         H5T_OPAQUE;
         OPAQUE_TAG "";
       "content";
      }
      DATASPACE  SIMPLE { ( 0 ) / ( 0 ) }
      DATA {
      }
   }
   DATASET "flt_var1-200001394" {
      DATATYPE  H5T_IEEE_F32LE
      DATASPACE  SIMPLE { ( 8 ) / ( 8 ) }
      DATA {
      (0): 0, 0, 0, 0, 0, 0, 0, 0
      }
   }
   DATASET "flt_var1-300001393" {
      DATATYPE  H5T_IEEE_F32LE
      DATASPACE  SIMPLE { ( 7 ) / ( 7 ) }
      DATA {
      (0): 1.01, 1.01, 1.01, 1.01, 1.01, 1.01, 1.01
      }
   }
   DATASET "int_var1-200001394" {
      DATATYPE  H5T_STD_I32LE
      DATASPACE  SIMPLE { ( 8 ) / ( 8 ) }
      DATA {
      (0): 0, 0, 0, 0, 0, 0, 0, 0
      }
   }
   DATASET "int_var1-300001393" {
      DATATYPE  H5T_STD_I32LE
      DATASPACE  SIMPLE { ( 7 ) / ( 7 ) }
      DATA {
      (0): 1, 1, 1, 1, 1, 1, 1
      }
   }
}
}
*/
