/* test create_vector and writing to disk */

#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <mpi.h>
#include <damsel.h>

#define CHECK_ERR(e) {if ((e.id!=DMSL_OK.id)) { printf("Error at line=%d msg=%s\n", __LINE__,DMSLerror_str(e)); goto fn_exit;}}

int main(int argc, char **argv) {
    char *filename;
    int rank, nprocs;
    damsel_err_t err;
#ifndef DAMSEL_SINGLETHREAD
    int res;
#endif
    short int buf[] = {3, 1, 4, 1, 5}; // some arbitrary data to write

#ifndef DAMSEL_SINGLETHREAD
    MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
    assert(res == MPI_THREAD_MULTIPLE);
#else
    MPI_Init(&argc, &argv);
#endif

    /* initialize Damsel environment */
    damsel_library lib;
    err = DMSLlib_init(&lib);
    CHECK_ERR(err)

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
    if (argc != 2) {
        if (!rank) printf("Usage: %s filename\n",argv[0]);
        goto fn_exit;
    }
    filename = argv[1];

    /* create a model, an in-memory store */
    damsel_model model;
    err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
    CHECK_ERR(err)

    /* register a tag named 'somedata' to the model, and the data to be
     * tagged by this tag is of short int type. The value of the handle,
     * 123 in this example, is given by the user.
     */
    damsel_handle tag_handle = 123;
    err = DMSLtag_define(model, &tag_handle, DAMSEL_DATA_TYPE_SHORTINT, "somedata");
    CHECK_ERR(err)

    /* create a vector handle collection with 5 handles in it. */
    damsel_handle vector_of_handles[] = { 0x20000, 0x21000, 0x22000,
                                          0x23000, 0x24000 };
    damsel_container container_id;
    err = DMSLcontainer_create_vector(model, vector_of_handles, 5, &container_id);
    CHECK_ERR(err)

    /* Tag the container and register I/O buffer (buf) */
    err = DMSLmodel_map_tag(buf, container_id, &tag_handle);
    CHECK_ERR(err)

    /* At this point, the in-memory store knows about all the data, as the
     * container has a tag and an I/O buffer pointing to the data. We are ready
     * to write the data out to disk. First, we need to invent the file-side
     * handles by mapping the in-memory store (model) to the file store. */

    /* make sure the file does not exist, ignore error if any */
    unlink(filename);

    /* set the file access mode */
    damsel_trait_bundle tb;
    err = DMSLtrait_create(&tb);
    CHECK_ERR(err)
    err = DMSLtrait_put(tb, "mode", "WRITE");
    CHECK_ERR(err)

    /* register the filename and access mode to model (in-memory store) */
    err = DMSLmodel_attach(model, filename, MPI_COMM_WORLD, tb);
    CHECK_ERR(err)

    err = DMSLtrait_release(tb);
    CHECK_ERR(err)

    /* invent/map file-side handle for each container */
    err = DMSLmodel_map_handles_inventing_file_handles(container_id);
    CHECK_ERR(err)

    /* create a container containing a vector of tags (one tag in this
     * example) */ 
    damsel_handle tag_collection_vector[] = {tag_handle};
    damsel_container tag_collection_id;
    err = DMSLcontainer_create_vector(model, tag_collection_vector,
                                      /* count = */ 1, &tag_collection_id);
    CHECK_ERR(err)

    /* maps the vector container into the file store */
    err = DMSLmodel_map_handles_inventing_file_handles(tag_collection_id);
    CHECK_ERR(err)

    /* Now we can tell damsel to execute everything. */
    damsel_request_t req;
    err = DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_WRITE, &req);
    CHECK_ERR(err)

    /* Other computations can be performed here, i.e. between the post of
     * asynchronous transfer and wait, as long as the registered data buffer,
     * buf, is not touched */

    /* wait for the I/O transfer to complete */
    damsel_status_t status;
    err = DMSLmodel_wait(req, &status);
    CHECK_ERR(err)

    /* close in-memory store */
    err = DMSLmodel_close(model);
    CHECK_ERR(err)

    /* close container handle */
    err = DMSLcontainer_close(tag_collection_id);
    CHECK_ERR(err)

    err = DMSLcontainer_close(container_id);
    CHECK_ERR(err)

fn_exit:
    /* exit Damsel environment */
    if (lib != DAMSEL_LIBRARY_INVALID) {
        err = DMSLlib_finalize(lib);
        CHECK_ERR(err)
    }

    MPI_Finalize();
    return 0;
}

/*
 Compile:
 mpicc -c -g -I/path/to/Damsel/include vector.c

 Link:
 mpicc -o vector vector.o -L/path/to/Damsel/lib -L/path/to/HDF5/lib -L/path/to/PnetCDF/lib -ldamsel -lhdf5 -lhdf5_hl -lm -lz -lpnetcdf

 Run: (on 2 process in parallel)
 mpiexec -n 2 vector output/dmsl.h5

 Output: (from command h5dump)
 h5dump output/dmsl.h5

HDF5 "output/dmsl.h5" {
GROUP "/" {
   DATASET "damsel-collections" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_I32LE "type";
         H5T_STD_U64LE "collection_handle";
         H5T_STD_U64LE "container_handle";
      }
      DATASPACE  SIMPLE { ( 0 ) / ( 0 ) }
      DATA {
      }
   }
   DATASET "damsel-entity-connectivity" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_I32LE "entity_type";
         H5T_STD_I32LE "vertices_per_entity";
         H5T_STD_U64LE "vertex_container_handle";
         H5T_STD_U64LE "entity_container_handle";
      }
      DATASPACE  SIMPLE { ( 0 ) / ( H5S_UNLIMITED ) }
      DATA {
      }
   }
   DATASET "damsel-handle-allocator" {
      DATATYPE  H5T_STD_I64LE
      DATASPACE  SIMPLE { ( 1 ) / ( 1 ) }
      DATA {
      (0): 2
      }
   }
   DATASET "damsel-seq-handle-containers" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_U64LE "handle";
         H5T_STD_U64LE "base";
         H5T_STD_I32LE "quantity";
         H5T_STD_I32LE "step";
      }
      DATASPACE  SIMPLE { ( 4 ) / ( H5S_UNLIMITED ) }
      DATA {
      (0): {
            8589939599,
            8589939594,
            5,
            1
         },
      (1): {
            8589939593,
            8589939592,
            1,
            1
         },
      (2): {
            12884906895,
            12884906890,
            5,
            1
         },
      (3): {
            12884906889,
            12884906888,
            1,
            1
         }
      }
   }
   DATASET "damsel-tag-containers" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_I64LE "tag_datatype";
         H5T_STD_U64LE "tag_handle";
         H5T_STRING {
            STRSIZE 256;
            STRPAD H5T_STR_NULLTERM;
            CSET H5T_CSET_ASCII;
            CTYPE H5T_C_S1;
         } "name";
      }
      DATASPACE  SIMPLE { ( 1 ) / ( H5S_UNLIMITED ) }
      DATA {
      (0): {
            7,
            8589939592,
            "somedata"
         }
      }
   }
   DATASET "damsel-vector-handle-containers" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_U64LE "handle";
         H5T_STD_I32LE "quantity";
         H5T_STD_I32LE "blocknumber";
         
         H5T_OPAQUE;
         OPAQUE_TAG "";
       "content";
      }
      DATASPACE  SIMPLE { ( 0 ) / ( 0 ) }
      DATA {
      }
   }
   DATASET "somedata-20000138f" {
      DATATYPE  H5T_STD_I16LE
      DATASPACE  SIMPLE { ( 5 ) / ( 5 ) }
      DATA {
      (0): 3, 1, 4, 1, 5
      }
   }
   DATASET "somedata-30000138f" {
      DATATYPE  H5T_STD_I16LE
      DATASPACE  SIMPLE { ( 5 ) / ( 5 ) }
      DATA {
      (0): 3, 1, 4, 1, 5
      }
   }
}
}
 */
