/* this is connectivity metadata routines */

#include <assert.h>
#include <stdio.h>

#include "damsel.h"
#include "damsel-connectivity.h"
#include "damsel-ndc.h"
#include "damsel-hdf5.h"
#include "damsel-debug.h"

#include <hdf5.h>

/* below is some entity API stuff */

/* how much can I build on top of tags? */

/* maybe I can have a damsel internal tag that stores the entity definitions
   as declared by DMSLentity_define - the tag would be on handle containers,
   which complicates lookup: there might be multiple handle containers
   covering a particular handle, and I'd need to find each one and then see
   if it had an entity definition.

   How, then, would I store connectivity in tags?

   If things are being declared as handle containers -> handle containers,
   then I can straightforwardly store stuff by associating the two handle
   containers in the file without any further processing.

   That's probably awkward to access later too.
   But whatever - this will do for a prototype, and as query uses arrive
   later, more can be implemented.

   So, I think, I'll need two more datasets in the HDF5 file - one for
   the entity definitions, and one to store connectivity.

   The connectivity one stores pairs of handles.
   The entity definition one stores the parameters to entity_define:
     handle of container, entity type and (normalised?) vertices per entity
     (normalised means when the entity type implies the vertices per entity,
      then that value)
*/

static hid_t get_connectivity_type(damsel_attached_file *file) {
  return file->hdf5_connectivity_type;
}

void initialize_connectivity_type(damsel_attached_file *file) {
  int ret;
  hid_t connectivity_type = H5Tcreate(H5T_COMPOUND,
    sizeof(damsel_infile_entity_definition));
  assert(connectivity_type >= 0);

  ret = H5Tinsert(connectivity_type, "entity_type",
    HOFFSET(damsel_infile_entity_definition, entity_type),
    H5T_NATIVE_INT);
  assert(ret >= 0);

  ret = H5Tinsert(connectivity_type, "vertices_per_entity",
    HOFFSET(damsel_infile_entity_definition, vertices_per_entity),
    H5T_NATIVE_INT);
  assert(ret >= 0);

  ret = H5Tinsert(connectivity_type, "vertex_container_handle",
    HOFFSET(damsel_infile_entity_definition, vertex_container_handle),
    H5T_NATIVE_ULLONG);
  assert(ret >= 0);

  ret = H5Tinsert(connectivity_type, "entity_container_handle",
    HOFFSET(damsel_infile_entity_definition, entity_container_handle),
    H5T_NATIVE_ULLONG);
  assert(ret >= 0);

  file->hdf5_connectivity_type = connectivity_type;
}


damsel_err_t DMSLentity_define(damsel_container app_container,
                               damsel_entity_type entity_type,
                               size_t vertices_per_entity,
                               damsel_container vertex_container) {
  push_printf_ndc("DMSLentity_define %p %d %zd %p", app_container, (int)entity_type, vertices_per_entity, vertex_container);

  assert(entity_type >= DAMSEL_ENTITY_TYPE_VERTEX);
  assert(entity_type < DAMSEL_ENTITY_TYPE_UNDEFINED);
  assert_valid_handle_container_id(app_container);

  DMSLcontainer_take(app_container);
  DMSLcontainer_take(vertex_container);

  damsel_container_internal *handle_container_struct = (damsel_container_internal *)app_container;
  damsel_store_internal *store = (damsel_store_internal *)handle_container_struct->model;
  assert_valid_store_id(store);
  assert(store->type == DAMSEL_STORETYPE_MEMORY);
  damsel_model_internal *model = (damsel_model_internal *)store;

  damsel_internal_entity_definition *intdef = malloc(sizeof(damsel_internal_entity_definition));
  assert(intdef!=NULL);
  intdef->vertex_container_id = vertex_container;
  intdef->entity_container_id = app_container;
  intdef->entity_type = entity_type;
// TODO: normalize based on entity_type (or assert that normalization?)
  intdef->vertices_per_entity = vertices_per_entity;

  linkedlist_prepend(&model->first_entity_definition, intdef);

  pop_ndc();
  return DMSL_OK;
}

damsel_err_t DMSLentity_define_fast(damsel_container entity_container_id, damsel_entity_type ent_type, size_t vertices_per_entity, damsel_handle_ptr vertex_handles) {

  damsel_err_t err;
  size_t num_entities = ((damsel_container_internal *)entity_container_id)->quantity;

  size_t count = num_entities * vertices_per_entity;

  damsel_model model = ( (damsel_container_internal *) entity_container_id)->model;
  damsel_container hc_id;
  err = DMSLcontainer_create_vector(model, vertex_handles, count, &hc_id);
  assert_valid_handle_container_id(hc_id);
  damsel_err_t ret = DMSLentity_define(entity_container_id, ent_type, vertices_per_entity, hc_id);
  DMSLcontainer_release(hc_id);
  return ret;
}

int DMSLmodel_get_entity_count(damsel_model m) {
  push_ndc("DMSLentity_get_count");

  damsel_model_internal *ms = (damsel_model_internal *)m;
  damsel_attached_file *file = ms->attached_file;

  hid_t spaceid = H5Dget_space(file->entities_dataset);
  assert(spaceid >= 0);

  hssize_t size;
  size = H5Sget_simple_extent_npoints(spaceid);

  printf_ndc("count of entity definitions: %d", (int)size);

  H5Sclose(spaceid);

  pop_ndc();
  return size;
}

damsel_err_t DMSLmodel_get_entity_infos(damsel_model m, damsel_entity_buf_type *buf) {
  push_ndc("DMSLentity_get_entity_infos");

  damsel_model_internal *ms = (damsel_model_internal *)m;
  damsel_attached_file *file = ms->attached_file;

  hid_t spaceid = H5Dget_space(file->entities_dataset);
  assert(spaceid >= 0);
  hssize_t size;
  size = H5Sget_simple_extent_npoints(spaceid);

  printf_ndc("count of entity definitions: %d", (int)size);

  hid_t datatype = get_connectivity_type(file);

  damsel_infile_entity_definition *conndata = malloc(size * sizeof(damsel_infile_entity_definition));

  int ret = H5Dread(file->entities_dataset, datatype, spaceid, spaceid,
    H5P_DEFAULT, conndata);

  printf_ndc("H5Dread returned %d", ret);

  H5Sclose(spaceid);

  int i;
  for(i=0; i < size; i++) {
    push_printf_ndc("transcribing entry %d", i);
    buf[i].entity_type = conndata[i].entity_type;
    buf[i].vertices_per_entity = conndata[i].vertices_per_entity;
    buf[i].entity_container = (damsel_container) damsel_handle_to_handle_container(file, conndata[i].entity_container_handle);
    assert_valid_handle_container_id(buf[i].entity_container);
    DMSLcontainer_take(buf[i].entity_container);
    buf[i].count = DMSLcontainer_count(buf[i].entity_container);
    buf[i].vertex_container = (damsel_container) damsel_handle_to_handle_container(file, conndata[i].vertex_container_handle);
    assert_valid_handle_container_id(buf[i].vertex_container);
    DMSLcontainer_take(buf[i].vertex_container);

// now need to generate a container of (file-side) tags that have values set
// on any member of this entity set.

    // this is pretty unscalable... but it should give me the semantics
    // for now.

    buf[i].tag_handle_container = container_of_tags_which_are_set_on(m, buf[i].entity_container);
    if(buf[i].tag_handle_container != NULL)
      buf[i].tag_count = DMSLcontainer_count(buf[i].tag_handle_container);
    else
      buf[i].tag_count = 0;

    // TODO: there are struct members for tags that need implementing.
    pop_ndc();
  }
  free(conndata);

  pop_ndc();
  return DMSL_OK;
}

damsel_err_t DMSLmodel_close_entity_infos(damsel_model m, damsel_entity_buf_type *buf) {
  int count = DMSLmodel_get_entity_count(m);
  int i;
  for(i=0;i<count;i++) {
    if(buf[i].tag_handle_container) DMSLcontainer_release(buf[i].tag_handle_container);
    if(buf[i].vertex_container) DMSLcontainer_release(buf[i].vertex_container);
    if(buf[i].entity_container) DMSLcontainer_release(buf[i].entity_container);
  }
  return DMSL_OK;
}

/* By the time store_entities runs, all mapping should have been
   done, so we should know 

   assumption here is that we're attached to an HDF5 file, so we
   should be able to do the same sort of stuff

   also, this should be called on all ranks simultaneously, as a
   collective operation.
*/
void store_entities(damsel_model_internal *s, damsel_attached_file *fs) {
  int local_connectivity_count;
  int myrank, maxrank;
  int i;
  push_ndc("in entities postmap");

  //damsel_attached_file *fs = s->attached_file;

  myrank = fs->myrank;
  maxrank = fs->maxrank;

  push_ndc("creating connectivity dataset");
  local_connectivity_count = linkedlist_length(s->first_entity_definition);
  printf_ndc("rank %d has %d connectivity definitions", myrank,
    local_connectivity_count);
  pop_ndc();

  int *per_rank_connectivity_count;
  per_rank_connectivity_count = malloc(sizeof(int) * maxrank);
  assert(per_rank_connectivity_count != NULL);

  MPI_Allgather(&local_connectivity_count, 1, MPI_INT,
                per_rank_connectivity_count, 1, MPI_INT,
                 fs->communicator);

  hsize_t hc_rows = 0;

  for(i=0; i<maxrank; i++) {
    hc_rows += per_rank_connectivity_count[i];
  }
  printf_ndc("rank %d counted %d connectivity rows in total", myrank,
    (int)hc_rows);
  hid_t connectivity_type = get_connectivity_type(fs);
  hid_t hc_dataset = -1;
  hid_t hc_dataspace = -1;
  const hsize_t hsize_one = 1;

  hsize_t write_offset = 0;

  if (H5Lexists(fs->file_id, DAMSEL_DATASET_CONNECTIVITY, H5P_DEFAULT) <= 0) {

    hsize_t hc_unlimited = H5S_UNLIMITED;
    hc_dataspace = H5Screate_simple(1, &hc_rows, &hc_unlimited);
    assert(hc_dataspace >= 0);

    hid_t connectivity_properties = H5Pcreate (H5P_DATASET_CREATE);
    assert(connectivity_properties != -1);

    int ret = H5Pset_chunk(connectivity_properties, 1, &hsize_one);
// need to pick a chunk size here. I'm setting it to 1 here.. that's rather arbitrary though. it presumably will have some effect on performance...

    assert(ret != -1);

    hc_dataset = H5Dcreate(fs->file_id, DAMSEL_DATASET_CONNECTIVITY,
      connectivity_type, hc_dataspace, H5P_DEFAULT, connectivity_properties, H5P_DEFAULT);
    assert(hc_dataset >= 0);
    ret = H5Pclose(connectivity_properties);
    assert(ret >= 0);
  } else {

    // need to extend the dataspace using h5t_extend.

    // hc_rows comes from the calculation we just did.
    // hc_existing_rows will be used later to determine the offset
    // for writing into the dataset

    hc_dataset = H5Dopen(fs->file_id, DAMSEL_DATASET_CONNECTIVITY, H5P_DEFAULT);

    hc_dataspace = H5Dget_space(hc_dataset);
    assert(hc_dataspace >= 0);

    hsize_t hc_existing_rows;
    hc_existing_rows = H5Sget_simple_extent_npoints(hc_dataspace);

    hsize_t hc_new_rows = hc_existing_rows + hc_rows;
    H5Dset_extent(hc_dataset, &hc_new_rows);
    write_offset = hc_existing_rows; 
  }
  assert(hc_dataspace >= 0);
  assert(hc_dataset >= 0);

  push_ndc("populating connectivity dataset");

  hid_t hc_dataspace_one = H5Screate_simple(1, &hsize_one, &hsize_one);
  assert(hc_dataspace_one >= 0);
  linkedlist *idlist = s->first_entity_definition;

  int my_hc_base = write_offset;
  for(i=0; i<myrank; i++) {
    my_hc_base += per_rank_connectivity_count[i];
  }

  printf_ndc("rank %d will put connectivity rows start at offset %d",
    myrank, my_hc_base);

  hid_t hc_plist_id = H5Pcreate (H5P_DATASET_XFER);
  H5Pset_dxpl_mpio (hc_plist_id, H5FD_MPIO_INDEPENDENT);
  assert(hc_plist_id >= 0);

  i=0;
  while(idlist!=NULL) {
    printf_ndc("stepping through internal entities list");

    assert(i<local_connectivity_count);

    damsel_internal_entity_definition *internal = (damsel_internal_entity_definition *)idlist->head;
    damsel_infile_entity_definition infile_entity_definition;

    infile_entity_definition.entity_type = internal->entity_type;
    infile_entity_definition.vertices_per_entity = internal->vertices_per_entity;

    damsel_container_internal *vertex_handle_container = (damsel_container_internal *)internal->vertex_container_id;
    damsel_container_internal *entity_handle_container = (damsel_container_internal *)internal->entity_container_id;

    damsel_err_t ret;
    ret = internal_push_container((damsel_container) vertex_handle_container, &infile_entity_definition.vertex_container_handle);
    assert_dmsl_ok(ret);
    ret = internal_push_container((damsel_container) entity_handle_container, &infile_entity_definition.entity_container_handle);
    assert_dmsl_ok(ret);

    int res = H5Sselect_none(hc_dataspace);
    assert(res >= 0);

    hsize_t coords = my_hc_base + i;
    res = H5Sselect_elements(hc_dataspace, H5S_SELECT_SET, 1, &coords);
    assert(res >= 0);
    res = H5Dwrite(hc_dataset, connectivity_type, hc_dataspace_one,
      hc_dataspace, hc_plist_id, &infile_entity_definition);
    assert(res >= 0);

    idlist=idlist->next;
    i++;
  }
  pop_ndc();

  H5Dclose(hc_dataset);
  H5Sclose(hc_dataspace);
  H5Sclose(hc_dataspace_one);

  free(per_rank_connectivity_count);

  pop_ndc();
}

/* this gets called to open bits of HDF5 for  */
void entities_initialize_for_read(damsel_attached_file *af) {

  push_ndc("entities_initialize_for_read");

  if(H5Lexists(af->file_id, DAMSEL_DATASET_CONNECTIVITY, H5P_DEFAULT) > 0) {
    af->entities_dataset = H5Dopen(af->file_id, DAMSEL_DATASET_CONNECTIVITY,
      H5P_DEFAULT);
    assert(af->entities_dataset >= 0);
  } else {
    af->entities_dataset = -1;
  }
  pop_ndc();
}

/* finds an entity definition where h is declared as an entity. */
void *find_entity_definition_for_handle(damsel_model_internal *m,
  damsel_handle h, int *pos) {

  push_printf_ndc("find_entity_definition_for_handle, model = %p, handle = %llx", m, h);

  linkedlist *ll;

  for(ll = m->first_entity_definition; ll!=NULL; ll=ll->next) {
    damsel_internal_entity_definition *d = (damsel_internal_entity_definition *) ll->head;
    // if the supplied handle is in the list of definitions, then this is
    // the one...
    int ix;
    push_printf_ndc("checking entity definition %p", d);
    ix = find_handle_in_container((damsel_container_internal *)d->entity_container_id, h);
    if(ix >= 0) {
      printf_ndc("found");
      pop_ndc();
      pop_ndc();
      *pos = ix;
      return d;
    }
    printf_ndc("not found");
    *pos = -1;
    pop_ndc();
  }
  pop_ndc();
  return NULL;
}

/* finds an entity definition where h is used as a vertex */
damsel_internal_entity_definition *find_vertex_definition_for_handle(damsel_model_internal *m,
  damsel_handle h, int *pos) {
  push_printf_ndc("find_vertex_definition_for_handle. model = %p, handle = %llx, will put returned pos into *%p", m, h, pos);

  linkedlist *ll;

  for(ll = m->first_entity_definition; ll!=NULL; ll=ll->next) {
    damsel_internal_entity_definition *d = (damsel_internal_entity_definition *) ll->head;
    // if the supplied handle is in the list of definitions, then this is
    // the one...
    int ix;
    push_printf_ndc("checking entity definition %p", d);
    ix = find_handle_in_container((damsel_container_internal *)d->vertex_container_id, h);
    if(ix >= 0) {
      printf_ndc("found");
      pop_ndc();
      pop_ndc();
      *pos = ix;
      return d;
    }
    printf_ndc("not found");
    *pos = -1;
    pop_ndc();
  }
  pop_ndc();
  return NULL;
}

/* a mapping permutation which arranges handles by connectivity */
void damsel_permute_connectivity(damsel_container appside, damsel_container fileside) {
  push_ndc("damsel_permute_connectivity");

  assert_valid_handle_container_id(appside);
  assert_valid_handle_container_id(fileside);

  // need to make a new vector container, that contains all of the entries
  // then I can have a think about how to make new ones.

  int count = DMSLcontainer_count(appside);
  damsel_err_t err;

#ifndef NDEBUG
  int fcount = DMSLcontainer_count(fileside);
  assert(count == fcount);
#endif

  damsel_handle *v = malloc(count * sizeof(damsel_handle));
  assert(v!=NULL);
  int vpos=0;

  damsel_model m = ((damsel_container_internal *) appside)->model;

  // status for each entry of appside
  // 0 means unprocessed
  // 1 means processed

  int *status = malloc(count * sizeof(int));
  assert(status != NULL);

  int ix;
  for(ix=0; ix<count; ix++) {
    status[ix] = 0;
  }

  while(vpos < count) {

    // find the first entry that has status 0
    int here;
    for(here = 0; here < count && status[here] != 0; here++);
    if(here >= count) { // we're done
    } else {
      // if proceed == 1 when we come out of the loop, it means we've moved
      // here to point to a new unprocessed entry.
      int proceed = 1;
      while(proceed == 1) {
        assert(here < count);
        assert(status[here] == 0);
        // record here.
        assert(vpos < count);
        damsel_handle h = DMSLcontainer_handle_at_position(appside, here);
        v[vpos] = h;
        status[here] = 1;
        vpos++;
        assert(vpos <= count);

        // now I've put that in the vector, can I find a new 'here' that is
        // close by topologically?

        proceed = 0; // because we haven't moved 'here'.

// we'll only find one entity definition here, even if the vertices are used in multiple entity definitions (for example, the faces of cubes, and the cubes themselves. depending on which we find, we'll get a different layout. */

        int pos;
        damsel_internal_entity_definition *entity_def;
        entity_def = find_vertex_definition_for_handle((damsel_model_internal *)m, h, &pos);
        if(entity_def == NULL) {
          // lets see if we can find an entity definition based on the handle being an entity handle:
          entity_def = find_entity_definition_for_handle((damsel_model_internal *)m, h, &pos);
// TODO here
          if(entity_def != NULL) {
            int num_vertices_per_entity = entity_def->vertices_per_entity;

            int vertex_offset = pos * num_vertices_per_entity;
            int tryvertex;
            for(tryvertex = 0 ; tryvertex < num_vertices_per_entity && proceed == 0; tryvertex++) {
              int trypos = vertex_offset + tryvertex;
              int v = DMSLcontainer_handle_at_position(entity_def->vertex_container_id, trypos);

              // now we have v, we need to find another entity that has a similar vertex, and that
              // has not already been output.

              int size2= DMSLcontainer_count(entity_def->vertex_container_id);
              int tryvertex2;
              for(tryvertex2 = 0; tryvertex2 < size2 && proceed == 0; tryvertex2++) {
                int v2 = DMSLcontainer_handle_at_position(entity_def->vertex_container_id, tryvertex2);
                if(v == v2) {
                  int entity2 = pos / num_vertices_per_entity;
                  assert(entity2 < count);
                  if(status[entity2] == 0) {
                    here = entity2;
                    proceed = 1;
                    printf_ndc("Successfully moved to position %d based on entities", here);
                  }
                }
              }
            }
          } else {
            printf_ndc("Trying to look up an entity definition which contains handle %llx as a vertex but could not find any such definition.", h);
            // we will continue in the loop without trying to find thing nearby, because we have no definition of nearby.
          }
        } else { // we did find a vertex definition (TODO: swap the sense of this if statement so that it flows more linearly)
          assert(entity_def != NULL);

          int num_vertices_per_entity = entity_def->vertices_per_entity;
          int entitypos = pos/num_vertices_per_entity;
          int vertex = pos%num_vertices_per_entity;
          printf_ndc("entitypos = %d, vertex = %d", entitypos, vertex);
          int tryvertex;
          for (tryvertex=0; tryvertex < num_vertices_per_entity && proceed == 0 ; tryvertex++) {
            int trypos = entitypos*num_vertices_per_entity + tryvertex;

            damsel_handle tryh = DMSLcontainer_handle_at_position(entity_def->vertex_container_id, trypos);
            printf_ndc("trying to find handle %llx", tryh);

            int tryhere = find_handle_in_container((damsel_container_internal *)appside, tryh);
            assert(tryhere >= 0);
            printf_ndc("trying position %d out of %d", tryhere, count);
            assert(tryhere < count);
            if(status[tryhere] == 0) {
              here = tryhere;
              proceed = 1;
              printf_ndc("Successfully moved to position %d based on vertices", here);
            } else {
              printf_ndc("Position %d already processed - not moving", tryhere);
            }
          }
        }
      }
    }
  }

  // when we reach here, the v array must contain all of the handles in
  // appside, permuted.

  assert(vpos == count);
  damsel_container newappside;
  err = DMSLcontainer_create_vector(m, v, count, &newappside);
  free(v);
  printf_ndc("newappside container = %p", newappside);

#ifndef NDEBUG
  push_ndc("checking permutation valid");
  int d_ix;
  for(d_ix = 0; d_ix < count; d_ix++) {
   damsel_handle h = DMSLcontainer_handle_at_position(newappside, d_ix);
   printf_ndc("Position %d has handle %llx", d_ix, h);
  }

  for(d_ix = 0; d_ix < count; d_ix++) {
   damsel_handle h = DMSLcontainer_handle_at_position(appside, d_ix);
   int e_ix;
   int found = 0;
   for(e_ix = 0; e_ix < count; e_ix++) {
     damsel_handle h2 = DMSLcontainer_handle_at_position(newappside, e_ix);
     if(h == h2) found=1;
   }
   assert(found == 1);
  }
  pop_ndc();
#endif

  free(status);

  DMSLmodel_map_handles(newappside, fileside);

  err = internal_push_container(newappside, NULL);
  assert(err.id == 0);

  DMSLcontainer_release(newappside);
  pop_ndc();
}

