
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>

#include <hdf5.h>
#include <hdf5_hl.h>

#ifdef DAMSEL_HDF5_MULTI
#ifndef HAVE_H5DWRITE_MULTI
# include "h5multidataset.h"
#endif
#endif

#include "damsel.h"
#include "damsel-hdf5.h"
#include "damsel-internal.h"
#include "damsel-ndc.h"
#include "damsel-types.h"
#include "damsel-debug.h"
#include "damsel-connectivity.h"
#include "damsel-versioning.h"
#ifdef DAMSEL_SUBFILE
#include "damsel-subfile.h"
#endif

#define  DEFAULT_NCHUNKS 100

static const hsize_t hsize_one = 1;

void *vlen_to_array_data(size_t npoints, damsel_vector_t *vector_data, hid_t tid);
size_t get_max_vlen_length(size_t npoints, damsel_vector_t *vector_data);
void *damsel_data_to_hdf5_data(damsel_model_internal *src_s, damsel_attached_file *fs, 
      damsel_cached_data_internal *dcd, damsel_tag_internal *tagdef, 
	  const char *datasetname, hsize_t hsize_mem);

static void initialize_sequence_handle_container_type(damsel_attached_file *file) {
  int ret;
  // define a compound type for our data set
  hid_t container_type = H5Tcreate(H5T_COMPOUND, sizeof(damsel_hdf5_infile_seq_handle_container));
  assert(container_type >= 0);
  ret = H5Tinsert(container_type, "handle", HOFFSET(damsel_hdf5_infile_seq_handle_container, handle), H5T_NATIVE_ULLONG);
  assert(ret >= 0);
  ret = H5Tinsert(container_type, "base", HOFFSET(damsel_hdf5_infile_seq_handle_container, base), H5T_NATIVE_ULLONG);
  assert(ret >= 0);
  ret = H5Tinsert(container_type, "quantity", HOFFSET(damsel_hdf5_infile_seq_handle_container, quantity), H5T_NATIVE_INT);
  assert(ret >= 0);
  ret = H5Tinsert(container_type, "step", HOFFSET(damsel_hdf5_infile_seq_handle_container, step), H5T_NATIVE_INT);
  assert(ret >= 0);
  file->hdf5_sequence_type = container_type;
}

static hid_t get_sequence_handle_container_type(damsel_attached_file *file) {
  return file->hdf5_sequence_type;
}

static void initialize_vector_handle_container_type(damsel_attached_file *file) {
  int ret;
  // define a compound type for our data set
  hid_t container_type = H5Tcreate(H5T_COMPOUND, sizeof(damsel_hdf5_infile_vector_handle_container_row));
  assert(container_type >= 0);
  ret = H5Tinsert(container_type, "handle", HOFFSET(damsel_hdf5_infile_vector_handle_container_row, handle), H5T_NATIVE_ULLONG);
  assert(ret >= 0);
  ret = H5Tinsert(container_type, "quantity", HOFFSET(damsel_hdf5_infile_vector_handle_container_row, quantity), H5T_NATIVE_INT);
  assert(ret >= 0);
  ret = H5Tinsert(container_type, "blocknumber", HOFFSET(damsel_hdf5_infile_vector_handle_container_row, blocknumber), H5T_NATIVE_INT);
  assert(ret >= 0);

  hid_t contenttype = H5Tcopy(H5T_NATIVE_OPAQUE);
  assert(contenttype >= 0);
  ret = H5Tset_size(contenttype, VECTOR_MAX*sizeof(long long int));
  assert(ret >= 0);
  ret = H5Tinsert(container_type, "content", HOFFSET(damsel_hdf5_infile_vector_handle_container_row, content), contenttype);
  assert(ret >= 0);
  file->hdf5_vector_type = container_type;
}

static hid_t get_vector_handle_container_type(damsel_attached_file *file) {
  return file->hdf5_vector_type;
}


static void initialize_tag_container_type(damsel_attached_file *file) {
  hid_t container_type = H5Tcreate(H5T_COMPOUND, sizeof(damsel_hdf5_infile_tag_container));
  assert(container_type >= 0);

  hid_t string_type = H5Tcopy(H5T_C_S1);
  assert(string_type >= 0);

  int ret = H5Tset_size(string_type, DAMSEL_HDF5_MAX_TAG_SIZE);
  assert(ret >= 0);

  ret = H5Tinsert(container_type, "tag_datatype", HOFFSET(damsel_hdf5_infile_tag_container, tag_datatype), H5T_NATIVE_LONG);
  assert(ret >= 0);

  ret = H5Tinsert(container_type, "tag_handle", HOFFSET(damsel_hdf5_infile_tag_container, tag_handle), H5T_NATIVE_ULLONG);
  assert(ret >= 0);

  ret = H5Tinsert(container_type, "name", HOFFSET(damsel_hdf5_infile_tag_container, name), string_type);
  assert(ret >= 0);

  file->hdf5_tag_type = container_type;
}

static hid_t get_tag_container_type(damsel_attached_file *file) {
  return file->hdf5_tag_type;
}

void initialize_collection_type(damsel_attached_file *file) {
  int ret;

  hid_t collection_type = H5Tcreate(H5T_COMPOUND, sizeof(damsel_hdf5_infile_collection));
  assert(collection_type >= 0);

  ret = H5Tinsert(collection_type, "type", HOFFSET(damsel_hdf5_infile_collection, type), H5T_NATIVE_INT);
  assert(ret >= 0);

  ret = H5Tinsert(collection_type, "collection_handle", HOFFSET(damsel_hdf5_infile_collection, collection_handle), H5T_NATIVE_ULLONG);
  assert(ret >= 0);

  ret = H5Tinsert(collection_type, "container_handle", HOFFSET(damsel_hdf5_infile_collection, container_handle), H5T_NATIVE_ULLONG);
  assert(ret >= 0);

  file->hdf5_collection_type = collection_type;
}

hid_t get_collection_type(damsel_attached_file *file) {
  return file->hdf5_collection_type;
}

static void initialize_hdf5_types(damsel_attached_file *s) {
  initialize_collection_type(s);
  initialize_tag_container_type(s);
  initialize_sequence_handle_container_type(s);
  initialize_vector_handle_container_type(s);
  initialize_connectivity_type(s);
  initialize_versioning_type(s);
}

static hid_t open_dataset_if_exists(hid_t file_id, char *name) {
  if(H5Lexists(file_id, name, H5P_DEFAULT) > 0) {
    hid_t ds = H5Dopen(file_id, name, H5P_DEFAULT);
    assert(ds >= 0);
    return ds;
  }
  else {
    return -1;
  }
}

static damsel_attached_file *damsel_internal_hdf5_reopen(const char *fn, MPI_Comm comm, damsel_trait_bundle tb, char *mode) {
  assert(fn != NULL);

  push_printf_ndc("reopening parallel hdf5 store for file '%s'", fn);

  damsel_attached_file *f = malloc(sizeof(damsel_attached_file));
  printf_ndc("store id = %p", f);
  assert(f!=NULL);
  f->store.type = DAMSEL_STORETYPE_HDF5;
  f->store.refcount = 1;
  f->first_pushed_container = NULL;
  f->remote_mappings = NULL;
  f->model_populated = NULL;
  f->container_push_allowed = 1;
  f->traits = tb;
  f->handle_allocator = 0;
  if(tb!=NULL) DMSLtrait_take(tb);
  initialize_hdf5_types(f);
  MPI_Comm_dup(comm, &(f->communicator));

  // non-collective
  hid_t mpiprop = H5Pcreate (H5P_FILE_ACCESS);
  assert(mpiprop != -1);
  // non-collective
  int ret = H5Pset_fapl_mpio(mpiprop, comm, MPI_INFO_NULL);
  assert(ret != -1);

  int ret_and_errno[2];
  MPI_Comm_rank(f->communicator, &(f->myrank));
  MPI_Comm_size(f->communicator, &(f->maxrank));
  if (f->myrank == 0) {
      struct stat statbuf;
      ret_and_errno[0] = stat(fn, &statbuf);
      ret_and_errno[1] = errno;
  }
  MPI_Bcast(ret_and_errno, 2, MPI_INT, 0, f->communicator);
  // collective
  printf_ndc("return is %d with errno %d", ret_and_errno[0], ret_and_errno[1] );
  if(ret_and_errno[0] == -1 && ret_and_errno[1] == ENOENT) {
      /* file does not exist.  are we reading or writing? */
      if (!strcmp("READ", mode) )  {
	  printf_ndc("Cannot open/reopen a file that does not exist\n");
	  goto fn_error;
      } else {
	  printf_ndc("Creating file");
	  f->file_id = H5Fcreate(fn, H5F_ACC_EXCL, H5P_DEFAULT, mpiprop);
      }
  } else {
   if(!strcmp("WRITE", mode)) {
      f->file_id = -1;
      printf_ndc("Trying to write to a file that already exists");
    } else {
      printf_ndc("Reopening file");
      f->file_id = H5Fopen(fn, H5F_ACC_RDWR, mpiprop);
    }
  }

  ret = H5Pclose(mpiprop);
  assert(ret >= 0);

  if(f->file_id >= 0) {


// this opening is going to interact in slightly strange ways with the other
// opening of this container that happens at write time! eep.
    f->seq_handle_container_dataset = open_dataset_if_exists(f->file_id, DAMSEL_DATASET_CONTAINER_SEQ);
    f->vector_handle_container_dataset = open_dataset_if_exists(f->file_id, DAMSEL_DATASET_CONTAINER_VECTOR);
    f->tag_container_dataset = open_dataset_if_exists(f->file_id, DAMSEL_DATASET_TAG);
    f->collection_dataset = open_dataset_if_exists(f->file_id, DAMSEL_DATASET_COLLECTION);
    entities_initialize_for_read(f);

    initialize_handle_allocator(f);
    pop_ndc();
    return f;
  }
fn_error:
    printf_ndc("file open did not work - returning NULL");
    free(f);
    pop_ndc();
    return NULL;
}

static void free_remote_mapping(void *m) {
  damsel_remote_mapping *rm = (damsel_remote_mapping *)m;
  DMSLcontainer_release(rm->app_container);
  DMSLcontainer_release(rm->file_container);
  free(m);
}

static void free_remote_mappings(damsel_attached_file *file) {
  linkedlist_free(file->remote_mappings, free_remote_mapping);
}

void damsel_close_hdf5_store(damsel_attached_file *file) {
  int ret;
  ssize_t norphans;

  assert(file->store.type == DAMSEL_STORETYPE_HDF5);
  if(file->seq_handle_container_dataset >= 0) {
    ret = H5Dclose(file->seq_handle_container_dataset);
    assert(ret >= 0);
  }
  if(file->tag_container_dataset >= 0) {
    ret = H5Dclose(file->tag_container_dataset);
    assert(ret >= 0);
  }


  if(file->vector_handle_container_dataset >= 0) {
    ret = H5Dclose(file->vector_handle_container_dataset);
    assert(ret >= 0);
  }

  if(file->collection_dataset >= 0) {
    ret = H5Dclose(file->collection_dataset);
    assert(ret >= 0);
  }

  if(file->entities_dataset >= 0) {
    ret = H5Dclose(file->entities_dataset);
    assert(ret >= 0);
  }
 
  norphans = H5Fget_obj_count(file->file_id, H5F_OBJ_ALL); 
  if (norphans > 1) { /* expect 1 for the file we have not closed */
      int i;
      H5O_info_t info;
      char name[64];
      hid_t * objects = calloc(norphans, sizeof(hid_t));
      H5Fget_obj_ids(file->file_id, H5F_OBJ_ALL, -1, objects);
      for (i=0; i<norphans; i++) {
	  H5Oget_info(objects[i], &info);
	  H5Iget_name(objects[i], name, 64);
	  printf_ndc("%d of %zd things still open: %d with name %s of type %d", i, norphans, objects[i], name, info.type);
      }
  }

  ret = H5Fclose(file->file_id);
  assert(ret >= 0);
  MPI_Comm_free(&(file->communicator));

  if(file->model_populated != NULL) {
    if(file->model_populated->seq_info != NULL) free(file->model_populated->seq_info);
    if(file->model_populated->vec_info != NULL) free(file->model_populated->vec_info);
    if(file->model_populated->tag_info != NULL) free(file->model_populated->tag_info);
    free(file->model_populated);
  }

  free_remote_mappings(file);

  int status;
  status = H5Tclose(get_collection_type(file));
  assert(status >= 0);
  status = H5Tclose(get_tag_container_type(file));
  assert(status >= 0);
  status = H5Tclose(get_sequence_handle_container_type(file));
  assert(status >= 0);
  status = H5Tclose(get_vector_handle_container_type(file));
  assert(status >= 0);

  DMSLtrait_release(file->traits);

  model_release((damsel_store_internal *)file);

  printf_ndc("closed HDF5 file.");
}

/* inner code to do the writes for a file store */
static void push_predefined_containers(damsel_model_internal *src_s, damsel_attached_file *fs) {
  push_ndc("push_predefined_containers: for mappings that are already defined, generating push data");

  linkedlist *mappings = src_s->mappings;
  while(mappings != NULL) {
    printf_ndc("processing a mapping");
    damsel_store_mapping *mapping = (damsel_store_mapping *) mappings->head;
    printf_ndc("mapping struct %p", mapping);

    internal_record_pushed_container(fs, mapping->app_container_id, mapping->file_container_id, get_new_handles(fs, 1), 0);

    mappings = mappings->next;
  }

  pop_ndc();
}

static void free_metadata_list(damsel_metadata_handle *mh) {
  if(mh == NULL) return;
  printf_ndc("freeing metadata struct %p", mh);
  free_metadata_list(mh->next);
  free(mh);
}

/* this will resolve all handles stored in model->first_metadata_handle to real
   mappings, and then release all the memory associated with that metadata list.
   it will apply permutation callbacks.
*/
static void resolve_handles_for_invented_mappings(damsel_model_internal *src_s, damsel_attached_file *fs) {
  damsel_err_t err;
  push_ndc("resolve_handles_for_invented_mappings: resolving handles for invented-handles mappings");
  damsel_metadata_handle *mh = src_s->first_metadata_handle;
  while(mh != NULL) {
    // map the supplied handle to *something*
    // I'll use my lame handle allocator for now.
    // Later on, maybe this file code will get to choose *all* of the
    // file handles and so this will be less of a weird case.
    // TODO: this doesn't work on multirank code - ugly handle allocator
    // issues duplicates on each rank.

    // now i need to build a handle container in the file store that has
    // that handle.

    // it should be the same size as the source handle

    damsel_container_internal *shc =
      (damsel_container_internal *)mh->container_id;
    assert_valid_handle_container_id(shc);

    damsel_handle allocator_base = get_new_handles(fs, shc->quantity);

    damsel_container dest_hc;
    err = DMSLcontainer_create_sequence((damsel_model)fs, allocator_base, shc->quantity, 1, &dest_hc);
    printf_ndc("added metadata handle container id %p in dest store %p with handle container id %p", mh->container_id, fs, dest_hc);

    internal_record_pushed_container(fs, mh->container_id, dest_hc, get_new_handles(fs,1), 0);


// TODO: this is the point that the permutation callback should happen.
// instead of passing in mh->container_id, we need to pass in a
// container that is a permutation of mh->container_id.
// or... we need to pass in mh->container_id and a replacement for
// dest_hc that is a permutation of dest_hc.
// if the map is made inside the callback, then I can let the choice
// of that (or some other arrangement even) happen inside the permutation.

    // DMSLmodel_map_handles(mh->container_id , dest_hc);
    (*mh->permutation_callback)(mh->container_id, dest_hc);

    DMSLcontainer_release(mh->container_id);
    DMSLcontainer_release(dest_hc);
    mh = mh->next;
  }
  push_ndc("freeing metadata list");
  free_metadata_list(src_s->first_metadata_handle);
  pop_ndc();
  src_s->first_metadata_handle = NULL;
  pop_ndc(); // invented handles mappings
}

void damsel_permute_id(damsel_container appside, damsel_container fileside) {
  damsel_err_t err = DMSLmodel_map_handles(appside, fileside);
  assert(err.id == DMSL_OK.id);
}

static void broadcast_send_container(damsel_model_internal *src_s, damsel_container c, int processingrank) {

  push_ndc("broadcast_send_container");
  assert_valid_handle_container_id(c);

  damsel_attached_file *fs = src_s->attached_file;

  damsel_container_internal *a = (damsel_container_internal *) c;
  printf_ndc("broadcasting container of %d handles", a->quantity);
  MPI_Bcast(&(a->type), 1, MPI_INT, processingrank, fs->communicator);

  if(a->type == DAMSEL_HANDLE_CONTAINER_TYPE_VECTOR) {
    printf_ndc("broadcasting vector container");
    int ix;
    for(ix=0; ix<a->quantity; ix++) {
      printf_ndc(" sending handle %d with value %llx", ix, ((damsel_vector_handle_container *)c)->contents[ix]);
    }
    printf_ndc("expecting to send %zd bytes", sizeof(damsel_vector_handle_container));
    MPI_Bcast((void *)a, sizeof(damsel_vector_handle_container), MPI_BYTE, processingrank, fs->communicator);
    MPI_Bcast( ((damsel_vector_handle_container *)c)->contents, sizeof(damsel_handle) * a->quantity, MPI_BYTE, processingrank, fs->communicator);
#ifdef USE_BSEARCH	
    MPI_Bcast( ((damsel_vector_handle_container *)c)->contents_sorted, sizeof(damsel_handle_index) * a->quantity, MPI_BYTE, processingrank, fs->communicator);
#endif	
  } else if(a->type == DAMSEL_HANDLE_CONTAINER_TYPE_SEQUENCE) {
    MPI_Bcast(a, sizeof(damsel_seq_handle_container), MPI_BYTE, processingrank, fs->communicator);
  } else if(a->type == DAMSEL_HANDLE_CONTAINER_TYPE_TREE) {
    // this is more complicated because it consists of a bunch of other
    // containers. those containers might have already been broadcast but
    // as at this point those containers cannot be modified, it is ok
    // to send them again.
    damsel_tree_handle_container *ta = (damsel_tree_handle_container *) a;
    linkedlist *treehere = ta->first_hc;
    int treelength = linkedlist_length(treehere);
    MPI_Bcast(&treelength, 1, MPI_INT, processingrank, fs->communicator);
    int verification_count = 0;
    while(treehere!=NULL) {
      damsel_container thiscontainer = (damsel_container)treehere->head;
      broadcast_send_container(src_s, thiscontainer, processingrank);
      verification_count++;
      treehere=treehere->next;
    }
    assert(verification_count == treelength);
  }
  else {
    assert(1==0 && "only know how to send sequence, vector and tree containers");
  }
  pop_ndc();
}

static damsel_container_internal *broadcast_receive_container(damsel_model_internal *src_s, int processingrank, damsel_model target_model) {
  push_ndc("broadcast_receive_container");
  damsel_attached_file *fs = src_s->attached_file;

  int containertype;
  MPI_Bcast(&containertype, 1, MPI_INT, processingrank, fs->communicator);

  damsel_container_internal *a;

  if(containertype == DAMSEL_HANDLE_CONTAINER_TYPE_VECTOR) {
    damsel_vector_handle_container *av;
    av=malloc(sizeof(damsel_vector_handle_container));
    char *ccc = (char *)av;
    int ix;
    for(ix=0;ix<sizeof(damsel_vector_handle_container); ix++) ccc[ix] = 0xFE;

    printf_ndc("expecting to receive %zd bytes", sizeof(damsel_vector_handle_container));
    MPI_Bcast(av, sizeof(damsel_vector_handle_container), MPI_BYTE, processingrank, fs->communicator);
    av->container.model = (damsel_model) src_s;
    model_take((damsel_store_internal *)av->container.model);
    a=(damsel_container_internal *) av;
    av->contents = malloc(sizeof(damsel_handle) * a->quantity);
    assert(av->contents != NULL);
    MPI_Bcast( av->contents, sizeof(damsel_handle) * a->quantity, MPI_BYTE, processingrank, fs->communicator);

#ifdef USE_BSEARCH	
	av->contents_sorted = malloc(sizeof(damsel_handle_index) * a->quantity);
    assert(av->contents_sorted != NULL);
    MPI_Bcast( av->contents_sorted, sizeof(damsel_handle_index) * a->quantity, MPI_BYTE, processingrank, fs->communicator);
#endif

    push_printf_ndc("received vector container with %d entries.", a->quantity);
    assert(av->container.quantity == a->quantity);
    for(ix=0; ix < a->quantity; ix++) {
      printf_ndc(" handle %d is %llx", ix, av->contents[ix]);
    }
    a->refcount = 1;
    pop_ndc();
  } else if(containertype == DAMSEL_HANDLE_CONTAINER_TYPE_SEQUENCE) {
    damsel_vector_handle_container *as;
    as=malloc(sizeof(damsel_seq_handle_container));
    MPI_Bcast(as, sizeof(damsel_seq_handle_container), MPI_BYTE, processingrank, fs->communicator);
    as->container.model = target_model;
    model_take((damsel_store_internal *)as->container.model);
    a=(damsel_container_internal *) as;
    a->refcount = 1;
  } else if(containertype == DAMSEL_HANDLE_CONTAINER_TYPE_TREE) {
    a = (damsel_container_internal *) DMSLcontainer_tree_create((damsel_model) src_s, DAMSEL_HANDLE_COLLECTION_TYPE_VECTOR);
    int quantity;
    MPI_Bcast(&quantity, 1, MPI_INT, processingrank, fs->communicator);
    int treeix;
    for(treeix=0; treeix<quantity; treeix++) {
      damsel_container newc;
      newc = (damsel_container) broadcast_receive_container(src_s, processingrank, target_model);
      DMSLcontainer_tree_add((damsel_container) a, newc);
    }
  } else {
    assert(1==0 && "unknown container type broadcast");
    return NULL;
  }
  assert_valid_handle_container_id(a);
  pop_ndc();
  return a;
}

/* distributes mappings from each rank onto all the other ranks.
   hm... is this going to need containers to be shipped over the wire too?
   because mappings exist as declarations between containers... yes.

   this is a pretty sequential implementation. maybe there can be
   a many-to-many parallelised version later on.
 */
static void resolve_remote_mappings(damsel_model_internal *src_s, damsel_attached_file *fs) {
  push_ndc("resolve_remote_mappings");

  int processingrank, maxrank, myrank;
  maxrank = fs->maxrank;
  myrank = fs->myrank;

  linkedlist *origmappings; // mappings before we mess with them
  origmappings = src_s->mappings;

  for(processingrank = 0; processingrank < maxrank; processingrank++) {
    printf_ndc("rank %d processing data for rank %d", myrank, processingrank);

    int mappingcount = linkedlist_length(origmappings);
    MPI_Bcast(&mappingcount, 1, MPI_INT, processingrank, fs->communicator);

    if(processingrank == myrank) {
      push_printf_ndc("rank %d being the master", myrank);


      linkedlist *here = origmappings;
      while(here != NULL) {
        printf_ndc("processing a mapping");

        damsel_store_mapping *mapping = (damsel_store_mapping *) here->head;

        broadcast_send_container(src_s, mapping->app_container_id, processingrank);
        broadcast_send_container(src_s, mapping->file_container_id, processingrank);
        damsel_handle fh = internal_push_container_lookup(mapping->app_container_id);
        MPI_Bcast(&fh, 1, MPI_LONG_LONG, processingrank, fs->communicator);
        here=here->next;
      }

    // and the mapping between the two is implicit because they
    // are sent together
      pop_ndc();
    } else {
      int mappingnum;
      for(mappingnum=0; mappingnum < mappingcount; mappingnum++ ) {
        damsel_container ac, fc;
        ac = (damsel_container) broadcast_receive_container(src_s, processingrank, (damsel_model) src_s);
        fc = (damsel_container) broadcast_receive_container(src_s, processingrank, (damsel_model) fs);
// TODO: these are both remote containers (or their mapping into the file is remote)
        damsel_handle fh;
        MPI_Bcast(&fh, 1, MPI_LONG_LONG, processingrank, fs->communicator);

// now store this all somewhere - as a mapping and maybe as a pushed container?
//        internal_push_container_remote(ac, fc, fh);
   /*     internal_record_pushed_container(fs, ac,
          fc, fh, 1);
        model_map_handles_internal(ac, fc, 1); */
        internal_record_remote_mapping(fs, ac, fc, fh, processingrank);
      }
    }
  }
  pop_ndc();
}

/* broadcast the global mappings to everyone */
static void resolve_global_mappings(damsel_model_internal *src_s, damsel_attached_file *fs) {

  push_ndc("resolve_global_mappings");

  assert(fs!=NULL);

  int processingrank, maxrank, myrank;
  maxrank = fs->maxrank;
  myrank = fs->myrank;

  int local_count_of_global_mappings = linkedlist_length(src_s->global_mappings);
  printf_ndc("rank %d has %d global mappings", myrank, local_count_of_global_mappings);

  int *per_rank_count_of_global_mappings = malloc(sizeof(int) * maxrank);
  assert(per_rank_count_of_global_mappings != NULL);

  MPI_Allgather(&local_count_of_global_mappings, 1, MPI_INT,
                per_rank_count_of_global_mappings, 1, MPI_INT,
                fs->communicator);
  for(processingrank = 0 ; processingrank < maxrank ; processingrank++) {


   if(processingrank == myrank) { // master

     linkedlist *ll = src_s->global_mappings;
  // assume that the first entries (up to the local_count_of_global_mappings-th entry)
  // are for this rank, and that the ones afterwards have been appended by this
  // routine and are not from this rank.

      printf_ndc("rank %d sending %d global mappings", myrank, local_count_of_global_mappings);
      // we already know the rank - its processingrank
      int ix;
      for(ix=0;ix<local_count_of_global_mappings;ix++) {
        assert(ll != NULL);
        damsel_internal_global_mapping *gm = (damsel_internal_global_mapping *) ll->head;
        assert(gm->magic == DAMSEL_GLOBAL_MAPPING_MAGIC);
        assert_valid_handle_container_id(gm->app_container);
        broadcast_send_container(src_s, gm->app_container, processingrank);
        int containersize = DMSLcontainer_count(gm->app_container);
        MPI_Bcast(gm->proc_ranks, containersize, MPI_INT, processingrank, fs->communicator);
        printf_ndc("global mapping to send: container = %p, proc_ranks = %p, remote_handles = %p", gm->app_container, gm->proc_ranks, gm->remote_handles);
        push_ndc("dump of remote handles to be sent");
        int ixx;
        for(ixx = 0 ; ixx < containersize; ixx++) {
          printf_ndc("position %d has remote app handle %llx", ixx, gm->remote_handles[ixx]);
        }
        pop_ndc();
        MPI_Bcast(gm->remote_handles, containersize, MPI_LONG_LONG, processingrank, fs->communicator);

        ll = ll->next;
      }

      // once we've broadcast the container, then we also know the lengths
      // of the other structures
    } else { // slave
      printf_ndc("rank %d receiving %d global mappings", myrank, per_rank_count_of_global_mappings[processingrank]);
      int ix;
      for(ix=0;ix<per_rank_count_of_global_mappings[processingrank];ix++) {
        push_ndc("receiving a mapping");
        damsel_internal_global_mapping *gm = malloc(sizeof(damsel_internal_global_mapping));
        gm->manage_memory_internally = 1;
        gm->magic = DAMSEL_GLOBAL_MAPPING_MAGIC;
        gm->rank = processingrank;
        printf_ndc("remote rank = %d", gm->rank);
        gm->app_container = (damsel_container) broadcast_receive_container(src_s, processingrank, (damsel_model) fs);
        assert_valid_handle_container_id(gm->app_container);
        int containersize = DMSLcontainer_count(gm->app_container);
        printf_ndc("container size = %d", containersize);
        gm->proc_ranks = malloc(sizeof(int) * containersize);
        assert(gm->proc_ranks != NULL);
        MPI_Bcast(gm->proc_ranks, containersize, MPI_INT, processingrank, fs->communicator);
        gm->remote_handles = malloc(sizeof(damsel_handle) * containersize);
        assert(gm->remote_handles != NULL);
        MPI_Bcast(gm->remote_handles, containersize, MPI_LONG_LONG, processingrank, fs->communicator);
        push_ndc("dump of remote handles just received");
        int ixx;
        for(ixx = 0 ; ixx < containersize; ixx++) {
          printf_ndc("position %d has remote app handle %llx", ixx, gm->remote_handles[ixx]);
        }
        pop_ndc();
        // THIS IS A LINKED-LIST APPEND - not linked list prepend,
        // because it will otherwise corrupt the ordering required for
        // the send part of this routine.
        linkedlist_append(&src_s->global_mappings, gm);
        pop_ndc();
      }
    }
  }

  free(per_rank_count_of_global_mappings);
  pop_ndc();
}

static void push_collection_containers(damsel_model_internal *src_s) {
  void *ctx = push_ndc("pushing containers used in collections");
  assert(src_s->direction == DAMSEL_TRANSFER_TYPE_WRITE);
  linkedlist *collection_container_list = src_s->first_collection;
  while(collection_container_list!=NULL) {
    damsel_collection_internal *ci = (damsel_collection_internal *) collection_container_list->head;
    assert(ci!=NULL);
    damsel_err_t ret = internal_push_container(ci->container, NULL);
// TODO: but only need to push this - this is failing in test-create-collection. not precisely sure why... some mapping is not happening, but what is it that is meant to be mapped here? I used to allow this to fail because i just wanted it on the list. hmm.
    assert_dmsl_ok(ret);
    collection_container_list = collection_container_list->next;
  }
  pop_ndc_context(ctx);
}

/*! TODO: this chunks code is not creating the right chunks at the moment. its creating too many. I need to
   figure out the algorithm that I'm using now, figure out what I want it do to.

foreach tag, tagdef:
  foreach cached_data, dcd:
    map app-side tag handle in dcd to fileside (-> fileside_dcd_tag_handle)
    does this mapping exist? if so:
      for each app-side handles in dcd, local_app_handle:
        map that local app hadnle to the fileside (-> fileside_handle)
        does that map? if so:
          for each local mapping 'mapping':
            find fileside-handle in the fileside of mapping
            is it found? if so, writepos is the position.
          now do something similar for remote mappings.
          if we found it somewhere
              assert (with de-deuplication) that we should create a chunk with these parameters: tagdef, file_container

BUG? I'm never checking that the tag referenced by dcd refers to the tag.

*/

static void create_chunks(damsel_model_internal *s, damsel_attached_file *fs) {
  void *ctx = push_ndc("create_chunks");

  // need to broadcast out details of datasets I want created.
  // that is, tags for which I have handle containers.

  linkedlist *ll;
  for(ll = s->first_tag_definition; ll != NULL; ll=ll->next) {
    damsel_tag_internal *tagdef = (damsel_tag_internal *)ll->head;
    void *ctx2 = push_printf_ndc("processing tag definition with pointer %p, name=%s, handle=0x%llx", tagdef, tagdef->name, tagdef->handle);

    damsel_cached_data_internal *dcd;
    for(dcd = s->first_cached_data; dcd != NULL; dcd=dcd->next) {
      void *ctx3 = push_printf_ndc("processing cached data write with pointer %p", dcd);
      // does this refer to the correct tag?
      if(tagdef->handle == dcd->tag_handle) {
      printf_ndc("handles_id = %p", dcd->handles_id);
      printf_ndc("tag_handle = %llx", dcd->tag_handle);
      printf_ndc("data = %p", dcd->data);
      damsel_handle fileside_dcd_tag_handle;
      damsel_err_t ret = handle_map(s, DAMSEL_TRANSFER_TYPE_WRITE, dcd->tag_handle, &fileside_dcd_tag_handle);
      if(ret.id == 0) { // a mapping exists
        assert_dmsl_ok(ret);
        int pos;
        int maxpos = DMSLcontainer_count(dcd->handles_id);
        for(pos=0; pos<maxpos; pos++) {
          push_printf_ndc("checking position %d out of %d", pos, maxpos);

          damsel_handle local_app_handle =
            DMSLcontainer_handle_at_position(dcd->handles_id, pos);

          damsel_handle fileside_handle;

          damsel_err_t ret = handle_map(s, DAMSEL_TRANSFER_TYPE_WRITE,
            local_app_handle, &fileside_handle);
          printf_ndc("local_app_handle 0x%llx maps to fileside handle 0x%llx",
            local_app_handle, fileside_handle);
          if(ret.id == 0) { // then we mapped ok
            hsize_t writepos = -1;
            damsel_handle file_container_handle = 0 ;
            damsel_container_internal *file_container = NULL;
            void *ctxlm = push_ndc("checking local mappings");

            linkedlist *lm;
            for(lm = s->mappings; lm != NULL && writepos == -1; lm = lm->next) {
              damsel_store_mapping *mapping = (damsel_store_mapping *)lm->head;
              push_printf_ndc("local mapping iteration, mapping = %p", mapping);
// BUG In here i think - it appears to be finding a valid writepos for mappings that are irrelevant
// perhaps i need to put in an extra check for something?
// why is this firing too many times?what am I actually firing on here?
              // does handle container contain fileside_handle, and if so, at what pos?
              writepos = (hsize_t) find_handle_in_container((damsel_container_internal *)mapping->file_container_id, fileside_handle);
              printf_ndc("find_handle_in_container returned position %d", (int) fileside_handle);
              file_container_handle = internal_push_container_lookup(mapping->app_container_id);
              file_container = (damsel_container_internal *) mapping->file_container_id;
              if(writepos != -1) printf_ndc("Found fileside_handle 0x%llx at position %d", fileside_handle, (int)writepos);
              pop_ndc();
            }

            pop_ndc_context(ctxlm);

            void *ctxrm = push_ndc("checking remote mappings");
            linkedlist *rm; /* of damsel_remote_mapping */
            for(rm = fs->remote_mappings; rm != NULL && writepos == -1; rm = rm->next) {
              push_printf_ndc("checking remote mapping %p", rm);
              damsel_remote_mapping *mapping = (damsel_remote_mapping *)rm->head;
              writepos = (hsize_t) find_handle_in_container((damsel_container_internal *)mapping->file_container, fileside_handle);
              file_container_handle = mapping->file_handle;
              file_container = (damsel_container_internal *) mapping->file_container;
              printf_ndc("remote mapping iteration, writepos = %d", (int)writepos);
              pop_ndc();
            }
            pop_ndc_context(ctxrm);

            if(writepos != -1) {
              // check (in a not very scalable fashion) if this chunk is already there
              linkedlist *llc;
              int found=0;
              for(llc=s->first_chunk; llc != NULL && found == 0; llc=llc->next) {
                damsel_internal_chunk *scanchunk = (damsel_internal_chunk *)llc->head;
                if(scanchunk->tag == tagdef
                  && scanchunk->container == file_container) found = 1;
              }
              if(found == 0) {
                printf_ndc("creating new chunk");
                assert(file_container_handle != 0); 
                assert(file_container != NULL);
                damsel_internal_chunk *chunk = malloc(sizeof(damsel_internal_chunk));
                assert(chunk!=NULL);
                chunk->magic = DAMSEL_CHUNK_MAGIC;
                chunk->tag = tagdef;
                chunk->container = file_container;
                DMSLcontainer_take((damsel_container)file_container);
                linkedlist_prepend(&(s->first_chunk), chunk);
              }
            }
          }
          pop_ndc();
        }
      } else {
#ifndef NDEBUG
        printf_ndc("Tag %p name=%s app handle=0x%llx has no mapping when trying to write dcd %p", tagdef, tagdef->name, tagdef->handle, dcd);
#endif
        assert(1==0 && "Tag has no mapping when trying to write data");
      }
      }
      pop_ndc_context(ctx3);
    }
    pop_ndc_context(ctx2);
  }
  pop_ndc_context(ctx);
}

/* get the structure of user defined datatype */
static void *get_defined_type_struct(damsel_data_type damsel_type) 
{
      long long i;
      linkedlist *here = defined_types;
	  
	  if (damsel_type <= DAMSEL_DATA_TYPE_PREDEFINED_WATERMARK)
		return NULL;
	  
	  // if its not a built-in type then we need to look up the type
      // definition structure, and then switch on that
      printf_ndc("need to look up damsel_data_type 0x%llx", damsel_type);
      printf_ndc("linked list has length %d", linkedlist_length(here));
      printf_ndc("DAMSEL_DATA_TYPE_PREDEFINED_WATERMARK = 0x%x", DAMSEL_DATA_TYPE_PREDEFINED_WATERMARK);
      printf_ndc("damsel_type - DAMSEL_DATA_TYPE_PREDEFINED_WATERMARK = 0x%llx", damsel_type - DAMSEL_DATA_TYPE_PREDEFINED_WATERMARK);
      for(i=1; i<damsel_type - DAMSEL_DATA_TYPE_PREDEFINED_WATERMARK; i++) {
        printf_ndc("i = %llx", i);
        assert(here!=NULL);
        here = here->next;
      }

      assert(here);
      assert(here->head);

      return here->head;
}
	  
typedef struct dataset_info_s {
    hsize_t dimsf;
    long internal_type;
    char name[DAMSEL_DSET_NAME_MAX]; // what happens when this is overflowed ?
} dataset_info;

static MPI_Datatype make_dataset_type(dataset_info *target) {
    MPI_Datatype newtype;
    int count=0;
    int lens[3] = {sizeof(target->dimsf), 
	sizeof(target->internal_type), sizeof(target->name) };
    MPI_Aint offsets[3];
    MPI_Datatype types[3] = {MPI_UNSIGNED_LONG_LONG, MPI_LONG, MPI_CHAR};

    MPI_Address(&(target->dimsf), &offsets[count++]);
    MPI_Address(&(target->internal_type), &offsets[count++]);
    MPI_Address(&(target->name), &offsets[count++]);

    MPI_Type_create_struct(count, lens, offsets, types,  &newtype);
    MPI_Type_commit(&newtype);

    return newtype;
}

static void generate_chunk_name(char *target, char *name, damsel_handle handle) {
  snprintf(target, DAMSEL_DSET_NAME_MAX-1, "%s-%llx", name, handle);
}

/** called by create_datasets to create a storage chunk
*/
static void create_dataset(damsel_attached_file *fs, int processingrank, damsel_tag_internal *tagdef, damsel_container_internal *hc) {

  int maxrank, myrank;
  dataset_info info;
  MPI_Datatype dataset_info_type;
 
  maxrank = fs->maxrank;
  myrank = fs->myrank; 

  damsel_container file_container = (damsel_container)hc;

  assert(hc != NULL);
  info.dimsf = hc->quantity;

  assert_is_damsel_type(tagdef->type);
  hid_t datatype = damsel_type_to_hdf5_type(tagdef->type);
  
  printf_ndc("returned datatype is %d", (int)datatype);
  damsel_handle desthandle = internal_push_container_lookup_fileside(file_container);

  generate_chunk_name(info.name, tagdef->name, desthandle);
/* these two creates are what I need to do collectively, so the parameters
   of these are what I need to broadcast. */
  int mode = 1; // indicate we're sending a create req
  assert(processingrank == myrank);

//
// need to convert size into something that can go over the wire
  info.internal_type = damsel_type_to_serializable_type(tagdef->type);

  dataset_info_type =  make_dataset_type(&info);

  MPI_Bcast(&mode, 1, MPI_INT, processingrank, fs->communicator);

  MPI_Bcast(MPI_BOTTOM, 1, dataset_info_type, myrank, fs->communicator);

  printf_ndc("rank %d creating dataspace with dims_as_int= %d", myrank, (int)info.dimsf);
  fflush(NULL);
  printf_ndc("screate barrier on master rank=%d", myrank);
  hid_t dataspace = H5Screate_simple(1, &info.dimsf, &info.dimsf);
  assert(dataspace >= 0);

  printf_ndc("B: returned dataspace is %d", (int)dataspace);
  push_ndc("create HDF5 dataset");
  printf_ndc("rank %d", myrank);
  printf_ndc("name %s", info.name);
  printf_ndc("size %d", (int)info.dimsf);
  printf_ndc("datatype %d", (int)datatype);

  /* dataset could already exist (?).  So don't re-create it if so. */
  hid_t dataset = H5I_BADID;
  hid_t o_exists;
  hid_t d_exists = H5Lexists(fs->file_id, info.name, H5P_DEFAULT);
  if (d_exists)  
      o_exists = H5Oexists_by_name(fs->file_id, info.name, H5P_DEFAULT);
  if (d_exists && o_exists)  {
      printf_ndc("skipping dataset creation: already present");
  } else {
      dataset = H5Dcreate(fs->file_id, info.name, datatype,
	      dataspace, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
  }
  pop_ndc();
  H5Fflush(fs->file_id, H5F_SCOPE_GLOBAL);
  H5Sclose(dataspace);
  if (dataset != H5I_BADID) H5Dclose(dataset);
  H5Fflush(fs->file_id, H5F_SCOPE_GLOBAL);

  MPI_Type_free(&dataset_info_type);

}

/* This goes over the list of chunks on the local rank and creates datasets for them. */
static void create_datasets_master(damsel_model_internal *src_s, int processingrank, damsel_attached_file *fs) {


#ifndef NDEBUG
  int thisrank;
  MPI_Comm_rank(fs->communicator, &thisrank);
  assert(processingrank == thisrank);
#endif

  printf_ndc("I am rank %d, in collective stage of hdf5 execute.", processingrank);
  fflush(NULL);
  // need to broadcast out details of datasets I want created.
 // that is, tags for which I have handle containers.

  linkedlist *llc;
  for(llc = src_s->first_chunk; llc != NULL; llc=llc->next) {
    damsel_internal_chunk *chunk = llc->head;
    create_dataset(fs, processingrank, chunk->tag, chunk->container);
  }

  int mode = 0; // indicate we're done
  MPI_Bcast(&mode, 1, MPI_INT, processingrank, fs->communicator);
}

static void create_datasets_slave(damsel_model_internal *src_s, int processingrank, damsel_attached_file *fs) {

  int myrank;
  MPI_Comm_rank(fs->communicator, &myrank);
  assert(processingrank != myrank);

  // need to receive broadcast details of datasets someone else
  // wants created.
  printf_ndc("I am rank %d, being a collective-slave for rank %d.", myrank, processingrank);
  // are we going to skip or to receive?
  int mode;
  dataset_info info;
  MPI_Datatype dataset_info_type = make_dataset_type(&info);
  while(1==1) {
    MPI_Bcast(&mode, 1, MPI_INT, processingrank, fs->communicator);
    if(mode == 1) {
      MPI_Bcast(MPI_BOTTOM, 1, dataset_info_type, processingrank, fs->communicator);
      damsel_data_type type = damsel_serializable_type_to_damsel_type(info.internal_type);
      printf_ndc("rank %d creating dataspace with dimsf= %d", myrank, (int)info.dimsf);
      fflush(NULL);
      hid_t dataspace = H5Screate_simple(1, &(info.dimsf), &(info.dimsf));
      printf_ndc("dataspace handle on rank %d is %x", myrank, (int)dataspace);
      assert(dataspace >= 0);

      assert_is_damsel_type(type);
      hid_t datatype= damsel_type_to_hdf5_type(type);

      printf_ndc("rank %d creating dataset with name '%s' and size %d and datatype %d.", myrank, info.name, (int)info.dimsf, (int)datatype);
      hid_t dataset = H5Dcreate(fs->file_id, info.name, datatype,
        dataspace, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
      // assert(dataset >= 0); // might fail because already exists

      printf_ndc("rank %d created dataset with name '%s' and size %d.", myrank, info.name, (int)info.dimsf);
      fflush(NULL);
      H5Fflush(fs->file_id, H5F_SCOPE_GLOBAL);

      H5Sclose(dataspace);
      H5Dclose(dataset);
      H5Fflush(fs->file_id, H5F_SCOPE_GLOBAL);
    } else if(mode == 0) {
      break;
    }
  }
  MPI_Type_free(&dataset_info_type);
}

/** creates the HDF5 datasets that will store chunks of damsel data. */
static void create_datasets(damsel_model_internal *src_s, damsel_attached_file *fs) {

  void *ctx = push_ndc("dataset creation");
  assert(src_s != NULL);

  assert(src_s->direction == DAMSEL_TRANSFER_TYPE_WRITE);

  assert(fs != NULL);

  int processingrank, maxrank, myrank;
  maxrank = fs->maxrank;
  myrank = fs->myrank;

  /* we find that processors are creating the same datasets.  This odd behavior
   * is not a bug, aside from a symptop of a need to rethink the whole dataset
   * creation process.  So, disable automatic error reporting */
  //herr_t (*old_func)(hid_t, void*);
  //void *old_data;
  //H5Eget_auto2(H5E_DEFAULT, &old_func, &old_data);
  //H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  for(processingrank=0; processingrank<maxrank; processingrank++) {
    push_printf_ndc("I am rank %d in collective iteration for processing rank %d", myrank, processingrank);
    fflush(NULL);
    printf_ndc("I am rank %d done with barrier", myrank);
    fflush(NULL);
    // at this stage, processingrank should be the same on all ranks
    if(processingrank == myrank) {
      create_datasets_master(src_s, processingrank, fs);
    } else {
      create_datasets_slave(src_s, processingrank, fs);
    }
    fflush(NULL);
    printf_ndc("ok, everyone barriering after creating datasets, rank=%d.",myrank);
    fflush(NULL);
    pop_ndc();
  }

  //H5Eset_auto2(H5E_DEFAULT, old_func, old_data);
  pop_ndc_context(ctx);

}

static void store_containers_seq(damsel_model_internal *model, damsel_attached_file *fs) {
  assert(model != NULL);
  assert(model->direction == DAMSEL_TRANSFER_TYPE_WRITE);

  linkedlist *ll;

  int maxrank, myrank;
  maxrank = fs->maxrank;
  myrank = fs->myrank;

  int local_seq_handle_container_count=0;
  push_ndc("creating sequence handle container dataset");
  ll = fs->first_pushed_container;
  while(ll!=NULL) {
    damsel_hdf5_internal_pushed_container *pc = (damsel_hdf5_internal_pushed_container *)ll->head;
    damsel_container_internal *handles = (damsel_container_internal *)pc->file_container;

    if(handles->type == DAMSEL_HANDLE_CONTAINER_TYPE_SEQUENCE && handles->written == DAMSEL_UNWRITTEN && pc->remote == 0) {
      local_seq_handle_container_count++;

#ifndef NDEBUG
/* this test is no use any more: file side handles don't directly encode the rank any more.
  int rank_in_mpi;
  MPI_Comm_rank(fs->communicator, &rank_in_mpi);
  int rank_in_file_handle = pc->handle >> 32;
  if(pc->remote==0) {
    assert(rank_in_file_handle == rank_in_mpi);
  } 
*/
#endif


    }
    ll = ll->next;
  }

  int *perrank_seq_handle_container_count;
  perrank_seq_handle_container_count = malloc(sizeof(int) * maxrank);
  assert(perrank_seq_handle_container_count != NULL);

  printf_ndc("rank %d local_seq_handle_container_count = %d", myrank,
    local_seq_handle_container_count);
  MPI_Allgather(&local_seq_handle_container_count, 1, MPI_INT,
                perrank_seq_handle_container_count, 1, MPI_INT,
                 fs->communicator);

  hsize_t hc_rows = 0;
  int i;
  for(i=0; i<maxrank; i++) {
    printf_ndc("rank %d: perrrank_handle_container_count[%d] = %d",
      myrank, i, perrank_seq_handle_container_count[i]);
    hc_rows += perrank_seq_handle_container_count[i];
  }
// this bit is collective ...

  hid_t container_type = get_sequence_handle_container_type(fs);
  assert(container_type >= 0);

  hid_t hc_dataspace_one = H5Screate_simple(1, &hsize_one, &hsize_one);
  assert(hc_dataspace_one >= 0);

  int my_hc_base = 0;

  hid_t hc_dataset = -1;
  hid_t hc_dataspace = -1;

  if(H5Lexists(fs->file_id, DAMSEL_DATASET_CONTAINER_SEQ, H5P_DEFAULT) <= 0) {
    hsize_t hc_unlimited = H5S_UNLIMITED, chunk_size=hc_rows/DEFAULT_NCHUNKS;
    hc_dataspace = H5Screate_simple(1, &hc_rows, &hc_unlimited);
    assert(hc_dataspace >= 0);
    hid_t chunked_plist = H5Pcreate(H5P_DATASET_CREATE);
	
	if (chunk_size<1) chunk_size = 1;
    H5Pset_chunk(chunked_plist, 1, &chunk_size);

    hc_dataset = H5Dcreate(fs->file_id, DAMSEL_DATASET_CONTAINER_SEQ,
      container_type, hc_dataspace, H5P_DEFAULT, chunked_plist, H5P_DEFAULT);
    assert(hc_dataset >= 0);

    H5Pclose(chunked_plist);

  } else {
    hc_dataset=H5Dopen(fs->file_id, DAMSEL_DATASET_CONTAINER_SEQ, H5P_DEFAULT);
    hc_dataspace = H5Dget_space(hc_dataset);
    hsize_t hc_offset = H5Sget_simple_extent_npoints(hc_dataspace);
    hsize_t hc_new_rows = hc_offset + hc_rows;
    printf_ndc("Extending sequence container dataset: hc_offset=%ld hc_rows=%ld hc_new_rows=%ld", (long)hc_offset, (long)hc_rows, (long)hc_new_rows);
    if(hc_new_rows > 0) H5Dset_extent(hc_dataset, &hc_new_rows);
    hc_dataspace = H5Dget_space(hc_dataset);
    my_hc_base += hc_offset;
  }

  assert(hc_dataset >= 0);
  assert(hc_dataspace >= 0);

  printf_ndc("proc %d barriering after handle container create.", myrank);

  pop_ndc();

  push_ndc("populating sequence container dataset");
// ... and this bit is independent ...

  int res;
  for(i=0; i<myrank; i++) {
    my_hc_base += perrank_seq_handle_container_count[i];
  }

  printf_ndc("rank %d will put handle container rows starting at offset %d",
    myrank, my_hc_base);

  hid_t hc_plist_id = H5Pcreate (H5P_DATASET_XFER);
  H5Pset_dxpl_mpio (hc_plist_id, H5FD_MPIO_INDEPENDENT);
  assert(hc_plist_id >= 0);

  ll = fs->first_pushed_container;
  i=0;
  while(ll != NULL) {
    assert(ll!=NULL);
    damsel_hdf5_internal_pushed_container *pc = (damsel_hdf5_internal_pushed_container *)ll->head;
    damsel_container_internal *handles = (damsel_container_internal *)pc->file_container;

    if( handles->type == DAMSEL_HANDLE_CONTAINER_TYPE_SEQUENCE && pc->remote == 0 && handles->written == DAMSEL_UNWRITTEN ) {
      assert(i<local_seq_handle_container_count);
      damsel_seq_handle_container *hsc = (damsel_seq_handle_container *)handles;
      damsel_hdf5_infile_seq_handle_container data;
      data.handle = internal_push_container_lookup_fileside((damsel_container) handles);
      data.base = hsc->base;
      data.quantity= handles->quantity;
      data.step = hsc->step;
      res = H5Sselect_none(hc_dataspace);
      assert(res >= 0);
      hsize_t coords = my_hc_base + i;
      res = H5Sselect_elements(hc_dataspace, H5S_SELECT_SET, 1, &coords);
      assert(res >= 0);
      printf_ndc("rank %d writing hc at %d", myrank, (int)coords);
      res = H5Dwrite(hc_dataset, container_type, hc_dataspace_one,
        hc_dataspace, hc_plist_id, &data);
      assert(res >= 0);
      i++;
    }
    ll = ll->next;
  }

  assert(i == local_seq_handle_container_count);
  assert(ll==NULL); // we should have hit the end of this linked list at
                    // the same time that we hit the end of the loop.

// and this bit is collective again.

  free(perrank_seq_handle_container_count);
  int status;
  status = H5Sclose(hc_dataspace);
  assert(status >= 0);
  status = H5Sclose(hc_dataspace_one);
  assert(status >= 0);
  status = H5Dclose(hc_dataset);
  assert(status >= 0);
  status = H5Pclose(hc_plist_id);
  assert(status >= 0);

  pop_ndc();

}

int unwritten_tag_filter(void *tv) {
  damsel_tag_internal *t = (damsel_tag_internal *)tv;
  assert_valid_tag_definition_structure(t);
  return (t->written == DAMSEL_UNWRITTEN);
}

static void store_tags(damsel_model_internal *model, damsel_attached_file *fs) {
  push_ndc("creating tag container");

  assert(model->direction == DAMSEL_TRANSFER_TYPE_WRITE);
  assert(model!=NULL);
  assert_valid_model_id(model);

  int myrank;
  MPI_Comm_rank(fs->communicator, &myrank);

  hid_t tag_container_type = get_tag_container_type(fs);
  assert(tag_container_type >= 0);

  linkedlist *tags_to_write;
  tags_to_write = linkedlist_filter(model->first_tag_definition, unwritten_tag_filter);

  hsize_t tc_rows = linkedlist_length(tags_to_write);

  printf_ndc("rank %d: we have %d tags", (int)myrank, (int) tc_rows);
// but some of these tags may already exist in the file and so we
// don't want to write them out again - and the count here then is too
// long sometimes. I need only new tag definitions that haven't been
// written to the file. use a marker byte for that.

  hid_t tc_dataset = -1;
  hid_t tc_dataspace = -1;
  hsize_t tc_offset = 0;

  if(H5Lexists(fs->file_id, DAMSEL_DATASET_TAG, H5P_DEFAULT) <= 0) {

    hsize_t tc_unlimited = H5S_UNLIMITED, chunk_size=tc_rows/DEFAULT_NCHUNKS;
    tc_dataspace = H5Screate_simple(1, &tc_rows, &tc_unlimited);
    assert(tc_dataspace >= 0);

    hid_t chunked_plist = H5Pcreate(H5P_DATASET_CREATE);
	if (chunk_size<1) chunk_size = 1;
    H5Pset_chunk(chunked_plist, 1, &chunk_size);
    tc_dataset = H5Dcreate(fs->file_id, DAMSEL_DATASET_TAG,
      tag_container_type, tc_dataspace, H5P_DEFAULT, chunked_plist, H5P_DEFAULT);
    assert(tc_dataset >= 0);
    H5Pclose(chunked_plist);
  } else {
    tc_dataset = H5Dopen(fs->file_id, DAMSEL_DATASET_TAG, H5P_DEFAULT);
    assert(tc_dataset >= 0);
    tc_dataspace = H5Dget_space(tc_dataset);
    tc_offset = H5Sget_simple_extent_npoints(tc_dataspace);

    hsize_t tc_new_rows = tc_offset + tc_rows;
    if(tc_new_rows > tc_offset) {
      printf_ndc("Extending from %d rows to %d rows", (int)tc_offset, (int)tc_new_rows);
      int res = H5Dset_extent(tc_dataset, &tc_new_rows);
      assert(res >= 0);
    } else {
      printf_ndc("Not extending: keeping %d rows", (int)tc_offset);
    }
    tc_dataspace = H5Dget_space(tc_dataset);
    assert(H5Sget_simple_extent_npoints(tc_dataspace) == tc_new_rows);
  }

  assert(H5Sget_simple_extent_npoints(tc_dataspace) >= tc_rows+tc_offset);
  assert(tc_dataset >= 0);
  assert(tc_dataspace >= 0);

/* ok lets write out the tags... do it all on rank 0... 
   I can write everything out all at once too
*/

  if(myrank == 0) {
    hid_t tc_plist_id = H5Pcreate (H5P_DATASET_XFER);
    H5Pset_dxpl_mpio(tc_plist_id, H5FD_MPIO_INDEPENDENT);
    assert(tc_plist_id >= 0);
    linkedlist *tdl = tags_to_write;

    damsel_hdf5_infile_tag_container *tc_data = malloc(sizeof(damsel_hdf5_infile_tag_container) * tc_rows);
    assert(tc_data != NULL);
    int i;
    for(i=0; i<tc_rows; i++) {
      damsel_tag_internal *td = (damsel_tag_internal *)tdl->head;
      memset(tc_data[i].name, 0, DAMSEL_HDF5_MAX_TAG_SIZE);
      strncpy(tc_data[i].name, td->name, DAMSEL_HDF5_MAX_TAG_SIZE-1);
      tc_data[i].name[DAMSEL_HDF5_MAX_TAG_SIZE-1] = 0;

      push_printf_ndc("Copying tag definition %s", tc_data[i].name);
      tc_data[i].tag_datatype = damsel_type_to_serializable_type(td->type);
      pop_ndc();
      damsel_handle fileside_handle;
      damsel_err_t ret = handle_map(model, DAMSEL_TRANSFER_TYPE_WRITE, td->handle, &fileside_handle);
      if(ret.id != 0) {
        printf_ndc("unhandled case: tag is not mapped into the file");
        // this is not a tag to write out. but this code is not structured to allowed that at the moment. so for now, we'll write out the zero handle; or more likely in debug mode, we'll die here. TODO/BUG
        fileside_handle = 0;
        assert(1==0);
      }
      printf_ndc("Mapped app side handle %llx to file side handle %llx", td->handle, fileside_handle);
      tc_data[i].tag_handle = fileside_handle;
      // TODO: write out file side handle rather than app side handle.
      tdl=tdl->next;
    }

    assert(tdl == NULL); // we should hit the end of the list after tc_rows

    hsize_t tc_dataspace_in_mem_size = tc_rows;
    hid_t tc_dataspace_in_mem = H5Screate_simple(1, &tc_dataspace_in_mem_size, &tc_dataspace_in_mem_size);
    assert(tc_dataspace_in_mem >= 0);

    printf_ndc("selecting hyperslap starting at tc_offset = %d for tc_rows = %d rows", (int)tc_offset, (int)tc_rows);
    int res = H5Sselect_hyperslab(tc_dataspace, H5S_SELECT_SET, &tc_offset, NULL, &tc_rows, NULL);
    assert(res >= 0);

    assert(H5Sget_simple_extent_npoints(tc_dataspace) >= tc_rows+tc_offset);
    assert(H5Sget_simple_extent_npoints(tc_dataspace_in_mem) == tc_rows);

    if (H5Sget_simple_extent_npoints(tc_dataspace_in_mem)>0) {
      res = H5Dwrite(tc_dataset, tag_container_type, tc_dataspace_in_mem, tc_dataspace, tc_plist_id, tc_data);
      assert(res >= 0);
    }

    free(tc_data);
    H5Pclose(tc_plist_id);
  } else {
    // non-zero rank - do nothing here
  }


  int ret;
  ret = H5Dclose(tc_dataset);
  assert(ret >= 0);
  ret = H5Sclose(tc_dataspace);
  assert(ret >= 0);

  linkedlist_free(tags_to_write, NULL);

  pop_ndc();

  fflush(NULL);
}

static void store_containers_vec(damsel_model_internal *model, damsel_attached_file *fs) {
  assert(model != NULL);
  assert(model->direction == DAMSEL_TRANSFER_TYPE_WRITE);



  push_ndc("creating vector handle containers");

  linkedlist *ll;
  int i;

  int maxrank, myrank;
  maxrank = fs->maxrank;
  myrank = fs->myrank;

  int local_vc_rows = 0;

  for(ll = fs->first_pushed_container; ll!=NULL; ll=ll->next) {

    damsel_hdf5_internal_pushed_container *pc = (damsel_hdf5_internal_pushed_container *)ll->head;
    damsel_container_internal *vhandles = (damsel_container_internal *)pc->file_container;
    if(vhandles->type == DAMSEL_HANDLE_CONTAINER_TYPE_VECTOR && pc->remote == 0 && vhandles->written == DAMSEL_UNWRITTEN) {
      local_vc_rows += vhandles->quantity / VECTOR_MAX + 1;
    }
  }

  int *per_rank_vc_count;
  per_rank_vc_count = malloc(sizeof(int) * maxrank);
  assert(per_rank_vc_count != NULL);

  printf_ndc("rank %d local_vc_rows = %d", myrank,
    local_vc_rows);
  MPI_Allgather(&local_vc_rows, 1, MPI_INT,
                per_rank_vc_count, 1, MPI_INT,
                 fs->communicator);

  hsize_t vc_rows = 0;
  for(i = 0; i < maxrank; i++) {
    vc_rows += per_rank_vc_count[i];
  }

  int my_vc_base = 0;

  for(i = 0; i < (myrank-1); i++) {
    my_vc_base += per_rank_vc_count[i];
  }

  hid_t vector_container_type = get_vector_handle_container_type(fs);
  assert(vector_container_type >= 0);

  hid_t vc_dataspace = -1;


  hid_t vc_dataspace_one = H5Screate_simple(1, &hsize_one, &hsize_one);
  assert(vc_dataspace_one >= 0);

  hid_t vc_dataset = -1;


  if(H5Lexists(fs->file_id, DAMSEL_DATASET_CONTAINER_VECTOR, H5P_DEFAULT) <= 0) {
    vc_dataspace = H5Screate_simple(1, &vc_rows, &vc_rows);
    assert(vc_dataspace >= 0);
    hid_t skip_attribute_update_plist = H5Pcreate(H5P_DATASET_CREATE);
    H5Pset_obj_track_times(skip_attribute_update_plist, 0);
    vc_dataset = H5Dcreate(fs->file_id, DAMSEL_DATASET_CONTAINER_VECTOR,
      vector_container_type, vc_dataspace, H5P_DEFAULT, 
      skip_attribute_update_plist, H5P_DEFAULT);
    assert(vc_dataset >= 0);
    H5Pclose(skip_attribute_update_plist);
  } else {

    vc_dataset = H5Dopen(fs->file_id, DAMSEL_DATASET_CONTAINER_VECTOR, H5P_DEFAULT);
    assert(vc_dataset >= 0);

    vc_dataspace = H5Dget_space(vc_dataset);
    hsize_t vc_offset = H5Sget_simple_extent_npoints(vc_dataspace);
    hsize_t vc_new_rows = vc_offset + vc_rows;
    H5Dset_extent(vc_dataset, &vc_new_rows);
    vc_dataspace = H5Dget_space(vc_dataset);
    my_vc_base += vc_offset;
  }

  assert(vc_dataset >= 0);
  assert(vc_dataspace >= 0);

// now I need to actually write out some values...
// this will be like the sequence handle container writing - independent mode,
// not collective mode

  pop_ndc();
  void *pvhc_ctx = push_ndc("populating vector handle containers");
  printf_ndc("rank %d will put vector handle container rows starting at offset %d",
    myrank, my_vc_base);

  hid_t vc_plist_id = H5Pcreate (H5P_DATASET_XFER);
  H5Pset_dxpl_mpio (vc_plist_id, H5FD_MPIO_INDEPENDENT);
  assert(vc_plist_id >= 0);

  i=0;

  for(ll = fs->first_pushed_container; ll != NULL; ll=ll->next) {

    damsel_hdf5_internal_pushed_container *pc = (damsel_hdf5_internal_pushed_container *)ll->head;
    damsel_container_internal *vhandles = (damsel_container_internal *)pc->file_container;

    assert(vhandles!=NULL);
    assert_valid_handle_container_id(vhandles);

    if(vhandles->type == DAMSEL_HANDLE_CONTAINER_TYPE_VECTOR && pc->remote == 0 && vhandles->written == DAMSEL_UNWRITTEN) {
      assert(i<local_vc_rows);
      damsel_vector_handle_container *vc = (damsel_vector_handle_container *)vhandles;
      assert(((damsel_store_internal *)(vc->container.model))->type == DAMSEL_STORETYPE_HDF5);

      damsel_hdf5_infile_vector_handle_container_row data;
      data.handle = internal_push_container_lookup_fileside((damsel_container) vhandles);
      data.quantity= vhandles->quantity;

      int blocknum;
      for(blocknum = 0; blocknum < (vhandles->quantity / VECTOR_MAX + 1) ; blocknum++) {
        push_printf_ndc("writing block %d", blocknum);
        assert(i<local_vc_rows);
        data.blocknumber = blocknum;
        // are we on the last block?
        int blockmax = (blocknum == vhandles->quantity / VECTOR_MAX) ? vhandles->quantity % VECTOR_MAX : VECTOR_MAX;
        int j;
        for(j=0; j < blockmax; j++) {
          assert(j < VECTOR_MAX);
          data.content[j] = vc->contents[j + blocknum * VECTOR_MAX];
        }

        // this loop ensures that the rest of the structure is initialised
        // so that valgrind is happy, and so that we don't write out random
        // values.
        for(;j<VECTOR_MAX; j++) {
          data.content[j] = 0;
        }
        assert(i<local_vc_rows);
        int res;
        res = H5Sselect_none(vc_dataspace);
        assert(res >= 0);
        hsize_t coords = my_vc_base + i;
        res = H5Sselect_elements(vc_dataspace, H5S_SELECT_SET, 1, &coords);
        assert(res >= 0);
        printf_ndc("rank %d writing vector container storage row %d (local row %d out of %d) for block %d of vector container %p", myrank, (int)coords, i, (int)local_vc_rows, blocknum, vhandles);
        res = H5Dwrite(vc_dataset, vector_container_type, vc_dataspace_one, vc_dataspace,
          vc_plist_id, &data);
        assert(res >= 0);
        i++;
        pop_ndc();
      }
    }
  }

  pop_ndc_context(pvhc_ctx);

  assert(i == local_vc_rows);
  assert(ll==NULL); // we should have hit the end of this linked list at
                         // the same time that we hit the end of the loop.

// now back into collective mode

  free(per_rank_vc_count);

  int ret;

  ret = H5Dclose(vc_dataset);
  assert(ret >= 0);
  ret = H5Sclose(vc_dataspace);
  assert(ret >= 0);
  ret = H5Sclose(vc_dataspace_one);
  assert(ret >= 0);

  printf_ndc("proc %d barriering after vector handle container create.", myrank);
  fflush(NULL);
}


static void store_collections(damsel_model_internal *src_s, damsel_attached_file *fs) {

// pull these straight out of the model, rather than via a mapped version.
// I think this makes things less tangled with handles?

  void *ctx = push_ndc("creating collection dataset");
  assert(src_s->direction == DAMSEL_TRANSFER_TYPE_WRITE);

  int i;

  int local_collection_rows = linkedlist_length(src_s->first_collection);

  int maxrank, myrank;
  maxrank = fs->maxrank;
  myrank = fs->myrank;

  int *per_rank_collection_count;
  per_rank_collection_count = malloc(sizeof(int) * maxrank);
  assert(per_rank_collection_count != NULL);

  printf_ndc("rank %d local_collection_rows = %d", myrank,
    local_collection_rows);
  MPI_Allgather(&local_collection_rows, 1, MPI_INT,
                per_rank_collection_count, 1, MPI_INT,
                 fs->communicator);

  hsize_t collection_rows = 0;
  for(i = 0; i < maxrank; i++) {
    collection_rows += per_rank_collection_count[i];
  }

  int my_collection_base = 0;
  for(i = 0; i < (myrank-1); i++) {
    my_collection_base += per_rank_collection_count[i];
  }

  hid_t collection_type = get_collection_type(fs);
  assert(collection_type >= 0);

  hid_t collection_dataspace_one = H5Screate_simple(1, &hsize_one, &hsize_one);
  assert(collection_dataspace_one >= 0);

  hid_t collection_dataspace = -1;
  hid_t collection_dataset = -1;
  if (H5Lexists(fs->file_id, DAMSEL_DATASET_COLLECTION, H5P_DEFAULT) <= 0) {
    collection_dataspace = H5Screate_simple(1, &collection_rows, &collection_rows);
    assert(collection_dataspace >= 0);
    collection_dataset = H5Dcreate(fs->file_id, DAMSEL_DATASET_COLLECTION,
      collection_type, collection_dataspace, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
    assert(collection_dataset >= 0);
  } else {
    collection_dataset = H5Dopen(fs->file_id, DAMSEL_DATASET_COLLECTION, H5P_DEFAULT);
    collection_dataspace = H5Dget_space(collection_dataset);
    hsize_t collection_offset = H5Sget_simple_extent_npoints(collection_dataspace);
    hsize_t collection_new_rows = collection_offset + collection_rows;
    H5Dset_extent(collection_dataset, &collection_new_rows);
    my_collection_base += collection_offset;
  }

  assert(collection_dataspace >= 0);
  assert(collection_dataset >= 0);

  pop_ndc_context(ctx);
  ctx = push_ndc("populating collection dataset");

  printf_ndc("rank %d will put collections starting at offset %d",
    myrank, my_collection_base);

  hid_t collection_plist_id = H5Pcreate (H5P_DATASET_XFER);
  H5Pset_dxpl_mpio (collection_plist_id, H5FD_MPIO_INDEPENDENT);
  assert(collection_plist_id >= 0);

  i=0;
  linkedlist *collectionlist = src_s->first_collection;
  while(collectionlist != NULL) {
    assert(collectionlist!=NULL);
    damsel_collection_internal *ci = (damsel_collection_internal *)collectionlist->head;
    assert(ci!=NULL);
    damsel_hdf5_infile_collection data;
    memset(&data, 0, sizeof(damsel_hdf5_infile_collection));
    damsel_err_t r = handle_map(src_s, DAMSEL_TRANSFER_TYPE_WRITE, ci->collection_handle, &data.collection_handle);
    assert_dmsl_ok(r);
    data.type = ci->type;
    data.container_handle = internal_push_container_lookup(ci->container);
    int res;
    res = H5Sselect_none(collection_dataspace);
    assert(res >= 0);
    hsize_t coords = my_collection_base + i;
    res = H5Sselect_elements(collection_dataspace, H5S_SELECT_SET, 1, &coords);
    assert(res >= 0);
    printf_ndc("rank %d writing collection at %d", myrank, (int)coords);
    res = H5Dwrite(collection_dataset, collection_type, collection_dataspace_one, collection_dataspace,
      collection_plist_id, &data);
    assert(res >= 0);
    i++;
    collectionlist = collectionlist->next;
  }
  assert(i == local_collection_rows);
  assert(collectionlist==NULL);

  free(per_rank_collection_count);

  int ret;
  ret = H5Dclose(collection_dataset);
  assert(ret >= 0);
  ret = H5Sclose(collection_dataspace);
  assert(ret >= 0);
  ret = H5Sclose(collection_dataspace_one);
  assert(ret >= 0);

  pop_ndc_context(ctx);
}
//  use a linked list to keep the dataset information for writing data once.
typedef struct dataset_write_info {
    damsel_model_internal *src_s; 
    damsel_attached_file *fs; 
	damsel_cached_data_internal *dcd; 
	damsel_tag_internal *tagdef;
    char name[DAMSEL_DSET_NAME_MAX];
    hsize_t count;
	hsize_t size;
    hsize_t *coord_f;
    hsize_t *coord_m;
} dataset_write_info;

/* Each dataset is uniquely identified by its name in the linked list. */
dataset_write_info *find_dataset_write_info(linkedlist *list, char *name) {
  linkedlist *head = list;
  dataset_write_info *info;
  
  while(head != NULL) {
    info = (dataset_write_info *) head->head;
	if (strcmp (info->name, name) == 0)
	  return info;
		
    head = head->next;
  }
  
  return NULL;
}

/* free memory held by the list and its members */
static void free_dataset_write_info(void *dset_write_info) {
  dataset_write_info *info = (dataset_write_info *) dset_write_info;
  if (info) {
    if (info->coord_f) {
      free(info->coord_f);
      info->coord_f = NULL;
    }
	  
    if (info->coord_m) {
      free(info->coord_m);
      info->coord_m = NULL;
    }
	  
    free(info);
    info = NULL;	
  }
}

/* add a information to the list. A new linked item is added to
   the list only when the name of the dataset entry does not exist
   in the linked list. If a dataset exists in the list, just add the
   data coordinate of the data point to the existing item.
*/
static void add_dataset_write_info(linkedlist **la, damsel_model_internal *src_s, 
      damsel_attached_file *fs, damsel_cached_data_internal *dcd, 
	  damsel_tag_internal *tagdef, const char *name, hsize_t size, 
	  hsize_t coord_f, hsize_t coord_m) 
{
  if (name == NULL || strlen(name)<=0)
    return;
	
  dataset_write_info *info = find_dataset_write_info( *la, (char*)(name));

  //printf("add_dataset_write_info: f=%lld,\t m=%lld\n", coord_f, coord_m);
  
  if (info && (info->count < info->size) && dcd) {
    info->coord_f[info->count] = coord_f;
    info->coord_m[info->count] = coord_m;
	info->count++;
    assert(coord_m < info->size);
    assert(size == info->size);
} else if (size>0) {
    info = malloc (sizeof (dataset_write_info));
	memset(info, 0, sizeof (dataset_write_info));
    info->src_s = src_s; 
    info->fs = fs; 
	info->dcd = dcd; 
	info->tagdef = tagdef;
    strcpy(info->name, name);
	info->coord_f = calloc(size, sizeof(hsize_t));
	info->coord_f[0] = coord_f;
	info->coord_m = calloc(size, sizeof(hsize_t));
	info->coord_m[0] = coord_m;
        assert(coord_m < size);
	info->size = size;
	info->count = 1;

	linkedlist_prepend(la, info);
  }
}

static void store_data(damsel_model_internal *src_s, damsel_attached_file *fs) {

  uint32_t local_cause, global_cause;
  void *ctx = push_ndc("store_data: Writing data");
  assert(src_s->direction == DAMSEL_TRANSFER_TYPE_WRITE);

  linkedlist *dataset_write_info_list = NULL; /* for writing data values once */
 
  hid_t plist_id = H5Pcreate (H5P_DATASET_XFER);

#ifndef TEST_COLLECTIVE_IO  
  H5Pset_dxpl_mpio (plist_id, H5FD_MPIO_INDEPENDENT);
#else  
  H5Pset_dxpl_mpio (plist_id, H5FD_MPIO_COLLECTIVE);
#endif
  
  assert(plist_id >= 0);

  // for each tag that is registered with this store:
  // output the tag definition
  // then go through the source handle container, and output the subset of
  // handles in that tag definition that are in the source handle container,
  // giving them the appropriate mapped destination handle.

  // as the handle containers can only be ranges at the moment, that should
  // be fairly straightforward.

  int myrank;
  MPI_Comm_rank(fs->communicator, &myrank);

  damsel_cached_data_internal *dcd;
  for(dcd = src_s->first_cached_data; dcd != NULL; dcd = dcd->next) {
    void *ctx_dcd = push_printf_ndc("in loop, dcd = %p", dcd);
    int pos;
    int maxpos = DMSLcontainer_count(dcd->handles_id);
	
    damsel_handle fileside_dcd_tag_handle;
    damsel_err_t ret = handle_map(src_s, DAMSEL_TRANSFER_TYPE_WRITE, dcd->tag_handle, &fileside_dcd_tag_handle);
    assert_dmsl_ok(ret);
    for(pos=0; pos<maxpos; pos++) {
      printf_ndc("processing position %d", pos);

      damsel_handle datapoint_app_side_handle = DMSLcontainer_handle_at_position(dcd->handles_id, pos);

      damsel_handle datapoint_file_side_handle;
      damsel_err_t ret = handle_map(src_s, DAMSEL_TRANSFER_TYPE_WRITE, datapoint_app_side_handle, &datapoint_file_side_handle);
      if(ret.id != 0) {
        printf_ndc("handle_map failed, datapoint_app_side_handle=%llx", datapoint_app_side_handle);
      }
      assert(ret.id == 0);
      linkedlist *chl;
      for(chl=src_s->first_chunk; chl!= NULL; chl = chl->next) {
        damsel_internal_chunk *chunk = (damsel_internal_chunk *)chl->head;
        assert(chunk!=NULL);

        damsel_tag_internal *tagdef = chunk->tag;

        void *ctx_tag = push_printf_ndc("Proc %d writing out tag %s to HDF5", myrank, tagdef->name);
        assert_valid_tag_definition_structure(tagdef);
        printf_ndc("tag %s has handle %llx", tagdef->name, tagdef->handle);

        printf_ndc("comparing dcd and tag definition. dcd->tag_handle = %llx tagdef->handle=%llx ret=%d", dcd->tag_handle, tagdef->handle, ret.id);

        if(ret.id == 0 && dcd->tag_handle == tagdef->handle) {
          void *ctx_mdcd = push_printf_ndc("found a matching cached data block for this tag, fileside_dcd_tag_handle=%llx", fileside_dcd_tag_handle);
          int writepos = find_handle_in_container(chunk->container, datapoint_file_side_handle);
          if(writepos >= 0) {
            assert(writepos < DMSLcontainer_count((damsel_container) chunk->container));
            damsel_handle file_container_handle = internal_push_container_lookup_fileside((damsel_container) chunk->container); // TODO: this might not do remotes right?
            assert(file_container_handle != 0);
            char datasetname[DAMSEL_DSET_NAME_MAX] = ""; // better hope this is big enough...
            generate_chunk_name(datasetname, tagdef->name, file_container_handle);
            printf_ndc("rank(%d): %s: datasetname=%s\n", myrank, __func__, datasetname);

            const hsize_t hsize_mem = maxpos;
            printf_ndc("pos = %d, hsize_mem = %d", (int)pos, (int)hsize_mem);
            assert(pos < hsize_mem);
			add_dataset_write_info(&dataset_write_info_list, src_s, fs, dcd, tagdef, 
			                        datasetname, hsize_mem, (hsize_t) writepos, (hsize_t) pos);
          }
          pop_ndc_context(ctx_mdcd);
        }
        pop_ndc_context(ctx_tag);
      }
    }
    pop_ndc_context(ctx_dcd);
  }
 
#ifdef DAMSEL_HDF5_MULTI
   size_t ndsets_to_write = 0;
   int ndsets_in_list = linkedlist_length(dataset_write_info_list);
   H5D_rw_multi_t *writeblock = NULL;
   if (ndsets_in_list>0)
       writeblock = calloc(ndsets_in_list, sizeof(H5D_rw_multi_t));
#endif	   
  
  linkedlist *lp = dataset_write_info_list;
  int res, i;
  while(lp) {
    int free_outdata=0;
	void *outdata = NULL;
	
    dataset_write_info *info = (dataset_write_info *)lp->head;
	
    outdata = damsel_data_to_hdf5_data(info->src_s, info->fs, 
	    info->dcd, info->tagdef, info->name, info->size);
	
	if (outdata)
	    free_outdata=1;
	else
		outdata = info->dcd->data;
	
	if (outdata) {
            printf_ndc("writing info->count=%d elements for memory block size info->size=%d", (int) info->count, (int) info->size);
            printf_ndc("first file offset is %d", (int)*(info->coord_f));
            printf_ndc("first mem offset is %d", (int)*(info->coord_m));
            hid_t dataset = H5Dopen(info->fs->file_id, info->name, H5P_DEFAULT);
            hid_t datatype = H5Dget_type(dataset);
            hid_t dataspace_mem = H5Screate_simple(1, &(info->size), NULL);
            hid_t dataspace_file = H5Dget_space(dataset);
            printf_ndc("dataset has space of size %d", (int)H5Sget_simple_extent_npoints(dataspace_file));
            res = H5Sselect_elements(dataspace_file, H5S_SELECT_SET, info->count, info->coord_f);
            assert(res >= 0);
            res = H5Sselect_elements(dataspace_mem,  H5S_SELECT_SET, info->count, info->coord_m);
            assert(res >= 0);
#ifdef DAMSEL_HDF5_MULTI
            writeblock[ndsets_to_write].dset_id = dataset;
            writeblock[ndsets_to_write].mem_type_id = datatype;
            writeblock[ndsets_to_write].mem_space_id = dataspace_mem;
            writeblock[ndsets_to_write].file_space_id = dataspace_file;
	    writeblock[ndsets_to_write].wbuf = outdata;
#if 0
	    writeblock[ndsets_to_write].flags = H5D_MULTI_CLOSE_ALL;
	    if(free_outdata && outdata)
		writeblock[ndsets_to_write].flags += H5D_MULTI_FREE_BUF;
#endif
	    ndsets_to_write++;
#else		
            /* collective IO will hang. H5Dwrite_multi() will handle collective IO */
            H5Pset_dxpl_mpio (plist_id, H5FD_MPIO_INDEPENDENT);	
            res = H5Dwrite(dataset, datatype, dataspace_mem, dataspace_file, plist_id, outdata);
            assert(res >= 0);
#ifdef HAVE_H5PGET_MPIO_NO_COLLECTIVE_CAUSE
	    H5Pget_mpio_no_collective_cause(plist_id, &local_cause, &global_cause);
	    printf_ndc("non-multi collective I/O carried out with flags 0x%x 0x%x\n", 
		    local_cause, global_cause);
#endif
            if(free_outdata && outdata) free(outdata);
            res = H5Tclose(datatype);
            assert(res >= 0);
            res = H5Sclose(dataspace_file);
            assert(res >= 0);
            res = H5Sclose(dataspace_mem);
            assert(res >= 0);
            res = H5Dclose(dataset);
            assert(res >= 0);
#endif			
	} /* if (outdata) { */
	
	lp = lp->next;
  }  /*  while(lp) { */

#ifdef DAMSEL_HDF5_MULTI
   H5Dwrite_multi(fs->file_id, ndsets_to_write, writeblock, plist_id);
#ifdef HAVE_H5PGET_MPIO_NO_COLLECTIVE_CAUSE
   H5Pget_mpio_no_collective_cause(plist_id, &local_cause, &global_cause);
   printf_ndc("multi-io collective I/O carried out with flags %x %x\n", 
	   local_cause, global_cause);
#endif

   for (i=0; i<ndsets_to_write; i++) {
       H5Dclose(writeblock[i].dset_id);
   }
   /*H5Dclose_multi(ndsets_to_write, (H5D_rw_multi_t *)writeblock); */
	 
   if (writeblock)	 
     free (writeblock);
#endif  
  
  linkedlist_free(dataset_write_info_list, free_dataset_write_info);
  
  printf_ndc("proc %d done writing all memory blocks", myrank);
  pop_ndc_context(ctx);
}

void damsel_execute_memory_to_hdf5(damsel_model_internal *src_s, damsel_attached_file *fs, MPI_Request *mpireq) {

  assert(src_s->direction == DAMSEL_TRANSFER_TYPE_WRITE);

  assert(fs->store.type == DAMSEL_STORETYPE_HDF5);

  void *ctx = push_ndc("damsel_execute_memory_to_hdf5 - write from memory to HDF5 file");

// invent handles for metadata handles - these are handles that need to
// end up in the HDF5 file, but that the application has not explicitly
// mapped.

  // todo: can i do this after the resolving handles stage so that it
  // captures both explicit and implicitly defined ones?
  push_predefined_containers(src_s, fs);
#ifdef DAMSEL_SUBFILE
  printf_ndc("src_s->num_subfiles=%d\n", src_s->num_subfiles);
  if(src_s->num_subfiles > 1)
    push_predefined_containers(src_s, src_s->attached_subfile);
#endif

  resolve_handles_for_invented_mappings(src_s, fs);
#if 0 /* metadata list is freed already. */
#ifdef DAMSEL_SUBFILE
  if(src_s->num_subfiles > 1)
    resolve_handles_for_invented_mappings(src_s, src_s->attached_subfile);
#endif
#endif

  src_s->more_mappings_allowed = 0;

  resolve_global_mappings(src_s, fs);
  resolve_remote_mappings(src_s, fs);
#ifdef DAMSEL_SUBFILE
  if(src_s->num_subfiles > 1) {
    ;
    //resolve_global_mappings(src_s, src_s->attached_subfile);
    //resolve_remote_mappings(src_s, src_s->attached_subfile);
  }
#endif

  void *se_ctx = push_ndc("calling store_entities()");
  store_entities(src_s, fs); // entity metadata hook
#ifdef DAMSEL_SUBFILE
  if(src_s->num_subfiles > 1)
    /* TODO: no need to store entities on subfiles? */
    ;//store_entities(src_s, src_s->attached_subfile); 
#endif
  pop_ndc_context(se_ctx);

#ifdef DAMSEL_SUBFILE
  if(src_s->num_subfiles > 1)
    store_subfiles(src_s, fs); /* store num_subfiles into the master */
#endif

  // define_fileside_tags(src_s, fs);

  push_collection_containers(src_s);
  // i need to do HDF5 meta data operations collectively.
  // that's everything except the writing of the data.

  // so lets iterate through the ranks, and allow each rank to specify
  // what is going to get done collectively, in turn.

// i think a lot of the following calls can commute with each other
// and also bits inside. more scope for parallelisation there...

  store_collections(src_s, fs);
#ifdef DAMSEL_SUBFILE
  if(src_s->num_subfiles > 1)
    store_collections(src_s, src_s->attached_subfile);
#endif
  store_tags(src_s, fs);
#ifdef DAMSEL_SUBFILE
  /* TODO: saving tags on subfiles breaks test-nconnectivitywrite.c */
  /* actually subfile doesn't need to maintain any tags? */
  if(src_s->num_subfiles > 1)
    ;//store_tags(src_s, src_s->attached_subfile);
#endif

  fs->container_push_allowed = 0;
#ifdef DAMSEL_SUBFILE
  /* TODO: if condition is really required? */
  if(src_s->num_subfiles > 1)
    src_s->attached_subfile->container_push_allowed = 0;
#endif

#ifndef DAMSEL_SUBFILE
  create_chunks(src_s, fs);
#else
  if(src_s->num_subfiles > 1)
    create_chunks(src_s, src_s->attached_subfile);
  else
    create_chunks(src_s, fs);
#endif

#ifndef DAMSEL_SUBFILE
  create_datasets(src_s, fs);
#else
  if(src_s->num_subfiles > 1)
    create_datasets(src_s, src_s->attached_subfile);
  else
    create_datasets(src_s, fs);
#endif
  store_containers_seq(src_s, fs);
#ifdef DAMSEL_SUBFILE
  if(src_s->num_subfiles > 1)
    store_containers_seq(src_s, src_s->attached_subfile);
#endif
  store_containers_vec(src_s, fs);
#ifdef DAMSEL_SUBFILE
  if(src_s->num_subfiles > 1)
    store_containers_vec(src_s, src_s->attached_subfile);
#endif

#ifndef DAMSEL_SUBFILE
  store_data(src_s, fs);
  store_version(src_s, fs);
  store_dependencies(src_s, fs);

#else
  if(src_s->num_subfiles > 1) {
    /* store subfile range info in the original (master) file */
    store_data_range(src_s, fs);
    store_version(src_s, fs);
    store_dependencies(src_s, fs);
    /* store data into subfiles */
    store_data(src_s, src_s->attached_subfile);
  }
  else { /* normal file */
    store_data(src_s, fs);
    store_version(src_s, fs);
    store_dependencies(src_s, fs);
  }
#endif
  int res = H5Fflush(fs->file_id,H5F_SCOPE_GLOBAL);
  assert(res >= 0);
#ifdef DAMSEL_SUBFILE
  if(src_s->num_subfiles > 1) {
    int res2 = H5Fflush(src_s->attached_subfile->file_id,H5F_SCOPE_GLOBAL);
    assert(res2 >= 0);
  }
#endif
  printf_ndc("mpi request to complete is %lx", (unsigned long)*mpireq);
  pop_ndc_context(ctx);
}


static void internal_read_tag(damsel_attached_file *src_s, damsel_model_internal *s, damsel_tag_internal *tagdef) {
  push_printf_ndc("Reading tag %s to memory", tagdef->name);

  assert(src_s != NULL);
  assert(s!=NULL);

  damsel_cached_data_internal *dcd = s->first_cached_data;
  while(dcd != NULL) {
    push_printf_ndc("processing cached data block %p", dcd);
    if(dcd->tag_handle == tagdef->handle) {


      linkedlist *ml = s->mappings;
      while(ml != NULL ) {
        damsel_store_mapping *writes = (damsel_store_mapping *) ml->head;
        assert(writes->magic == DAMSEL_STORE_MAPPING_MAGIC);
        push_printf_ndc("Iterating over a mapping, writes = %p", writes);
        assert(writes!=NULL);
/* in a memory store execute, the damsel_cached_data_internal object is on the
   destination store (i.e. the memory store)
*/

      // so here, I know that i want to read any tag data for this tag from:
      //   disk store:   writes->src_store
      //   handle container:   writes->src_container
      // and read it into memory. How do I know where that memory is
      // defined? I need to search my own cached data list.
        printf_ndc("found a matching cached data block for tag %s", tagdef->name);
        // now does this cached data block overlap with any mapped
        // handle container?
        // we can cheat and short circuit here by comparing handle container
        // pointers - that will not give false positives but might give false
        // negatives
// TODO: this is the comparison which needs to turn into deeper evaluation
// of the in-file handle container table - each handle container may
// contain some but not all of the handles; and each handle container
// may or may not have tag values associated with the tag that we're trying
// to read.
// to begin with i should move along the handles in the in-memory container,
// reading in as many as I can greedily from each in-file container in turn.
// there might be fancier things to do involving figuring out how differing
// handle containers align. but not now.

        if(dcd->handles_id == writes->app_container_id  || handle_containers_overlap(dcd->handles_id, writes->app_container_id)) {
          printf_ndc("... and the handle containers match (handle container id = %p). dumping from file to memory.", dcd->handles_id);

/* TODO: both here and in the writer, we need to be using the handle in
   file space, not in memory space - the memory space handle will be
   different each time, potentially. (although maybe not in proto1 due to
   static allocation.
*/
          assert_is_damsel_type(tagdef->type);
          hid_t datatype=damsel_type_to_hdf5_type(tagdef->type);

// TODO: bug here - we can only read from sequence handle containers, not
// vector handle containers... but with the advent of selections, I need to
// be able to read from vector handle containers...

// TODO: rather than reading in from a single dataset based on the handle
// name, we need to go through the table of handle containers and find one
// which contains the first handle that we want.

          damsel_container_internal *specified_handle_container = (damsel_container_internal *)writes->file_container_id;
          printf_ndc("handle ptr = %p", specified_handle_container);

          assert(specified_handle_container != NULL);
          assert(specified_handle_container->type == DAMSEL_HANDLE_CONTAINER_TYPE_VECTOR || specified_handle_container->type == DAMSEL_HANDLE_CONTAINER_TYPE_SEQUENCE);

          // so now, I can figure out the first handle that I want to copy
          // into memory - the first one from this handle container.
// find the first handle in that container (from
// in-memory structures)

          damsel_handle first_file_handle;

          hsize_t dest_offset=0; // does this increase with the right stride?

          int position_in_container;
          int size_of_container = specified_handle_container->quantity;

          printf_ndc("size of container is %d", (int)size_of_container);
          push_ndc("scanning over container");
          for(position_in_container = 0;
              position_in_container < size_of_container;
              position_in_container++) {

            first_file_handle = DMSLcontainer_handle_at_position((damsel_container)specified_handle_container, position_in_container);

// TODO make it read into the appropriate client-side slot
            printf_ndc("(file-side) handle to read is %llx", first_file_handle);

// go through the handle containers on disk until we find the first
// on-disk handle container that contains the handle we want.
// this is stored in the DAMSEL_DATASET_CONTAINER_SEQ HDF5 dataset.

// BUG: its not necessarily the *first* on-disk handle container that
// is the one that has the storage dataset for the handle we are looking for
// so maybe I need to make all of this into a bigger loop that goes
// multiple times, for each entry to read.

// we need to have this open already because hdf5 dataset opening is a
// collective operation and this loop is doing things independently.
// ... and indeed we do.

// so could read whole table in to memory

            int handle_seq_container_count;
            int handle_vec_row_count;

// TODO: there's now a distinction between the count of vector containers
// and the number of HDF5 dataset rows needed to represent those containers.
            DMSLhdf5_get_container_counts(src_s, &handle_seq_container_count, &handle_vec_row_count);
            hssize_t hcseqsize = handle_seq_container_count;
            hssize_t hcvecsize = handle_vec_row_count;

            damsel_hdf5_infile_seq_handle_container *hcseqrows =
              malloc(handle_seq_container_count *
              sizeof(damsel_hdf5_infile_seq_handle_container));

            damsel_hdf5_infile_vector_handle_container_row *hcvecrows =
              malloc(handle_vec_row_count *
              sizeof(damsel_hdf5_infile_vector_handle_container_row));
#ifndef DAMSEL_SUBFILE
            DMSLhdf5_get_container_seq_infos(src_s, hcseqrows);
            DMSLhdf5_get_container_vec_infos(src_s, hcvecrows);
#else
            if(s->num_subfiles==0) {
              DMSLhdf5_get_container_seq_infos(src_s, hcseqrows);
              DMSLhdf5_get_container_vec_infos(src_s, hcvecrows);
            }
            else {
              assert(s->num_subfiles>1);
              DMSLhdf5_get_container_seq_infos(s->attached_file, hcseqrows);
              DMSLhdf5_get_container_vec_infos(s->attached_file, hcvecrows);
            }
#endif
            damsel_handle infile_container_handle = 0;
            int ofound=0;
            hsize_t found_offset=0;
            {
              hssize_t i;
              for(i=0;i<(hcseqsize+hcvecsize) && ofound == 0;i++) {
                int ifound = 0;
                push_printf_ndc("checking container %d", (int)i);
                if(i<hcseqsize) {
                  printf_ndc("checking sequence handle container %d", (int)i);
                  int vi = i;
                  int j;
                  int qty = hcseqrows[vi].quantity;
                  for(j=0; j<qty && ifound == 0 ; j++) {
                    damsel_handle candidate_handle = hcseqrows[vi].base + j*hcseqrows[vi].step;
                    if(first_file_handle == candidate_handle) {
                      printf_ndc("found");
                      ifound=1;
                      infile_container_handle = hcseqrows[vi].handle;
                      found_offset = j;
                    }
                  }
                } else {
                  int vi = i - hcseqsize;
                  push_printf_ndc("checking vector handle container %d", (int)i);
                  int j;
                  int qty = hcvecrows[vi].quantity;
                  for(j=0; j<qty && ifound == 0 ; j++) {
                    printf_ndc("probing offset %d", j);
                    damsel_handle candidate_handle = hcvecrows[vi].content[j];
                    if(first_file_handle == candidate_handle) {
                      printf_ndc("found");
                      ifound=1;
                      infile_container_handle = hcvecrows[vi].handle;
                      found_offset = j + VECTOR_MAX * hcvecrows[vi].blocknumber;
                    }
                  }
                  pop_ndc();
                }

// TODO: this is where nondeterminism should happen (that is, we should keep looping until
//       the next block of code runs ok).

                if(ifound == 1) {
            assert(ifound == 1);
            assert(infile_container_handle != 0);
            printf_ndc("found at infile_container_handle = %lld 0x%llx, found_offset=%d", infile_container_handle,
              infile_container_handle, (int)found_offset);
// for now read in a single data value and then loop...
// this will presumably be substantially less efficient than before,
// but there is clear scope for making this use ranges instead of looping,
// once I have this code actually working.

            char datasetname[DAMSEL_DSET_NAME_MAX] = ""; // better hope this is big enough...
            generate_chunk_name(datasetname, tagdef->name, infile_container_handle);
            assert(infile_container_handle != 0); // not sure about this one - it might be legitimate to have a handle of 0... but probably thats a problem elsewhere where I'm asserting that it isn't NULL?

                         // - we can't use tag name because we might write to
                         // the same tag repeatedly, on different file handles.
                         // perhaps we can make it a combination of the
                         // tag name and file container handle
            printf_ndc("trying to open HDF5 dataset '%s'.", datasetname);
	    hid_t dataset = -1;
#ifndef DAMSEL_SUBFILE
	    if (H5Lexists(src_s->file_id, datasetname, H5P_DEFAULT)==1) {
		dataset = H5Dopen(src_s->file_id, datasetname, H5P_DEFAULT);
	    }
#else
            if(s->num_subfiles == 0) {
              if (H5Lexists(src_s->file_id, datasetname, H5P_DEFAULT)==1) {
		dataset = H5Dopen(src_s->file_id, datasetname, H5P_DEFAULT);
              }
            }
            else {
              if (H5Lexists(s->attached_subfile->file_id, datasetname, H5P_DEFAULT)==1) {
		dataset = H5Dopen(s->attached_subfile->file_id, datasetname, H5P_DEFAULT);
              }
            }
#endif              

            if(dataset >= 0 ) { // open worked
// TODO: this is what might fail and should result into moving onto the next candidate
// fileside container.
              assert(dataset >= 0);

// hc_rows needs to be the number of elements in the on-disk handle container
// perhaps can get the dataspace for the container instead? (becuase
// we've opened it)
              hid_t el_dataspace = H5Dget_space(dataset);
              assert(el_dataspace >= 0);
              int ret = H5Sselect_none(el_dataspace);
              assert(ret >= 0);
              ret = H5Sselect_elements(el_dataspace, H5S_SELECT_SET, 1, &found_offset); // do we need to munge found_offset somehow (eg by a step size? for what i've done in prototype apps, file side handles always differ by a step of one but maybe not in general...)
              assert(ret >= 0);
              damsel_container_internal *dcdhc = (damsel_container_internal *)dcd->handles_id;
              assert(dcdhc != NULL); // if this fails, we've been given a handle container which doesn't exist
              assert(dcdhc->type == DAMSEL_HANDLE_CONTAINER_TYPE_SEQUENCE); // BUG: for now we can only deal with sequence types here... but should I be able to deal with other types?

              const hsize_t hsize_target = dcdhc->quantity /* size of target memory handle container */;
              hid_t el_dataspace_target = H5Screate_simple(1, &hsize_target, &hsize_target);
              assert(el_dataspace_target >= 0);
              printf_ndc("target dataset size is %d",(int)hsize_target);

              ret = H5Sselect_none(el_dataspace_target);
              assert(ret >= 0);

              ret = H5Sselect_elements(el_dataspace_target, H5S_SELECT_SET, 1, &dest_offset);
              assert(ret >= 0);

              push_printf_ndc("will read source dataset at offset %d into target offset %d of memory block at %p", (int) found_offset, (int)dest_offset, dcd->data);
// TODO: need this to write out to an offset inside dcd->data rather than
// the base point of dcd.
// how do I determine *where* in dcd that happens?
// dcd is a void pointer to a buffer...
// how do I add one to it? it depends on the size of the elements
// which depends on the tag...
// so I need to know the memory offset which is
//  element size * offset count

// whats the offset count?
// I know that because its the number of times i've gone round this loop
// or is it related to handle somehow?
// when handle is a pointer, yes related to handle, but when its not, then no...
// but round-loop count should be fine?
// or maybe i could use the select_elements call to pick out the appropriate
// dataspace on the memory side too? rather than always sleecting one out of
// a set of one... because the dataspace stuff was OK for reading before and
// maybe its what I want for reading in slices nayway?
// so el_dataspace_one needs to be a dataset the size of dcd->data
// which is the size of the input handle set...

              ret = H5Dread(dataset, datatype, el_dataspace_target, el_dataspace, H5P_DEFAULT, (void*)dcd->data);
              printf_ndc("H5Dread returned %d", (int) ret);
              // int bytepos; for(bytepos=0; bytepos<8; bytepos++) { printf_ndc("pos %d has value %d", bytepos, (int) (  ( (char *)dcd->data)[bytepos])); };
              assert(ret != -1);
              pop_ndc();


// i can get out of range errors here when reducing the number of ranks on
// a restart in proto2d. what's going on? we should be finding a different
// dataset and getting data from near the start of that, rather than
// one that's out of range from an earlier data set

              ret = H5Sclose(el_dataspace_target);
              assert(ret >= 0);
              ret = H5Sclose(el_dataspace);
              assert(ret >= 0);
              ofound = 1; // if we get this far...
			  ret = H5Dclose(dataset);
			  assert(ret >= 0);
            } else {
              printf_ndc("that dataset doesn't exist or couldn't be opened - looping");
            }
                }

                pop_ndc();
              }
            }

            if(ofound != 1) {
              printf_ndc("Could not find file-side handle %llx in file", first_file_handle);
              dump_ndc();
            }
            free(hcseqrows);
            free(hcvecrows);

// ok, now we've read in a single value. we need to loop...
            dest_offset++;
          }
          pop_ndc();
        } else {
            printf_ndc("handle containers do not overlap.");
        }

        ml = ml->next;
        pop_ndc();
      }
    } else {

      printf_ndc("skipping non-matching cached data block for tag %s", tagdef->name);
    }
    dcd = dcd->next;
    pop_ndc();
  }

  push_ndc("checking if we need to perform damsel-side conversion");
  if(tagdef->type == DAMSEL_DATA_TYPE_HANDLE) {

// now we need to iterate over all the read data: that is, the dcd
// entries. but only the relevant dcd entries. which is when: (dcd->tag_handle == tagdef->handle)

      damsel_cached_data_internal *dcd = s->first_cached_data;
      for(dcd = s->first_cached_data; dcd != NULL; dcd=dcd->next) {
        if(dcd->tag_handle == tagdef->handle) {
          int dest_offset;
          for(dest_offset=0; dest_offset<DMSLcontainer_count(dcd->handles_id); dest_offset++) {
            damsel_handle *targetdata = (damsel_handle *)dcd->data;
            damsel_handle new_handle;
            damsel_err_t r = handle_map(s, DAMSEL_TRANSFER_TYPE_READ, targetdata[dest_offset], &new_handle);
            if(r.id == 0) {
              targetdata[dest_offset] = new_handle;
            } else {
              targetdata[dest_offset] = 0; // BUG: we don't have an invalid handle to put into application side handles. We could perhaps have a boolean array of validity; or allow the user to pass in an invalid handle value.
            }
          }
        }
      }
  }
  pop_ndc();

  pop_ndc();
}

void damsel_execute_hdf5_to_memory(damsel_attached_file *src_s, damsel_model_internal *s, MPI_Request *mpireq)  {

  void *ctx = push_ndc("damsel_execute_hdf5_to_memory");

  assert(s!=NULL);
  assert(s->store.type == DAMSEL_STORETYPE_MEMORY);

  assert(src_s!=NULL);
  assert(src_s->store.type == DAMSEL_STORETYPE_HDF5);


/* so memory writes are writing data into memory (i.e. reading from disk) */

  // so we will have a bunch of tag reads that will specify a file store and a
  // handle container
  // and we need to read (some piece) of that handle container into memory

  linkedlist *rtl = s->first_tag_definition;
  while(rtl != NULL) {
#ifndef DAMSEL_SUBFILE
    internal_read_tag(src_s, s, (damsel_tag_internal *) rtl->head );
#else
    if(s->num_subfiles == 0)
      internal_read_tag(src_s, s, (damsel_tag_internal *) rtl->head );
    else {
      assert(s->num_subfiles > 1);
      damsel_attached_file *src_s_subfile = s->attached_subfile;
      assert(src_s_subfile!=NULL);
      assert(src_s_subfile->store.type == DAMSEL_STORETYPE_HDF5);
      internal_read_tag(src_s, s, (damsel_tag_internal *) rtl->head );
    }
#endif
    rtl = rtl->next;
  }

  printf_ndc("mpi request to complete is %lx", (unsigned long)*mpireq);
  pop_ndc_context(ctx);
}

void DMSLhdf5_get_container_counts(damsel_attached_file *attached_file,
  int *seq_count, int *vec_rows) {
  assert(attached_file != NULL);
  assert(attached_file->model_populated != NULL);
  if(seq_count != NULL) {
    *seq_count = attached_file->model_populated->seq_count;
  }

  if(vec_rows != NULL) {
    *vec_rows = attached_file->model_populated->vec_rows;
  }
}

void DMSLhdf5_get_container_seq_infos(damsel_attached_file *s, damsel_hdf5_infile_seq_handle_container *col) {
  assert(s != NULL);
  assert(s->model_populated != NULL);
  memcpy(col, s->model_populated->seq_info, s->model_populated->seq_count * sizeof(damsel_hdf5_infile_seq_handle_container));
}

void DMSLhdf5_get_container_vec_infos(damsel_attached_file *s, damsel_hdf5_infile_vector_handle_container_row *col) {
  assert(s != NULL);
  assert(s->model_populated != NULL);
  memcpy(col, s->model_populated->vec_info, s->model_populated->vec_rows * sizeof(damsel_hdf5_infile_vector_handle_container_row));
}

/*! returns the number of containers stored in this file, split up by storage type.
 */
void populate_hdf5_get_container_counts(damsel_attached_file *file, int *seq_count, int *vec_rows) {

  push_ndc("populate_hdf5_get_container_counts");

  hid_t spaceid;
  hssize_t hcsize;

  assert(file->store.type == DAMSEL_STORETYPE_HDF5);

  assert(file->seq_handle_container_dataset >= 0);
  spaceid = H5Dget_space(file->seq_handle_container_dataset);
  assert(spaceid >= 0);
  hcsize= H5Sget_simple_extent_npoints(spaceid);
  printf_ndc("there are %d sequence containers",
    (int) hcsize);
  int ret = H5Sclose(spaceid);
  assert(ret >= 0);
  *seq_count = hcsize;

  assert(file->vector_handle_container_dataset >= 0);
  spaceid = H5Dget_space(file->vector_handle_container_dataset);
  assert(spaceid >= 0);
  hcsize= H5Sget_simple_extent_npoints(spaceid);
  printf_ndc("vector handle container has %d rows",
    (int) hcsize);
  ret = H5Sclose(spaceid);
  assert(ret >= 0);
  *vec_rows = hcsize;

  pop_ndc();
}

void populate_hdf5_get_container_seq_infos(damsel_attached_file *s, damsel_hdf5_infile_seq_handle_container *col) {
  assert(s->store.type == DAMSEL_STORETYPE_HDF5);
  hid_t ct = get_sequence_handle_container_type(s);
  assert(ct>=0);
  int ret = H5Dread(s->seq_handle_container_dataset, ct,
            H5S_ALL, H5S_ALL, H5P_DEFAULT, col);
  assert(ret >= 0);
}


void populate_hdf5_get_container_vec_infos(damsel_attached_file *s, damsel_hdf5_infile_vector_handle_container_row *col) {
  assert(s->store.type == DAMSEL_STORETYPE_HDF5);
  hid_t ct = get_vector_handle_container_type(s);
  assert(ct>=0);
  int ret = H5Dread(s->vector_handle_container_dataset, ct,
            H5S_ALL, H5S_ALL, H5P_DEFAULT, col);
  assert(ret >= 0);
}

hid_t damsel_type_to_hdf5_type(damsel_data_type damsel_type) 
{
  int i;
  hid_t datatype=-1;

  assert_is_initialized();
  void *ctx = push_ndc("damsel_type_to_hdf5_type");

  if(damsel_type == DAMSEL_DATA_TYPE_SHORTINT) {
    printf_ndc("this is a short int");
    datatype = H5T_NATIVE_SHORT;
  } else if (damsel_type == DAMSEL_DATA_TYPE_BYTES) {
      printf_ndc("these are bytes");
      datatype = H5T_NATIVE_UINT8;
  } else if(damsel_type == DAMSEL_DATA_TYPE_FLOAT) {
      printf_ndc("this is a float");
      datatype = H5T_NATIVE_FLOAT;
  } else if(damsel_type == DAMSEL_DATA_TYPE_DOUBLE) {
      printf_ndc("this is a double");
      datatype = H5T_NATIVE_DOUBLE;
  } else if (damsel_type == DAMSEL_DATA_TYPE_INTEGER) {
      printf_ndc("this is an integer");
      datatype = H5T_NATIVE_INT;
  } else if (damsel_type == DAMSEL_DATA_TYPE_INT64) {
      printf_ndc("this is a 64 bit integer");
      datatype = H5T_NATIVE_INT64;
  } else if (damsel_type == DAMSEL_DATA_TYPE_HANDLE) {
      printf_ndc("this is a handle, which will be stored as a 64 bit integer");
      datatype = H5T_NATIVE_INT64;
  } else {
      void *datatype_struct = get_defined_type_struct(damsel_type);
      if(datatype_struct == NULL) {
        printf_ndc("could not find damsel type %llx to convert to hdf5 type.", damsel_type);
        pop_ndc_context(ctx);
        return -1;
      }

      switch(*(int *)datatype_struct) {
        case DAMSEL_I_TYPE_OPAQUE_MAGIC:
          {
            damsel_i_type_opaque *opaque = (damsel_i_type_opaque *) datatype_struct;
            printf_ndc("defining HDF5_NATIVE_OPAQUE with size %zu", opaque->bytecount);

            datatype = H5Tcreate(H5T_OPAQUE, opaque->bytecount);
            printf_ndc("hdf5 datatype = %x", (unsigned) datatype);
            assert(datatype >= 0);
          }
          break;

        case DAMSEL_I_TYPE_ARRAY_MAGIC:
          {
            damsel_i_type_array *array = (damsel_i_type_array *) datatype_struct;
            hid_t base_datatype;
            hsize_t dims[DAMSEL_I_TYPE_ARRAY_MAX_NDIMS];
            unsigned u;

            printf_ndc("defining HDF5_NATIVE_ARRAY with ndims %u", array->ndims);
            base_datatype = damsel_type_to_hdf5_type(array->base_type);
            printf_ndc("defining HDF5_NATIVE_ARRAY with base datatype %x", base_datatype);

            /* Translate array dims between types */
            for(u = 0; u < array->ndims; u++)
                dims[u] = array->dims[u];
            datatype = H5Tarray_create(base_datatype, array->ndims, dims);
            printf_ndc("hdf5 datatype = %x", (unsigned) datatype);
            assert(datatype >= 0);
          }
          break;
        case DAMSEL_I_TYPE_VECTOR_MAGIC:
          {
            damsel_i_type_vector *datetype_vector = (damsel_i_type_vector *) datatype_struct;
            hid_t base_datatype;

            printf_ndc("defining H5T_VLEN datatype");
            base_datatype = damsel_type_to_hdf5_type(datetype_vector->base_type);
            printf_ndc("defining H5T_VLEN with base datatype %x", base_datatype);

			/* TODO: HDF5 parallel IO does not support writing variable-length data. 
			   For a temporary solution, we will use fixed length array type of 
			   size max_len. 
			datatype = H5Tvlen_create (base_datatype);
			*/
			hsize_t dims[] = {(hsize_t)(datetype_vector->max_len)};
            datatype = H5Tarray_create(base_datatype, 1, dims);

			/* need to close the base type or it will cause memory leak.
			   we use TRY to handle error from closing predefined datatypes. */
			H5E_BEGIN_TRY {
              H5Tclose(base_datatype);
            } H5E_END_TRY

            printf_ndc("hdf5 datatype = %x", (unsigned) datatype);
            assert(datatype >= 0);
          }
          break;
        case DAMSEL_I_TYPE_ENUM_MAGIC:
          {
            damsel_i_type_enum *datetype_enum = (damsel_i_type_enum *) datatype_struct;

            printf_ndc("defining H5T_ENUM datatype");

			datatype = H5Tenum_create (H5T_NATIVE_INT);
			for (i = 0; i <datetype_enum->nmembs; i++) {
                                int ret = H5Tenum_insert (datatype, datetype_enum->names[i], &(datetype_enum->values[i]));
                                assert(ret >= 0);
                              }


            printf_ndc("hdf5 datatype = %x", (unsigned) datatype);
            assert(datatype >= 0);
          }
          break;
        case DAMSEL_I_TYPE_STRUCT_MAGIC:
          {
            hid_t *member_datatypes;
            damsel_i_type_struct *datetype_struct = (damsel_i_type_struct *) datatype_struct;

			if (datetype_struct->nmembs<1)
				return -1;

				
			member_datatypes = (hid_t *)calloc(datetype_struct->nmembs, sizeof(hid_t));

            printf_ndc("defining H5T_COMPOUND datatype");
			
			/* need to figure out the member types first */
			for (i=0; i<datetype_struct->nmembs; i++)
				member_datatypes[i] = damsel_type_to_hdf5_type(datetype_struct->types[i]);
	
			datatype = H5Tcreate(H5T_COMPOUND, datetype_struct->tsize);
			for (i=0; i<datetype_struct->nmembs; i++) {
                int ret = H5Tinsert( datatype, datetype_struct->names[i], datetype_struct->offsets[i], member_datatypes[i]);
                assert(ret >= 0);
            }

			/* need to close the base type or it will cause memory leak.
			   we use TRY to handle error from closing predefined datatypes. */
			for (i=0; i<datetype_struct->nmembs; i++) {
				H5E_BEGIN_TRY {
                H5Tclose(member_datatypes[i]);
				} H5E_END_TRY
			}

			free (member_datatypes);
			
            printf_ndc("hdf5 datatype = %x", (unsigned) datatype);
            assert(datatype >= 0);
          }
          break;
        default:
          assert(0 && "Unknown data type!");
          break;
      } /* end switch */

// TODO: going to need to close this type somewhere. but for now don't bother
// because I'm expecting to rework type handling eventually.
// TODO: i should probably cache the hdf5 datatype against the damsel type
// so that I can reuse it.
  }
  pop_ndc_context(ctx);
  return datatype;
}

void DMSLhdf5_get_tag_tuples(damsel_attached_file *s, damsel_hdf5_infile_tag_container *col) {
  assert(s!=NULL);
  assert(s->model_populated != NULL);
  assert(col!=NULL);
  memcpy(col, s->model_populated->tag_info, s->model_populated->tag_count * sizeof(damsel_hdf5_infile_tag_container));
}

void populate_hdf5_get_tag_count(damsel_attached_file *file, int *tag_count) {
  assert(file->tag_container_dataset >= 0);
  hid_t spaceid = H5Dget_space(file->tag_container_dataset);
  assert(spaceid >= 0);
  hssize_t hcsize= H5Sget_simple_extent_npoints(spaceid);
  printf_ndc("tag container has %d points",
    (int) hcsize);
  int ret = H5Sclose(spaceid);
  assert(ret >= 0);
  if(tag_count != NULL) {
    *tag_count = hcsize;
    printf_ndc("returning tag count to application.");
  } else {
    printf_ndc("skipping return of tag count to application.");
  }

}

void populate_hdf5_get_tag_tuples(damsel_attached_file *s, damsel_hdf5_infile_tag_container *col) {
  assert(s->store.type == DAMSEL_STORETYPE_HDF5);
  hid_t ct = get_tag_container_type(s);
  assert(ct>=0);
  int ret = H5Dread(s->tag_container_dataset, ct,
            H5S_ALL, H5S_ALL, H5P_DEFAULT, col);
  assert(ret >= 0);
}

damsel_err_t damsel_internal_hdf5_open(damsel_model_internal *mi, const char *filename, MPI_Comm comm, damsel_trait_bundle tb) {
  void *ctx = push_ndc("damsel_internal_hdf5_open");

// we may have a trait that describes if we should
// be opening for read or for write.
// if not, then do things automatically, or fail (undecided...)

  damsel_err_t err;
  char *mode;
  err = DMSLtrait_get(tb, "mode", &mode);
  printf_ndc("mode = %s", mode);

#ifdef DAMSEL_SUBFILE
  char *num_subfiles;
  err = DMSLtrait_get(tb, "dmsl_num_subfiles", &num_subfiles);
  int maxrank;
  MPI_Comm_size(comm, &maxrank);
  //printf_ndc("maxrank=%d, num_subfiles=%s\n", maxrank, num_subfiles);
  /* do not partition if maxrank is less than num_subfiles */
  if(num_subfiles != NULL && maxrank >= atoi(num_subfiles))
    mi->num_subfiles = atoi(num_subfiles);
#endif

#ifdef DAMSEL_SUBFILE
  if(mode != NULL && !strcmp(mode, "READ")) {
    mi->direction = DAMSEL_TRANSFER_TYPE_READ;
    mi->attached_file = damsel_internal_hdf5_reopen(filename, comm, tb, "READ");
    printf_ndc("checking whether file is partitioned...\n");
    /* read num_subfiles field from the master */
    if(H5Lexists(mi->attached_file->file_id, DAMSEL_DATASET_NUM_SUBFILES, H5P_DEFAULT) > 0) {
      int ret;
      hid_t dataset;

      dataset = H5Dopen(mi->attached_file->file_id, DAMSEL_DATASET_NUM_SUBFILES, H5P_DEFAULT);
      if(dataset >= 0) { /* open worked */
        assert(dataset >= 0);

        ret = H5Dread(dataset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &mi->num_subfiles);
        assert(ret >= 0);
        if(mi->num_subfiles > 1) {
          /* open the attached subfile */
          printf_ndc("file is partitioned into %d\n", mi->num_subfiles);
          int myrank, nprocs;
          MPI_Comm subfile_comm;
          char subfilename[1024];

          myrank = mi->attached_file->myrank;
          nprocs = mi->attached_file->maxrank;
          double ratio = (double)nprocs/(double)(mi->num_subfiles);
          int color = (int)(double)myrank/ratio;
          MPI_Comm_split(comm, color, myrank, &subfile_comm);

          sprintf(subfilename, "%s.subfile_%i", filename, color);
          printf_ndc("attached subfile is %s\n", subfilename);
          mi->attached_subfile = damsel_internal_hdf5_reopen(subfilename, subfile_comm, tb, "REOPEN");
        }
        ret = H5Dclose(dataset);
        assert(ret >= 0);
      }
    }
    else {
      printf_ndc("file is not partitioned. (num_subfiles=%d)\n",
		 mi->num_subfiles);
    }
    
  } else if(mode != NULL && !strcmp(mode, "WRITE")) {
    mi->attached_file = damsel_internal_hdf5_reopen(filename, comm, tb, "WRITE");

    if(mi->num_subfiles > 1 && mi->attached_file->maxrank >= mi->num_subfiles) {
        int myrank, nprocs;
        MPI_Comm subfile_comm;
        char subfilename[1024];

        myrank = mi->attached_file->myrank;
        nprocs = mi->attached_file->maxrank;
        double ratio = (double)nprocs/(double)(mi->num_subfiles);
        int color = (int)(double)myrank/ratio;
        MPI_Comm_split(comm, color, myrank, &subfile_comm);

        sprintf(subfilename, "%s.subfile_%i", filename, color);
        mi->attached_subfile = damsel_internal_hdf5_reopen(subfilename, subfile_comm, tb, "REOPEN");
    }
  } else if(mode != NULL && !strcmp(mode, "REOPEN")) {
    mi->attached_file = damsel_internal_hdf5_reopen(filename, comm, tb, "REOPEN");
  } else {
    printf_ndc("unknown mode %s", mode);
    assert(1==0);
  }
#else /* normal file open */
  if(mode != NULL && !strcmp(mode, "READ")) {
    mi->direction = DAMSEL_TRANSFER_TYPE_READ;
    mi->attached_file = damsel_internal_hdf5_reopen(filename, comm, tb, "READ");
  } else if(mode != NULL && !strcmp(mode, "WRITE")) {
    mi->attached_file = damsel_internal_hdf5_reopen(filename, comm, tb, "WRITE");
  } else if(mode != NULL && !strcmp(mode, "REOPEN")) {
    mi->attached_file = damsel_internal_hdf5_reopen(filename, comm, tb, "REOPEN");
  } else {
    printf_ndc("unknown mode %s", mode);
    assert(1==0);
  }
#endif

  if(mi->attached_file == NULL) {
    printf_ndc("Attached file is NULL - returning DMSL_ERROR_GENERAL");
    pop_ndc_context(ctx);
    return DMSL_ERROR_GENERAL;
  }

  assert(mi->attached_file != NULL);
  printf_ndc("Attached file is %p - returning DMSL_OK", mi->attached_file);
#ifdef DAMSEL_SUBFILE
  if(mi->num_subfiles > 1)
    assert(mi->attached_subfile != NULL);
    printf_ndc("Attached subfile is %p - returning DMSL_OK", mi->attached_subfile);
#endif
  pop_ndc_context(ctx);
  return DMSL_OK;
}

damsel_container_internal *damsel_handle_to_handle_container(damsel_attached_file *s,
  damsel_handle h) {
  // TODO: this is only going to work if we've pulled in all the containers
  // from the file already. I should at least assert that...
  push_ndc("damsel_handle_to_handle_container");
  assert(s->model_populated != NULL);
  printf_ndc("Looking up handle container referenced by handle (not id) %llx in store %p", h, s);
  linkedlist *ll = s->first_pushed_container;
  while(ll != NULL) {
    damsel_hdf5_internal_pushed_container *pc = (damsel_hdf5_internal_pushed_container *)ll->head;
    if(pc->handle == h) {
      printf_ndc("found!");
      pop_ndc();
      return (damsel_container_internal *)pc->file_container;
    }
    ll = ll->next;
  }
  printf_ndc("Couldn't find that handle container.");
  pop_ndc();
  return NULL;
}

/* convert vlen data to fixed length array data */
void *vlen_to_array_data(size_t npoints, damsel_vector_t *vector_data, hid_t tid)
{
    int i;
    void *data = NULL;
	
    if (npoints<=0)
	    return NULL;
	
    hsize_t adims[1];	
	hid_t atid = H5Tget_super(tid);
	size_t asize = H5Tget_size(atid);
	int ret = H5Tget_array_dims2(tid, adims);
          assert(ret >= 0);
	ret = H5Tclose(atid);
          assert(ret >= 0);

	data = calloc (npoints*adims[0], asize );
	damsel_vector_t val;
	size_t offset=0;
	for (i=0; i<npoints; i++) {
	    offset = i*adims[0]*asize;
		val = vector_data[i];
		size_t len = val.len;
		if (len > adims[0]) len = adims[0]; /* must not exceed the length of array. */
		memcpy ( data+offset, val.p, len*asize);
	}
	
	return data;
}

/* get the max length of the vlen data */
size_t get_max_vlen_length(size_t npoints, damsel_vector_t *vector_data)
{
    int i;
    size_t max_len = 0;
	
	for (i=0; i<npoints; i++) {
		if (vector_data[i].len > max_len)
		    max_len = vector_data[i].len;
	}
	
	return max_len;
}

/* converting damsel data to HDF5 data */
void *damsel_data_to_hdf5_data(damsel_model_internal *src_s, damsel_attached_file *fs, 
      damsel_cached_data_internal *dcd, damsel_tag_internal *tagdef, 
	  const char *datasetname, hsize_t hsize_mem)
{
    int i;
    void *outdata=NULL;
	
    // this is the point where we could translate from in-memory
    // values to in-file values when HDF5 cannot do it - for
    // example, when we're writing handles.
    if(tagdef->type == DAMSEL_DATA_TYPE_HANDLE) 
	{
        int count = DMSLcontainer_count(dcd->handles_id);
        damsel_handle *houtdata = calloc(count, sizeof(damsel_handle));
        outdata=houtdata;
        assert(outdata != NULL);
        for(i=0;i<count;i++) {
            handle_map(src_s, DAMSEL_TRANSFER_TYPE_WRITE, ((damsel_handle *)dcd->data)[i], &(houtdata[i]));
        }
    } else {
        void *datatype_struct = get_defined_type_struct(tagdef->type);
        if ( datatype_struct != NULL && ( *((int *)datatype_struct) == DAMSEL_I_TYPE_VECTOR_MAGIC) ) {
            //TODO: HDF5 parallel IO does not support writing VL datatypes yet. Use fixed length array	
            size_t new_len = get_max_vlen_length(hsize_mem, (damsel_vector_t *)dcd->data);
            damsel_i_type_vector *datetype_vector = (damsel_i_type_vector *) datatype_struct;
					
         	hid_t dataset = H5Dopen(fs->file_id, datasetname, H5P_DEFAULT); 
            hid_t datatype= H5Dget_type(dataset);				
            if (new_len > datetype_vector->max_len) {
                /* create a new datatype using a larger size */
                hsize_t adims[1] = {new_len};
                hid_t atid = H5Tget_super(datatype);
                int ret = H5Tclose(datatype);
                assert(ret >= 0);
                datatype = H5Tarray_create2( atid, 1, adims);
                ret = H5Tclose(atid);
                assert(ret >= 0);

                hid_t sid = H5Dget_space(dataset);

                /* create a new dataset */
                ret = H5Dclose(dataset);
                assert(ret >= 0);
                ret = H5Ldelete( fs->file_id, datasetname, H5P_DEFAULT);
                assert(ret >= 0);

                dataset = H5Dcreate(fs->file_id, datasetname, datatype, sid, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
                ret = H5Sclose(sid);
                assert(ret >= 0);
            }

            /* check if need to new dataset since the size of array type is too small */
            outdata = vlen_to_array_data(hsize_mem, (damsel_vector_t *)dcd->data, datatype);
            herr_t ret = H5Tclose(datatype);
            assert(ret >= 0);
            ret = H5Dclose(dataset);
            assert(ret >= 0);
        }
    }
	
    return outdata;
}			
			

