
// HDF5 specific stuff
// perhaps a goal should be for this to not be included into the damsel
// core code? (though I'll need to get the HDF5 initialiser declaration
// from elsewhere in that case then...)

#ifndef __DAMSEL_HDF5_H
#define __DAMSEL_HDF5_H

#ifdef DAMSEL_HDF5
#ifdef __cplusplus
extern "C" {
#endif

#include <hdf5.h>
#include <stdlib.h>

#include "damsel.h"
#include "damsel-internal.h"

#define DAMSEL_HDF5_MAX_TAG_SIZE 256

/* some HDF5 datasets are created with a maximum name of 256.  Is this an
 * internal limit or is it user-visible?  */
#define DAMSEL_DSET_NAME_MAX DAMSEL_HDF5_MAX_TAG_SIZE


/*!its dodgy the way that these are native types, but handles are a different
   type. if I want to move away from handles being always longs, then this
   must change */
typedef struct {
  /*! the handle for this handle container */
  long long int handle;

   /*! the base handle for handles in this container */
  long long int base;
  int quantity;
  int step;
} damsel_hdf5_infile_seq_handle_container;

typedef struct {
  long long int handle;
  int quantity;

/*! in the case that VECTOR_MAX is not enough to store the container,
      it is stored in several blocks. these are identified (within a
      specific container) by their blocknumber, starting at 0. so all
      containers have a block with block number 0. how will the code
      determine if there is a subsequent block?*/
  int blocknumber;

/*! length of contents depends on value of quantity... not sure how to do
    that in hdf5 types and I'm not online when writing this...
    so for now, I'll limit it to VECTOR_MAX maximum and store it inefficiently. */
  long long int content[VECTOR_MAX];

} damsel_hdf5_infile_vector_handle_container_row;

typedef struct {

  /*! still undecided what i'll use for datatypes in the very short term...
      I use sizeof elsewhere but that sucks. i'll use a numeric placeholder for
      now.
  */
  long tag_datatype;

  /*! this is a handle in the file's handle space */
  damsel_handle tag_handle;

  /*! ok, I'll limit this to a fixed size */
  char name[DAMSEL_HDF5_MAX_TAG_SIZE];
} damsel_hdf5_infile_tag_container;

/*! structure that holds info populated at read time by model_populate call */
struct populated_info {
  int seq_count;
  damsel_hdf5_infile_seq_handle_container *seq_info;

  /* this is the count of rows in the HDF5 table */
  int vec_rows;

  damsel_hdf5_infile_vector_handle_container_row *vec_info;

  int tag_count;
  damsel_hdf5_infile_tag_container *tag_info;
};

typedef struct damsel_attached_file {
  damsel_store_internal store;

  /*! the MPI communciator that will be used for this file. */
  MPI_Comm communicator;

  hid_t file_id;

  /*! In the present implementation, this will be set to a meaningful
      value only when opening an existing HDF5 damsel file. There's no
      reason why it should be restricted that way other than lazy
      implementation.
      If the dataset has not been opened, then it will be set to a
      negative value - valid hid_t values must be positive.

      This stores definitions of handle containers that are defined as
      sequences.
  */
  hid_t seq_handle_container_dataset;

  /*! this stores definitions of handle containers that are defined
      as vectors. */
  hid_t vector_handle_container_dataset;

  hid_t tag_container_dataset;
  // hid_t entity_definition_dataset;
  // hid_t connectivity_definition_dataset;

  hid_t entities_dataset;

  hid_t collection_dataset;

  int containers_loaded; // only meaningful in read mode; 0 == not loaded, 1 == loaded, other values invalid

  /*! only meaningful in read mode: NULL means not populated, otherwise points
     to a struct of the information loaded by populating. */
  struct populated_info *model_populated;

  int container_push_allowed; // only meaningful in write mode. 1 means that adding containers to the push list is allowed; 0 means that no more container pushing is allowed (for example, because the containers have been written out already)

  /* list of damsel_hdf5_internal_pushed_container - a list of application side containers that have been mirrored into the file, and their file-side handle. they may have ended up in the file by mapping, or they may have ended up there by pushing. either way, they end up in this structure. */
  linkedlist *first_pushed_container;

  /* contains damsel_remote_mapping */
  linkedlist *remote_mappings;

  /* the next handle that this rank is allowed to allocate, or 0 if the
     handle allocation mechanism has not been initialised yet. */
  damsel_handle handle_allocator;

  hid_t hdf5_collection_type;
  hid_t hdf5_tag_type;
  hid_t hdf5_sequence_type;
  hid_t hdf5_vector_type;
  hid_t hdf5_connectivity_type;
  hid_t hdf5_versioning_type;

  damsel_trait_bundle traits;

  int maxrank;
  int myrank;

} damsel_attached_file;

typedef struct {
  damsel_container app_container;
  damsel_container file_container;
  damsel_handle file_handle;
  int rank;
} damsel_remote_mapping;

typedef struct {
  damsel_container container; // on app side
  damsel_container file_container; // on file side
  damsel_handle handle; // on file side
  int remote; // 0 if this container originated on this rank, 1 if we received it from another rank and so should not put it onto disk because the source rank will do that

} damsel_hdf5_internal_pushed_container;

typedef struct {
  int type;
  damsel_handle collection_handle;
  damsel_handle container_handle;
} damsel_hdf5_infile_collection;

void DMSLhdf5_get_container_counts(damsel_attached_file *, int *seq_count, int *vec_rows);
void DMSLhdf5_get_container_seq_infos(damsel_attached_file *, damsel_hdf5_infile_seq_handle_container *col);
void DMSLhdf5_get_container_vec_infos(damsel_attached_file *, damsel_hdf5_infile_vector_handle_container_row *col);

void populate_hdf5_get_container_counts(damsel_attached_file *, int *seq_count, int *vec_rows);
void populate_hdf5_get_container_seq_infos(damsel_attached_file *, damsel_hdf5_infile_seq_handle_container *col);
void populate_hdf5_get_container_vec_infos(damsel_attached_file *, damsel_hdf5_infile_vector_handle_container_row *col);

void populate_hdf5_get_tag_count(damsel_attached_file *, int *);
void populate_hdf5_get_tag_tuples(damsel_attached_file *, damsel_hdf5_infile_tag_container *col);
void DMSLhdf5_get_tag_tuples(damsel_attached_file *, damsel_hdf5_infile_tag_container *col);

hid_t damsel_type_to_hdf5_type(damsel_data_type damsel_type);

damsel_container_internal *damsel_handle_to_handle_container(damsel_attached_file *s,
  damsel_handle h);

/** this should be called (perhaps internally to damsel rather than in
    the public API?) at file open time, when opening to read; it should
    restore the in-memory state of tags from the file. */
void internal_map_read_tags_to_mem(damsel_model);

void initialize_collection_type(damsel_attached_file *);
hid_t get_collection_type(damsel_attached_file *);

void initialize_connectivity_type(damsel_attached_file *);

/** records the a remote rank, processingrank, has a mapping from ac to fc,
    and that the file-side container has handle fh */
void internal_record_remote_mapping(damsel_attached_file *fs,
  damsel_container ac, damsel_container fc, damsel_handle fh,
  int processingrank);

/*! initializes the handle allocator in the attached file - before this is
    called, get_new_handles will not be able to allocated handles for the
    file. */
void initialize_handle_allocator(damsel_attached_file *file);

/*! returns the base of a contiguous sequence of handles, of length as
    specified - so that the caller may use base, base+1, ... base+quantity-1
*/
damsel_handle get_new_handles(struct damsel_attached_file *s, int quantity);



void damsel_close_hdf5_store(struct damsel_attached_file *);
void damsel_execute_memory_to_hdf5(damsel_model_internal *src_s, struct damsel_attached_file *s, MPI_Request *mpireq);
void damsel_execute_hdf5_to_memory(struct damsel_attached_file *src_s, damsel_model_internal *s, MPI_Request *mpireq);
damsel_err_t damsel_internal_hdf5_open(damsel_model_internal *mi, const char *fn, MPI_Comm comm, damsel_trait_bundle);

#define check_handle_is_handle_container(store, h) { assert(damsel_handle_to_handle_container(store, h) != NULL); }

#ifdef __cpluspluc
}
#endif
#endif
#endif
