
#include <assert.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include "damsel.h"
#include "damsel-internal.h"
#include "damsel-ndc.h"

FILE * NDC_output;

pthread_key_t ndc_pthread_key;

struct ndcframe {
  char *message;
  int freeme;
};

#ifdef DAMSEL_VERBOSE_SPEW

static void release_context_stack(void *m) {
  free(m);
}

/*! returns the thread-specific NDC stack */
static linkedlist **get_context_stack() {
  linkedlist **m = pthread_getspecific(ndc_pthread_key);
  if(m == NULL) {
    m = malloc(sizeof(linkedlist *));
    *m = NULL;
    assert(m != NULL);
    pthread_setspecific(ndc_pthread_key, m);
  }
  assert(m!=NULL);
  return m;
}

void init_ndc(void) {
  char *fname;
  fname = getenv("DAMSEL_LOG_OUTPUT");
  if (fname) NDC_output = fopen(fname, "w");
  if (NDC_output == NULL) NDC_output = stdout;

  pthread_key_create(&ndc_pthread_key, release_context_stack);
  push_ndc("Top level context");
}

/* returns a buffer adequately indented. caller must free this buffer.  Why do
 * it this way? mpich2 mpiexec with -prepend-rank will set the stdout buffer to
 * unbufferd, so each time a 'putchar' is done it gets a '[0]' prepended.    */

static char * indents_str() {
    /* mpich2 mpiexec with '-prepend-rank' will do this char-at-a-time and it
     * looks pretty bad */
  int indent;
  char *indent_buf;

  linkedlist **context = get_context_stack();

  indent = linkedlist_length(*context) - 1;
  indent_buf = malloc(indent+1);
  memset(indent_buf, ' ', indent);
  indent_buf[indent] = '\0';
  return (indent_buf);
}

void *push_ndc_internal(char *msg, int freeme) {


  struct ndcframe *newframe = malloc(sizeof(struct ndcframe));

  newframe->message=msg;
  newframe->freeme=freeme;

  linkedlist **ll = get_context_stack();
  linkedlist_prepend(ll, newframe);

  char *indent = indents_str();
  fprintf(NDC_output, "%s+%s\n", indent, msg);
  fflush(NDC_output);
  free (indent);
  return newframe;
}

void *push_ndc(char *msg) {
  return push_ndc_internal(msg,0);
}


void pop_ndc(void) {
  pop_ndc_context(NULL);
}


/* pops a frame. if ctx is supplied, it must be a pointed returned by the
 * corresponding push call, to allow debugging of not-match-up pushes and
 * pops.
 */
void pop_ndc_context(void *ctx) {
  linkedlist **llc = get_context_stack();
  assert(*llc != NULL);
  char *indent = indents_str();
  struct ndcframe *frame = (*llc)->head;
  fprintf(NDC_output, "%s-%s", indent, frame->message);
#ifndef DAMSEL_NO_VERBOSE_SPEW
  if(ctx == NULL) {
    fprintf(NDC_output, " (not checked)");
  }
#endif
  fprintf(NDC_output, "\n");
  fflush(NDC_output);
  free(indent);

#ifndef DAMSEL_NO_VERBOSE_SPEW
  if(ctx != NULL) {
    if(ctx != frame) {
      fprintf(NDC_output, "ndc stack mismatch: ");
      linkedlist *ll;
      int found=0;
      for(ll=(*llc)->next; ll!=NULL; ll=ll->next) {
        if(ll->head == ctx) found=1;
      }
      if(found == 1) {
        fprintf(NDC_output, "attempted to pop a context too soon - there are other contexts that must be popped first\n");
        fprintf(NDC_output, "attempted to pop context %s\n", 
		((struct ndcframe *)ctx)->message);
        dump_ndc();
      } else {
        fprintf(NDC_output, "attempted to pop a context that has already been popped\n");
      }
    }
  }
  assert(ctx == NULL || ctx == frame); // we do this assert after printing the message, so that we can see what the message was.
#endif
  if(frame->freeme == 1) free(frame->message);
  linkedlist *old = *llc;
  *llc = (*llc)->next;
  free(frame);
  free(old);
}

void dump_ndc(void) {
  int i=0;
  fprintf(NDC_output, "Nested diagnostic context:\n");
  linkedlist **llc = get_context_stack();
  linkedlist *here = *llc;
  while(here!=NULL) {
    struct ndcframe *frame = here->head;
    fprintf(NDC_output, " %d: %s\n", i, frame->message);
    i++;
    here=here->next;
  }
  fprintf(NDC_output, "End of NDC dump\n");
  fflush(NDC_output);
}

void finalize_ndc() {
    linkedlist **m = pthread_getspecific(ndc_pthread_key);
    if (m != NULL) {
	free(m);
    }
// TODO: delete the pthread key?
}

void *push_printf_ndc(const char *fmt, ...)
{
  va_list args;
  va_start(args, fmt);
  char *s = malloc(256); // memory leak
  vsnprintf(s, 255, fmt, args);
  void *r = push_ndc_internal(s,1);
  va_end(args);
  return r;
}

void printf_ndc(const char *fmt, ...)
{
  va_list args;
  va_start(args, fmt);
  char *indents = indents_str();
  fprintf(NDC_output, "%s ", indents);
  vfprintf(NDC_output, fmt, args);  // seems to not work?
  va_end(args);
  fprintf(NDC_output, "\n");
  free (indents);
  fflush(NDC_output);
}
#else
void init_ndc() {}
void dump_ndc() {}
void finalize_ndc() {}

#endif
