#include <assert.h>
#include <stddef.h>
#include <stdio.h>
#include <memory.h>

#include "damsel.h"
#include "damsel-internal.h"
#include "damsel-types.h"
#include "damsel-ndc.h"
#include "damsel-debug.h"

linkedlist *defined_types = NULL;

static damsel_data_type add_type(void *newtype) {
  linkedlist_append(&defined_types, newtype);
  return linkedlist_length(defined_types)
    + DAMSEL_DATA_TYPE_PREDEFINED_WATERMARK;
}

void init_types() {
// no type initialisation code now, but I'll keep this hook in for now.
}

void release_types() {
  linkedlist_free(defined_types, free);
  defined_types = NULL;
}

damsel_data_type DMSLtype_opaque_create(size_t bytecount) {
  damsel_i_type_opaque *o = NULL;

  assert_is_initialized();
  compile_assert(offsetof(damsel_i_type_opaque, magic) == 0);

  /* Allocate opaque datatype structure */
  o = calloc(1, sizeof(damsel_i_type_opaque));
  assert(o);
  printf_ndc("creating opaque type with byte count %zd with id %p", bytecount, o);

  /* Initialize opaque datatype info */
  o->magic = DAMSEL_I_TYPE_OPAQUE_MAGIC;
  o->bytecount = bytecount;

  /* Add to [internal] type dictionary */
  damsel_data_type num = add_type(o);

  return num;
}

damsel_data_type DMSLtype_array_create(damsel_data_type base_type,
                                        unsigned ndims,
                                        const size_t dims[])
{
  damsel_i_type_array *a = NULL;

  assert_is_initialized();
  compile_assert(offsetof(damsel_i_type_array, magic) == 0);
  assert(ndims <= DAMSEL_I_TYPE_ARRAY_MAX_NDIMS);

  /* Allocate array datatype structure */
  a = calloc(1, sizeof(damsel_i_type_array));
  assert(a);
  printf_ndc("creating array type with ndims %u with id %p", ndims, a);

  /* Initialize array datatype info */
  a->magic = DAMSEL_I_TYPE_ARRAY_MAGIC;
  a->base_type = base_type;             /* <-- This probably needs to reference count the base type */
  a->ndims = ndims;
  memcpy(a->dims, dims, ndims * sizeof(dims[0]));

  /* Add to [internal] type dictionary */
  damsel_data_type num = add_type(a);

  return num;
}

damsel_data_type DMSLtype_vector_create(damsel_data_type base, size_t max_len)
{
  damsel_i_type_vector *type = NULL;

  assert_is_initialized();
  compile_assert(offsetof(damsel_i_type_vector, magic) == 0);

  /* Allocate datatype structure */
  type = calloc(1, sizeof(damsel_i_type_vector));

  if (!type)
    return -1;
	  
  if (max_len<=0)
    max_len = DAMSEL_I_TYPE_MAX_VECTOR_SIZE;
	  
  /* Initialize array datatype info */
  type->magic = DAMSEL_I_TYPE_VECTOR_MAGIC;
  type->base_type = base; 
  type->max_len = max_len;

  /* Add to type dictionary */
  damsel_data_type type_id = add_type(type);

  return type_id;
}

damsel_data_type DMSLtype_enum_create(unsigned nmembs, char **names, int *values)
{
  int i;
  damsel_i_type_enum *type = NULL;

  assert_is_initialized();
  compile_assert(offsetof(damsel_i_type_enum, magic) == 0);

  /* make sure the no. of enum members is not out of bound */
  if (nmembs<=0 || nmembs>DAMSEL_I_TYPE_MAX_NMEMBS)
    return -1;
	  
  /* Allocate datatype structure */
  type = calloc(1, sizeof(damsel_i_type_enum));

  if (!type)
      return -1;
	  
  /* Initialize array datatype info */
  type->magic = DAMSEL_I_TYPE_ENUM_MAGIC;
  type->nmembs = nmembs; 
  

  for (i=0; i<nmembs; i++) {
	strncpy ( type->names[i], names[i], DAMSEL_I_TYPE_MAX_NAME_LEN-1);
	if (values)
		type->values[i] = values[i];
	else
		type->values[i] = (int)i; /* default values */
  }

  /* Add to type dictionary */
  damsel_data_type type_id = add_type(type);

  return type_id;
}

damsel_data_type DMSLtype_struct_create(unsigned nmembs, char **names, 
    damsel_data_type *types, size_t *offsets, size_t tsize)
{
  int i;
  damsel_i_type_struct *type = NULL;

  assert_is_initialized();
  compile_assert(offsetof(damsel_i_type_struct, magic) == 0);

  /* make sure the no. of enum members is not out of bound */
  if (nmembs<=0 || nmembs>DAMSEL_I_TYPE_MAX_NMEMBS)
    return -1;
	  
  /* Allocate datatype structure */
  type = calloc(1, sizeof(damsel_i_type_struct));

  if (!type)
      return -1;
	  
  /* Initialize array datatype info */
  type->magic = DAMSEL_I_TYPE_STRUCT_MAGIC;
  type->nmembs = nmembs; 
  
  for (i=0; i<nmembs; i++) {
	strncpy ( type->names[i], names[i], DAMSEL_I_TYPE_MAX_NAME_LEN-1);
	type->types[i] = types[i];
	type->offsets[i] = offsets[i];
	type->tsize = tsize;
  }

  /* Add to type dictionary */
  damsel_data_type type_id = add_type(type);

  return type_id;
}

