#include <errno.h>
#include <mpi.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <assert.h>

#include "damsel-hdf5.h"
#include "damsel.h"
#include "damsel-internal.h"
#include "damsel-ndc.h"
#include "damsel-debug.h"
#include "damsel-connectivity.h"
#include "damsel-types.h"

#ifndef DAMSEL_SINGLETHREAD
#include <pthread.h>
#endif
#ifdef DAMSEL_SUBFILE
#include "damsel-subfile.h"
#endif

#define FALSE (1==0)
#define TRUE (1==1)

int globals_initialized = 0;

int test_no_tag_with_this_handle(damsel_model model, damsel_handle h) {
  assert_valid_model_id(model);
  linkedlist *ll = ((damsel_model_internal *)model)->first_tag_definition;
  while(ll != NULL) {
    damsel_tag_internal *tag_definition = (damsel_tag_internal *)ll->head;
    if(tag_definition->handle == h) return (1==0);
    ll=ll->next;
  }
  return (1==1);
}

/*! returns the base of a contiguous sequence of handles, of length as
    specified - so that the caller may use base, base+1, ... base+quantity-1
*/
damsel_handle get_new_handles(damsel_attached_file *file, int quantity) {
  push_ndc("get_new_handles");
  assert(file != NULL);
  assert(file->handle_allocator != 0);
  damsel_handle base;
  base = file->handle_allocator;
  file->handle_allocator+=quantity;
  printf_ndc("get_new_handles allocated %d handles beginning at %llx for attached file %p", quantity, base, file);

  pop_ndc();
  return base;
}

void initialize_handle_allocator(damsel_attached_file *file) {
  assert(file->handle_allocator == 0);
  damsel_handle blockid;

  int myrank, maxrank;
  myrank = file->myrank;
  maxrank = file->maxrank;


  hid_t datatype = H5T_NATIVE_INT64;
  assert(datatype > 0);
 
  hsize_t one = 1;
  hid_t dataspace = H5Screate_simple(1, &one, &one);  // single element

  assert(dataspace > 0);
  hid_t dataset = -1;
  if(H5Lexists(file->file_id, DAMSEL_DATASET_HANDLE_ALLOCATOR, H5P_DEFAULT) <= 0) {
    dataset = H5Dcreate(file->file_id, DAMSEL_DATASET_HANDLE_ALLOCATOR, datatype, dataspace, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
    assert(dataset > 0);
    blockid = 0;
  } else { // handles previously allocated.
    dataset  = H5Dopen(file->file_id, DAMSEL_DATASET_HANDLE_ALLOCATOR, H5P_DEFAULT);
    assert(dataset >= 0);
    blockid=0;
    int ret = H5Dread(dataset, datatype, dataspace, dataspace, H5P_DEFAULT, &blockid);
    assert(ret >= 0);
  }

  assert(dataset >= 0);
  printf_ndc("read blockid %llx", blockid);
  // at this point, blockid will be the first blockid we can use: either 0,
  // if none recorded in the file, or the value read from the file.

  blockid += maxrank;

  if(myrank == 0) {
    int res = H5Dwrite(dataset, H5T_NATIVE_INT64, dataspace, dataspace, H5P_DEFAULT, &blockid);
    assert(res >= 0);
  }
  printf_ndc("new blockid written to file: %llx", blockid);

  H5Dclose(dataset);

  blockid += myrank;
  file->handle_allocator = (((damsel_handle)blockid) << 32) + 5000;
}

int handle_container_unique(damsel_container id) {
  push_printf_ndc("handle_container_unique, container = %p", id);
  assert_valid_handle_container_id(id);

  damsel_container_internal *hc = (damsel_container_internal *)id;
  assert(hc != NULL);
  int li;
  for(li=0; li<hc->quantity; li++) {
    push_printf_ndc("checking LHS position %d", li);
    damsel_handle lh = DMSLcontainer_handle_at_position(id, li);
    int ri;
    for(ri=0; ri < li ; ri++) {
      push_printf_ndc("checking RHS position %d", ri);
      damsel_handle rh = DMSLcontainer_handle_at_position(id, ri);
      if(lh == rh) {
        printf_ndc("found duplicate handle %llx at positions %d and %d", lh, li, ri);
        pop_ndc();
        return 1;
      }
      pop_ndc();
    }
    pop_ndc();
  }
  printf_ndc("handle container contains no duplicates.");
  pop_ndc();
  return 0;
}



static void free_tag(void *m) {
  damsel_tag_internal *t = (damsel_tag_internal *) m;
  free(t->name);
  t->name = NULL;
  free(t);
}

static void free_tags(damsel_model_internal *mi) {
  linkedlist_free(mi->first_tag_definition, free_tag);
  mi->first_tag_definition = NULL;
}

static void free_dcd(damsel_cached_data_internal *dcd) {
  if(dcd == NULL) return;
  free_dcd(dcd->next);

  DMSLcontainer_release(dcd->handles_id);
  free(dcd);
}

static void free_mapping(void *m) {
  damsel_store_mapping *mi = (damsel_store_mapping *)m;
  damsel_container_internal *ac = (damsel_container_internal *) mi->app_container_id;
  damsel_container_internal *fc = (damsel_container_internal *) mi->file_container_id;
  DMSLcontainer_release((damsel_container)ac);
  DMSLcontainer_release((damsel_container)fc);
  free(m);
}

static void free_global_mapping(void *m) {
  damsel_internal_global_mapping *gm = (damsel_internal_global_mapping *)m;
  assert(gm->magic == DAMSEL_GLOBAL_MAPPING_MAGIC);

  if(gm->manage_memory_internally) {
    free(gm->proc_ranks);
    free(gm->remote_handles);
  }
  DMSLcontainer_release(gm->app_container);
  free(m);
}

static void free_global_mappings(damsel_model_internal *mi) {
  linkedlist_free(mi->global_mappings, free_global_mapping);
  mi->global_mappings = NULL;
}

static void free_mappings(damsel_model_internal *mi) {
  linkedlist_free(mi->mappings, free_mapping);
  mi->mappings = NULL;
}

static void free_pushed_container(void *m) {
  damsel_hdf5_internal_pushed_container *mi = (damsel_hdf5_internal_pushed_container *)m;
  if(mi == NULL) return;
  if(mi->container) DMSLcontainer_release(mi->container);
  DMSLcontainer_release(mi->file_container);
  free(m);
}

static void free_pushed_containers(damsel_model_internal *mi) {
  if(mi == NULL) return;
  if(mi->attached_file == NULL) return;
  linkedlist_free(mi->attached_file->first_pushed_container, free_pushed_container);
  mi->attached_file->first_pushed_container = NULL;
#ifdef DAMSEL_SUBFILE
  if(mi->num_subfiles > 1) {
    if(mi->attached_subfile == NULL) return;
    linkedlist_free(mi->attached_subfile->first_pushed_container, free_pushed_container);
    mi->attached_subfile->first_pushed_container = NULL;
  }
#endif
}

static void free_collection(void *m) {
  damsel_collection_internal *coll = (damsel_collection_internal *)m;
  DMSLcontainer_release(coll->container);
  free(m);
}

static void free_collections(damsel_model_internal *mi) {
  if(mi == NULL) return;
  linkedlist *ll = mi->first_collection;
  linkedlist_free(ll, free_collection);
  mi->first_collection = NULL;
}

static void free_entity_definition(void *m) {
  damsel_internal_entity_definition *ed = (damsel_internal_entity_definition *)m;
  DMSLcontainer_release(ed->vertex_container_id);
  DMSLcontainer_release(ed->entity_container_id);
  free(m);
}

static void free_entities(damsel_model_internal *mi) {
  if(mi == NULL) return;
  linkedlist *ll = mi->first_entity_definition;
  linkedlist_free(ll, free_entity_definition);
  mi->first_entity_definition = NULL;
}

static void free_chunk(void *m) {
  assert(m!=NULL);
  damsel_internal_chunk *c = (damsel_internal_chunk *)m;
  assert(c->magic == DAMSEL_CHUNK_MAGIC);
  DMSLcontainer_release((damsel_container) c->container);
  free(m);
}

static void free_chunks(damsel_model_internal *mi) {
  if(mi == NULL) return;
  linkedlist *ll = mi->first_chunk;
  linkedlist_free(ll, free_chunk);
  mi->first_chunk = NULL;
}

void model_take(damsel_store_internal *s) {
  assert_valid_store_id(s);
  s->refcount++;
  printf_ndc("model_take: taking reference on model/store %p, refcount is now %d", s, s->refcount);
}

void model_release(damsel_store_internal *s) {
  assert_valid_store_id(s);
  push_printf_ndc("model_release: model/store %p", s);
  assert(s->refcount > 0);
  s->refcount--;
  if(s->refcount == 0) {
    printf_ndc("model no longer used - freeing");
    free(s);
  } else {
    printf_ndc("model still in use - not freeing. There are %d remaining references", s->refcount);
  }
  pop_ndc();
}

damsel_err_t DMSLmodel_close(damsel_model m) {
  push_ndc("DMSLmodel_close");
  damsel_model_internal *mi = (damsel_model_internal*) m;
  free_pushed_containers(mi);
  if(mi->attached_file) damsel_close_hdf5_store(mi->attached_file);
#ifdef DAMSEL_SUBFILE
  if(mi->attached_subfile) 
    damsel_close_hdf5_store(mi->attached_subfile);
#endif
  free_tags(mi);
  free_mappings(mi);
  free_global_mappings(mi);
  free_collections(mi);
  free_entities(mi);
  free_chunks(mi);
  model_release(&(mi->store));
  pop_ndc();
  return DMSL_OK;
}

/*! this does not map the contained handles into the file, though
    it does map the specified collection handle into the file.
    the contained handles need to be mapped in elsewhere by the
    application.
    (perhaps a 'push' rather than a 'map' would be relevant here if
    I implement something like that - I need to ensure that the handles
    in the supplied container all appear in the file, but they don't
    necessarily appear mapped one-to-one to this container...)
 */
damsel_collection DMSLcoll_create(damsel_model model_id,
                    damsel_handle_ptr app_handle_ptr,
                    damsel_container container_id,
                    damsel_collection_type coll_type) {

  DMSLcontainer_take(container_id);

  damsel_model_internal *m = (damsel_model_internal *)model_id;

  damsel_collection_internal *ci = malloc(sizeof(damsel_collection_internal));
  assert(ci != NULL);

  ci->collection_handle = *app_handle_ptr;
  ci->container = container_id;
  ci->type = coll_type;

  linkedlist_prepend(&m->first_collection, ci);

  return (damsel_collection)ci;
}

damsel_err_t check_tag_does_not_exist(damsel_model m, const char *name) {
  linkedlist *ll = ((damsel_model_internal *)m)->first_tag_definition;
  while(ll!=NULL) {
    damsel_tag_internal *t = (damsel_tag_internal *)ll->head;
    if(!strcmp(t->name, name)) {
      printf_ndc("tag already exists with name %s", name);
      return DMSL_ERROR_GENERAL;
    }
    ll=ll->next;
  }
  return DMSL_OK;
}

/* does this function have handle registering behaviour too? or is this
   a "different kind" of handle, that cannot be annotated by tags? */

damsel_err_t DMSLtag_define_internal(damsel_model model_id, damsel_handle_ptr tag_handle_ptr, damsel_data_type type, const char *name, int written) {
  assert_is_initialized();

  push_ndc("DMSLtag_define");
  assert(model_id!=NULL);
  assert(tag_handle_ptr != NULL);
  assert(name!=NULL);

  damsel_err_t err = check_tag_does_not_exist(model_id, name);
  if(err.id != DMSL_OK.id) {
    pop_ndc();
    return DMSL_ERROR_GENERAL;
  }

  damsel_model_internal *model = (damsel_model_internal *)model_id;

  damsel_handle taghandle = *tag_handle_ptr;

  printf_ndc("defining tag '%s' with element type id %llx, tag handle %llx", name, type, taghandle);

  damsel_tag_internal *tagdef = malloc(sizeof(damsel_tag_internal));
  assert(tagdef != NULL);
  tagdef->magic = DAMSEL_TAG_MAGIC;
  tagdef->written = written;

  tagdef->name = malloc(strlen(name)+1);
  assert(tagdef->name != NULL);
  strcpy(tagdef->name, name);

  assert_is_damsel_type(type);
  tagdef->type = type;
  tagdef->handle = taghandle;

  assert_valid_tag_definition_structure(tagdef);

  linkedlist_prepend(&model->first_tag_definition, tagdef);
  printf_ndc("damsel tag id = %p", tagdef);
  pop_ndc();
  return DMSL_OK;
}

damsel_err_t DMSLtag_define(damsel_model model_id, damsel_handle_ptr tag_handle_ptr, damsel_data_type type, const char *name) {
  printf_ndc("DMSLtag_define, name = %p %s", name, name);
  return DMSLtag_define_internal(model_id, tag_handle_ptr, type, name, DAMSEL_UNWRITTEN);
}

// the in-memory version is going to cache the location of data but not
// copy it.
// so I need to store the parameters of this call for later use, but
// don't need to copy any other data around.

// i have an open enquiry with quincey and tim about what happens when
// i call this with a file-based store. for now, we fail.

// void DMSLtag_set(damsel_store *store, damsel_handle collh, damsel_handle tag_handle, void *data) {
// Why is tag_handle not a tag ID in the form of damsel_id_t?
// Now I'm going to need to find out the model / memory store
// from one of these data structures,
damsel_err_t
DMSLmodel_map_tag(void * const      data,
                  damsel_container  app_handles_container,
                  damsel_handle_ptr tag_handle_ptr)
{
  void *ctx = push_printf_ndc("DMSLmodel_map_tag, data=%p, app_handles_container=%p, tag_handle_ptr=%p", data, app_handles_container, tag_handle_ptr);

  assert(data != NULL);

  damsel_container_internal *hc = (damsel_container_internal *)app_handles_container;
  printf_ndc("handle container is %p", hc);
  assert(hc != NULL);
  assert_valid_handle_container_id(hc);
  assert(hc->model != NULL);

  assert(tag_handle_ptr != NULL);
  damsel_handle tag_handle = *((damsel_handle *) tag_handle_ptr);
  printf_ndc("tag_handle = %llx", tag_handle);

  damsel_store_internal *store = (damsel_store_internal *)hc->model;
  damsel_model_internal *model_internal = (damsel_model_internal *)hc->model;
  assert(store!=NULL);

  assert(store->type == DAMSEL_STORETYPE_MEMORY);
  damsel_model_internal *model = (damsel_model_internal *)store;

  assert_valid_tag_handle(tag_handle, model_internal);

  DMSLcontainer_take(app_handles_container);

  damsel_cached_data_internal *dc = malloc(sizeof(damsel_cached_data_internal));
  assert(dc != NULL); /* if (dc == NULL) return DMSL_EMALLOC; */

  dc->handles_id = (damsel_container) hc;

  printf_ndc("dc->handles_id = %p", dc->handles_id);
  assert_valid_handle_container_id(dc->handles_id);

  dc->tag_handle = tag_handle;
  dc->data = data;

  dc->next = model->first_cached_data;
  model->first_cached_data = dc;

  pop_ndc_context(ctx);
  return DMSL_OK;
}


damsel_err_t DMSLmodel_map_handles_inventing_file_handles_with_permute(damsel_container app_container, void *permutation_callback) {
  push_printf_ndc("mapping handle container with damsel-decided file handles");
  printf_ndc("container id = %p", app_container);

  damsel_container_internal *hc = ((damsel_container_internal *) app_container);
  assert_valid_handle_container_id(hc);

  damsel_model_internal *model = (damsel_model_internal *) hc->model;
  assert_valid_model_id(model);

  DMSLcontainer_take(app_container);

  damsel_metadata_handle *metadata_handle = malloc(sizeof(damsel_metadata_handle));
  metadata_handle->container_id = (damsel_container) hc;
  metadata_handle->permutation_callback = permutation_callback;
  metadata_handle->next = model->first_metadata_handle;
  model->first_metadata_handle = metadata_handle;

  printf_ndc("new allocated metadata handle structure is %p", metadata_handle);

  pop_ndc();
  return DMSL_OK;
}

damsel_err_t DMSLmodel_map_handles_inventing_file_handles(damsel_container app_container) {
  return DMSLmodel_map_handles_inventing_file_handles_with_permute(app_container, damsel_permute_id);
}

/* This is used in DMSLmodel_map_handles.

   The intention is to detect (and abort) if a mapping is being created that
   would write values onto the same handles as have already been mapped.

   Its unclear to benc if this is too restrictive: perhaps it is possible for
   one tag's values on a handle to be set through one mapping; and another
   tag's values on that same handle to be set through another mapping?

   The supplied container must be a file-side handle container.

 */
static void check_mapping_unique(damsel_model_internal *model,
  damsel_container container) {
  push_printf_ndc("check_mapping_unique, model=%p, file container=%p", model,
    container);
  push_ndc("checking handle contains no duplicate handles");

  assert(!handle_container_unique(container));

  pop_ndc();

  push_ndc("checking for overlaps with other mapped file-side containers in this model");

  linkedlist *ll = model->mappings;

  while( ll != NULL) {
    damsel_store_mapping *mapping = (damsel_store_mapping *) ll->head;
    printf_ndc("checking mapping %p", mapping);
    assert(!handle_containers_overlap(mapping->file_container_id,container));
    ll=ll->next;
  }
  pop_ndc();
  pop_ndc();
}

damsel_err_t DMSLmodel_map_single_handle_inventing_file_handle(damsel_model model, damsel_handle h) {
  damsel_err_t rc;
  push_ndc("mapping single handle.");
  damsel_container container_id;
  rc = DMSLcontainer_create_sequence(model, h, 1, 1, &container_id);
  printf_ndc("created container is %p", container_id);
  rc = DMSLmodel_map_handles_inventing_file_handles(container_id);
  DMSLcontainer_release(container_id);
#ifdef DAMSEL_SUBFILE
  damsel_model_internal *mi = (damsel_model_internal *)model;
#if 0
  if(mi->num_subfiles > 1) {
    damsel_container subfile_container_id;
    rc = DMSLcontainer_create_sequence(model, h, 1, 1, &subfile_container_id);
    printf_ndc("created subfile container is %p", subfile_container_id);
    rc = DMSLmodel_map_handles_inventing_file_handles(subfile_container_id);
    DMSLcontainer_release(subfile_container_id);
  }
#endif
#endif
  pop_ndc();
  return rc;
}


damsel_err_t DMSLmodel_map_handles(damsel_container app_container,
                                   damsel_container file_container) {
  damsel_err_t err = model_map_handles_internal(app_container, file_container, 0);

  if(err.id == DMSL_OK.id) {
  /* in read mode, this might have caused one or more tags to be
     mapped from the file into app handle space. */ 
    damsel_model m = ((damsel_container_internal *)app_container)->model;
    damsel_model_internal *mi = (damsel_model_internal *)m;
    if(mi->attached_file->model_populated != NULL) {
      DMSLmap_resolve_read_tags_internal(m);
    }
  }
  return err;
}

damsel_err_t model_map_handles_internal(damsel_container app_container,
  damsel_container file_container, int late_additions_allowed) {
  push_printf_ndc("DMSLmodel_map_handles, app_container = %p, file_container = %p", app_container, file_container);

  assert_valid_handle_container_id(app_container);
  assert_valid_handle_container_id(file_container);
  damsel_container_internal *a = (damsel_container_internal *)app_container;
  damsel_container_internal *b = (damsel_container_internal *)file_container;
  assert( ((damsel_store_internal *) a->model)->type == DAMSEL_STORETYPE_MEMORY);
  assert( ((damsel_store_internal *) b->model)->type == DAMSEL_STORETYPE_HDF5);

  DMSLcontainer_take(app_container);
  DMSLcontainer_take(file_container);

  damsel_model_internal *model = (damsel_model_internal *)a->model;

  if(late_additions_allowed == 0) assert(model->more_mappings_allowed == 1);

  if(a->quantity != b->quantity) {
    printf_ndc("model_map_handles_internal was given containers of different sizes, but they must be the same size. app_container->quantity = %d, file_container->quantity = %d", a->quantity, b->quantity);
    return DMSL_ERROR_GENERAL;
  }

  assert(a->quantity == b->quantity);

/* the file container needs to not overlap with any other mapped file
   container (though it can overlap with other file containers that are
   not used in mapping)
*/
  check_mapping_unique(model, file_container);

  damsel_store_mapping *mapping = malloc(sizeof(damsel_store_mapping));
  printf_ndc("mapping struct memory address %p", mapping);
  assert(mapping != NULL);
  assert(a != NULL);
  assert(((damsel_store_internal *)a->model)->type == DAMSEL_STORETYPE_MEMORY || ((damsel_store_internal *)a->model)->type == DAMSEL_STORETYPE_HDF5);

  mapping->app_container_id = app_container;
  mapping->file_container_id = file_container;
  mapping->magic = DAMSEL_STORE_MAPPING_MAGIC;

  linkedlist_prepend(&((damsel_model_internal *)a->model)->mappings, mapping);
  pop_ndc();
  return DMSL_OK;
}


/* inner code to do the writes for a file store */
typedef struct {
  damsel_model_internal *model;
  damsel_attached_file *file_backend;
  damsel_transfer_type direction;
  MPI_Request *request;
} damsel_execute_args;

damsel_execute_args * damsel_execute_args_create(damsel_model_internal *model,
  MPI_Request *request, damsel_transfer_type direction, damsel_attached_file *file_backend) {

  damsel_execute_args * args;
  args = (damsel_execute_args *) malloc(sizeof(damsel_execute_args));

  args->model = model;
  args->request = request;
  args->direction = direction;
  args->file_backend = file_backend;

  return (args);
}

static void damsel_execute_args_destroy(damsel_execute_args *args) {
  free(args);
}


static int damsel_query_fn(void *extra_state, MPI_Status *status)
{
    /* can never cancel */
    MPI_Status_set_cancelled(status, 0);
    /* choose not to return a value for this */
    status->MPI_SOURCE = MPI_UNDEFINED;
    /* tag has no meaning for this generalized request */
    status->MPI_TAG = MPI_UNDEFINED;
    /* this generalized request never fails */
    return MPI_SUCCESS;

}

static int damsel_free_fn(void *extra_state)
{
   damsel_execute_args *args = (damsel_execute_args*)extra_state;
   damsel_execute_args_destroy(args);

   return MPI_SUCCESS;

}

static int damsel_cancel_fn(void *extra_state, int complete) {
	/* cannot cancel */
	return MPI_SUCCESS;
}


int DMSLcontainer_count(damsel_container c) {
  if (c)
    return ((damsel_container_internal *)c)->quantity;
  else 
    return 0;
}

int DMSLcontainer_get_type(damsel_container c) {
  return ((damsel_container_internal *)c)->type;
}

/*! returns the i-th handle in the specified container. undefined behaviour
    if there are not at least i entries in the store. the container is
    specified as a damsel_container_internal struct. */
damsel_handle DMSLcontainer_handle_at_position(damsel_container c,
  int i) {
  damsel_container_internal *hc = (damsel_container_internal *)c;
  printf_ndc("DMSLcontainer_handle_at_position, container id %p, offset %d", hc, i);
  assert(i<hc->quantity);
  if(hc->type == DAMSEL_HANDLE_CONTAINER_TYPE_SEQUENCE) {
    damsel_seq_handle_container *hsc = (damsel_seq_handle_container *)hc;
    return hsc->base + hsc->step * i;
  } else if (hc->type == DAMSEL_HANDLE_CONTAINER_TYPE_VECTOR) {
    damsel_vector_handle_container *hvc =
      (damsel_vector_handle_container *)hc;
    return hvc->contents[i];
  } else if(hc->type == DAMSEL_HANDLE_CONTAINER_TYPE_TREE) {

    linkedlist *here = ((damsel_tree_handle_container *)hc)->first_hc;
    while(here!=NULL) {
      damsel_container_internal *subhc = here->head;
      assert_valid_handle_container_id(subhc);
      if(i < subhc->quantity) { // then it is in this subcontainer
        return DMSLcontainer_handle_at_position((damsel_container)subhc, i);
      }
      // otherwise move along one...
      i -= subhc->quantity;
      here = here->next;
    }
    assert(1==0 && "reached end of linked list unexpectedly");
  } else {
    assert(1==0 && "undefined handle container type encountered");
    return 0;
  }
  return DAMSEL_HANDLE_TYPE_INVALID;
}

int get_offset_in_container(damsel_container_internal *hc, damsel_handle src_handle);

damsel_err_t handle_map_rank_specific(damsel_model_internal *model, damsel_transfer_type direction,
  damsel_handle src_handle, damsel_handle *dest_handle, int rank, int recursion_allowed) {
  void *ctx = push_printf_ndc("handle_map_rank_specific: model %p direction %d mapping handle %llx", model, direction, src_handle);

  void *ctxrs = push_ndc("checking rank-specific mappings");

  damsel_attached_file *fs = model->attached_file;

  int myrank, maxrank;
  assert(model != NULL);
  assert(model->attached_file != NULL);
  myrank = model->attached_file->myrank;
  maxrank = model->attached_file->maxrank;


// TODO: this iteration needs to happen over a different set of mappings if we're not checking against the local rank
// it needs to check the subset of the other mappings list (?) that apply to the chosen rank.

// can i structure this in a way that looks nicer than two huge code blocks that are almost identical?

  if(myrank == rank) {
    printf_ndc("taking myrank path");

    linkedlist *ml = model->mappings;

    while(ml != NULL) {
      damsel_store_mapping *mapping = (damsel_store_mapping *) ml->head;

      void *ctxm = push_printf_ndc("checking mapping app_container_id=%p file_container_id=%p", mapping->app_container_id, mapping->file_container_id);

      if(direction == DAMSEL_TRANSFER_TYPE_WRITE) {
        damsel_container_internal *src_container =
          (damsel_container_internal *) mapping->app_container_id;
        int i = get_offset_in_container(src_container, src_handle);
        if(i>=0) { // found!
          printf_ndc("source handle is in container %p at offset %d", src_container, i);
          damsel_handle hh = DMSLcontainer_handle_at_position(mapping->file_container_id, i);
          printf_ndc("found handle 0x%llx - successfully mapped %llx to %llx", hh, src_handle, hh);
          *dest_handle = hh;
          pop_ndc_context(ctxm);
          pop_ndc_context(ctxrs);
          pop_ndc_context(ctx);
          return DMSL_OK;
        }
      } else {
        assert(direction == DAMSEL_TRANSFER_TYPE_READ);
        damsel_container_internal *src_container =
          (damsel_container_internal *) mapping->file_container_id;
        int i = get_offset_in_container(src_container, src_handle);
        if(i>=0) { // found!
          printf_ndc("source handle is in container %p at offset %d", src_container, i);
          damsel_handle hh = DMSLcontainer_handle_at_position(mapping->app_container_id, i);
          printf_ndc("found handle 0x%llx - successfully mapped %llx to %llx", hh, src_handle, hh);
          *dest_handle = hh;
          pop_ndc_context(ctxm);
          pop_ndc_context(ctxrs);
          pop_ndc_context(ctx);
          return DMSL_OK;
        }
      }

      ml=ml->next;
      pop_ndc_context(ctxm);
    }
  } else {
    push_ndc("taking not-my-rank path");
// TODO we're looking at the local mappings for a different rank.
// so do the same as above but over a different set of structs.
    assert(direction == DAMSEL_TRANSFER_TYPE_WRITE); // none of this on read until i implement it
    linkedlist *ml = fs->remote_mappings;
    while(ml != NULL) {
      printf_ndc("checking remote mapping %p", ml);
      damsel_remote_mapping *mapping = (damsel_remote_mapping *) ml->head;
      damsel_container_internal *src_container =
        (damsel_container_internal *) mapping->app_container;
        int i = get_offset_in_container(src_container, src_handle);
        if(i>=0 && mapping->rank == rank) {
          damsel_handle hh = DMSLcontainer_handle_at_position(mapping->file_container, i);
          *dest_handle = hh;
          pop_ndc();
          pop_ndc_context(ctxrs);
          pop_ndc_context(ctx);
          return DMSL_OK;
        }
      ml = ml->next;
    }
    pop_ndc();
  }

  printf_ndc("did not find any mapping locally");
  pop_ndc_context(ctxrs);
  if(recursion_allowed == 1) {
    push_ndc("checking global mappings");

  // so i start with a handle on this rank, and I want to end up
  //   with a file rank
  // the above has tried a direct route.
  // what i can instead try is:
  //  is this handle mapped to a local handle on another rank?
  //  (which can happen in two different directions in a global 
  //   mapping structure)

  linkedlist *gml = model->global_mappings;

  printf_ndc("there are %d global mappings", linkedlist_length(gml));

  while(gml!=NULL) {
    push_ndc("processing a global mapping");


    damsel_internal_global_mapping *mapping = (damsel_internal_global_mapping *) gml->head;
  //     TODO: check from app_container -> farside
  //           this one we only need to check mappings that are from
  //           the "current rank" (in plain handle_map this rank, but
  //           in the richer version, whatever the current rank is)
    damsel_container c = mapping->app_container;
    int csize = DMSLcontainer_count(c);

#ifndef NDEBUG
    push_ndc("dump of this mapping");
    printf_ndc("mapping rank %d", mapping->rank);
    assert_valid_rank(mapping->rank, maxrank);
    printf_ndc("count %d", csize);
    int dumpix;
    for(dumpix=0; dumpix<csize; dumpix++) {
      printf_ndc("position %d, local app space handle %llx   remote app space handle %llx  remote rank %d", dumpix, DMSLcontainer_handle_at_position(c, dumpix),
        mapping->remote_handles[dumpix], mapping->proc_ranks[dumpix]);
      assert_valid_rank(mapping->proc_ranks[dumpix], maxrank);
    }
    pop_ndc();
#endif

    if(mapping->rank == rank ){
      // scan over the app container in this mapping to see if it matches up
      // with the source handle

      int cix = get_offset_in_container((damsel_container_internal *)c, src_handle);
// BUG: might be in the source container multiple times, so this needs to be
// an iteration over all of those possible values.

      if(cix != -1) {
        printf_ndc("offset in container, cix = %d", cix);
        assert(cix>=0 && cix < csize);
        damsel_handle hh = mapping->remote_handles[cix]; // new source handle
        printf_ndc("new source handle hh=%llx", hh);
        int hrank = mapping->proc_ranks[cix];
        assert_valid_rank(hrank, maxrank);
        printf_ndc("looking for mapping on hrank = %d", hrank);
        // so now we have an equivalent handle on a different rank.
        // so I can run the same handle lookup call with that different rank.
        damsel_err_t ret;
        ret = handle_map_rank_specific(model, direction, hh, dest_handle, hrank, 0);
        if(ret.id == 0) {
          pop_ndc();
          pop_ndc();
          pop_ndc_context(ctx);
          return DMSL_OK; // found it
        }
      }
    }

    printf_ndc("not found in forward remote map; checking reverse remote map");

    int rix;
    for(rix = 0; rix<csize; rix++) {
      assert_valid_rank(rank, maxrank);
      assert_valid_rank(mapping->proc_ranks[rix],maxrank);
      if(mapping->proc_ranks[rix] == rank) {
        // then check if the far end handle matches, in which case, we've found a handle on the source rank.
        if(mapping->remote_handles[rix] == src_handle) {
          // found it...
          damsel_handle hh = DMSLcontainer_handle_at_position(c, rix);
          damsel_err_t ret;
          ret = handle_map_rank_specific(model, direction, hh, dest_handle, mapping->rank, 0);
          if(ret.id == 0) {
            pop_ndc();
            pop_ndc();
            pop_ndc_context(ctx);
            return DMSL_OK; // found it
          }
        }
      }
    }
    printf_ndc("not found in either forward or reverse remote map");
    gml=gml->next;
    pop_ndc();
  }
  //     now if we reach this I have a handle on that rank.

  //     apply same mechanism recursively:
  //        that handle might resolve 'locally' to a file handle,
  //     or that handle might resolve again in the global handle table
  //        to a handle on another rank. and we go round again, with the current rank parameter being that other rank - we can pretty much
  //             forget that we're running on the current rank at all...

  // so I perhaps want a version of handle_map that can use the other
  //   form of mapping table? the one that comes remotely and I suppose
  //   has an extra "rank-of-tables-to-check" parameter (equivalently,
  //   it identities the namespace of the source handle)...
  // and can then make handle_map the local version of that, that sticks the
  //   rank in...

// TODO: a test case that checks multirank global handle mapping in the *read* case, not just the write case.

  pop_ndc();
  }

  pop_ndc_context(ctx);
  return DMSL_ERROR_GENERAL;
}

/*! handle_map is a helper that, given source and destination stores,
    and a handle in that source store, will give the corresponding handle
    in the destination store, returning 0, or otherwise will return non-zero.

    it will use indirection via other rank global mappings (if they have been
    synchronised between ranks) to find the result.
*/

damsel_err_t handle_map(damsel_model_internal *model, damsel_transfer_type direction,
  damsel_handle src_handle, damsel_handle *dest_handle) {
  void *ctx = push_ndc("handle_map");
  int rank;
  assert(model != NULL);
  assert(model->attached_file != NULL);
  rank = model->attached_file->myrank;
  damsel_err_t ret;
  ret = handle_map_rank_specific(model, direction, src_handle,
    dest_handle, rank, 1);
  printf_ndc("handle_map returning %d", ret.id);
  pop_ndc_context(ctx);
  return ret;
}


static void *damsel_transfer_thread(void *ptr) {

  damsel_execute_args *args;
  args = (damsel_execute_args*)ptr;

  /* generalized request completed inside these funcs */
  switch(args->direction) {
    case DAMSEL_TRANSFER_TYPE_WRITE:
      damsel_execute_memory_to_hdf5(args->model, args->file_backend,args->request);
      break;
    case DAMSEL_TRANSFER_TYPE_READ:
      damsel_execute_hdf5_to_memory(args->file_backend, args->model, args->request);
      break;
    default:
      printf("invalid enum value in damsel_transfer_thread\n");
      assert(1==0 && "invalid enum value");
      break;
  }

  free_dcd(args->model->first_cached_data);
  args->model->first_cached_data = NULL;

  MPI_Grequest_complete(*(args->request));

  return NULL;
}

/* this is only implemented for HDF5 at the moment */
void DMSLmodel_get_tuple_count(damsel_model model, int *container_count, int *tag_count) {
  push_ndc("DMSLget_tuple_count");
  assert(model != NULL);
  damsel_model_internal *mi = (damsel_model_internal *) model;
  damsel_attached_file *file = mi->attached_file;
  assert(file != NULL);

  if(container_count!=NULL) {
    *container_count =
      linkedlist_length(file->first_pushed_container);
    printf_ndc("container count is %d", *container_count);
  }

  if(tag_count != NULL) {
    assert(file->model_populated != NULL);
    *tag_count = file->model_populated->tag_count;
  }
  pop_ndc();
}

damsel_err_t DMSLmodel_get_container_infos(damsel_model model_id,
  damsel_container_buf_type *container_buf) {
  push_ndc("DMSLmodel_get_container_infos");
  assert(model_id != NULL);
  assert(container_buf != NULL);

  int handle_container_count;
  damsel_attached_file *src_hdf5_store =
    (damsel_attached_file *) ((damsel_model_internal *) model_id)->attached_file;
  DMSLmodel_get_tuple_count(model_id, &handle_container_count, NULL);

  assert(src_hdf5_store->store.type == DAMSEL_STORETYPE_HDF5);

  assert(src_hdf5_store->model_populated != NULL);
  push_ndc("transcribing pointers into return structure");
  // the output should perhaps be a transcription of store's linked list of
  // containers... pull the container IDs for each across.

// this should use the pushed containers list? (which is slowly turning into
// the definitive container list for the file side?
  linkedlist *ll = src_hdf5_store->first_pushed_container;

  int i=0;
  while(ll!=NULL) {
    damsel_hdf5_internal_pushed_container *pc =
      (damsel_hdf5_internal_pushed_container *) ll->head;
    damsel_container hc = pc->file_container;
    printf_ndc("i = %d, handle_container_count = %d, hc = %p", i, handle_container_count, hc);
    assert(i<handle_container_count);
    assert_valid_handle_container_id(hc);
    container_buf[i] = (damsel_container)hc;
    i++; ll=ll->next;
  }
  assert(i==handle_container_count);
  pop_ndc();
  pop_ndc();
  return DMSL_OK;
}

damsel_err_t DMSLmodel_get_tag_infos(damsel_model model_id,
  damsel_tag_buf_type *tag_buf) {
  // i think the tag_colelction_dataset is already open (at least,
  // get_tuple_count seems to think so
  // so then I need to iterate over that and construct appropraite structs
  // in tag_buf for each entry.

  int tag_count;

  DMSLmodel_get_tuple_count(model_id, NULL, &tag_count);

  damsel_model_internal *mi = (damsel_model_internal *)model_id;
  damsel_hdf5_infile_tag_container *col = mi->attached_file->model_populated->tag_info;

  // don't use DMSLhdf5_get_tag_tuples(file, col) here, because
  // i want to reference the memory inside the cached tag_info
  // struct when putting pointers to string in. that way, they
  // stay allocated until the model is closed.

  hssize_t ix;
  for(ix=0; ix<tag_count; ix++) {
    printf_ndc("model_get_tag_infos: transcribing tag %d", (int)ix);
    tag_buf[ix].name=col[ix].name;
    tag_buf[ix].tag_datatype=damsel_serializable_type_to_damsel_type(col[ix].tag_datatype);
    tag_buf[ix].tag_handle=&(col[ix].tag_handle);
  }

  return DMSL_OK;
}

/* this is going to share quite a lot of code with DMSLselect_range -
   is there a nice way I can factor that common code? */
damsel_container DMSLselect_handles_with_values(damsel_model model,
  damsel_handle index_tag) {

  damsel_err_t err;

  push_ndc("DMSLselect_handles_with_values");

  assert(model!=NULL);

  damsel_model_internal *ms =(damsel_model_internal *)model;
  assert(ms->store.type == DAMSEL_STORETYPE_MEMORY);

  damsel_attached_file *store = (damsel_attached_file *)ms->attached_file;

  assert(store->store.type == DAMSEL_STORETYPE_HDF5);
  damsel_attached_file *filestore = (damsel_attached_file *)store;

#ifdef DAMSEL_SUBFILE
  damsel_attached_file *subfile_store;
  damsel_attached_file *subfile_filestore;
  if(ms->num_subfiles > 1) {
      subfile_store = (damsel_attached_file *)ms->attached_subfile;
      assert(subfile_store->store.type == DAMSEL_STORETYPE_HDF5);
      subfile_filestore = (damsel_attached_file *)subfile_store;
  }
#endif

  int found_handles_size = VECTOR_MAX;
  damsel_handle *found_handles = malloc(sizeof(damsel_handle) * found_handles_size);
  int num_found_handles = 0;

  int seq_container_count, vec_row_count;
#ifndef DAMSEL_SUBFILE
  DMSLhdf5_get_container_counts(filestore, &seq_container_count, &vec_row_count);
#else
  if(ms->num_subfiles == 0)
    DMSLhdf5_get_container_counts(filestore, &seq_container_count, &vec_row_count);
  else
    /* TODO: this info should be read from master or subfile? */
    /* TODO: presently reading from the master */
    DMSLhdf5_get_container_counts(filestore, &seq_container_count, &vec_row_count);
#endif
  printf_ndc("seq_container_count=%d, vec_row_count=%d\n", 
             seq_container_count, vec_row_count);
  damsel_hdf5_infile_seq_handle_container *seq_containers = malloc(sizeof(damsel_hdf5_infile_seq_handle_container) * seq_container_count);
  damsel_hdf5_infile_vector_handle_container_row *vec_container_rows = malloc(sizeof(damsel_hdf5_infile_vector_handle_container_row) * vec_row_count);

#ifndef DAMSEL_SUBFILE
  DMSLhdf5_get_container_seq_infos(store, seq_containers);
#else
  if(ms->num_subfiles == 0) /* file is not partitioned */
    DMSLhdf5_get_container_seq_infos(store, seq_containers);
  else
    /* TODO: for now, reading from the master */
    DMSLhdf5_get_container_seq_infos(store, seq_containers);
#endif

  damsel_container empty_container;
  err = DMSLcontainer_create_vector((damsel_model)store, NULL, 0, &empty_container);

  // get list of handle containers in file
  // for each handle container:
  int seq_container_num;
  for(seq_container_num = 0; seq_container_num < seq_container_count;
    seq_container_num++) {
    printf_ndc("probing sequence container %d with handle %llx", seq_container_num, seq_containers[seq_container_num].handle);
    // try to open tag+handle

    char datasetname[256] = ""; // better hope this is big enough... BUG
    char *tn = tagname(ms, index_tag);
    if (tn == NULL) { /* tagname not found */
	pop_ndc();
	return empty_container;
    }
#ifndef DAMSEL_SUBFILE
    snprintf(datasetname, 255, "%s-%llx", tn, seq_containers[seq_container_num].handle);
#else
    if(ms->num_subfiles == 0)
      snprintf(datasetname, 255, "%s-%llx", tn, seq_containers[seq_container_num].handle);
    else
    {
      assert(ms->num_subfiles > 1);
      snprintf(datasetname, 255, "%s-%llx-len", tn, seq_containers[seq_container_num].handle);
    }
#endif
    free(tn);

    printf_ndc("trying to open %s", datasetname);
    hid_t dataset = -1;

#ifndef DAMSEL_SUBFILE
    if (H5Lexists(filestore->file_id, datasetname, H5P_DEFAULT)>0)
      dataset = H5Dopen(filestore->file_id, datasetname, H5P_DEFAULT);
#else
    if(ms->num_subfiles == 0) {
      if(H5Lexists(filestore->file_id, datasetname, H5P_DEFAULT)>0)
        dataset = H5Dopen(filestore->file_id, datasetname, H5P_DEFAULT);
    }
    else {
      if(H5Lexists(filestore->file_id, datasetname, H5P_DEFAULT)>0)
        dataset = H5Dopen(filestore->file_id, datasetname, H5P_DEFAULT);
    }
#endif

    if(dataset > 0) {
      printf_ndc("successfully opened %s", datasetname);

      hid_t hc_dataspace = H5Dget_space(dataset);
      assert(hc_dataspace >= 0);
      int ret;

      hsize_t dims, maxdims;
#ifndef DAMSEL_SUBFILE
      ret = H5Sget_simple_extent_dims(hc_dataspace, &dims, &maxdims);
#else
      if(ms->num_subfiles == 0)
        ret = H5Sget_simple_extent_dims(hc_dataspace, &dims, &maxdims);
      else
      {
        assert(ms->num_subfiles > 1);
        ret = H5Dread(dataset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &dims);
      }
#endif
      assert(ret >= 0);

      printf_ndc("this sequence handle container has %d elements.",(int)dims);

      int dataset_pos;
      for(dataset_pos=0; dataset_pos < dims; dataset_pos++) {
        printf_ndc("copying handle at offset %d", dataset_pos);
        damsel_handle found_handle = seq_containers[seq_container_num].base
          + dataset_pos * seq_containers[seq_container_num].step;
        assert(num_found_handles < found_handles_size);
        found_handles[num_found_handles] = found_handle;
        num_found_handles++;
        if(num_found_handles >= found_handles_size) {
          printf_ndc("reallocing: old found_handles = %p", found_handles);
          found_handles_size *= 2;
          printf_ndc("new found_handles_size = %d", found_handles_size);
          found_handles = realloc(found_handles, found_handles_size * sizeof(damsel_handle));
          printf_ndc("new found_handles = %p", found_handles);
        }
        assert(num_found_handles <= found_handles_size);
      }

      ret = H5Sclose(hc_dataspace);
      assert(ret >= 0);
      ret = H5Dclose(dataset);
      assert(ret >= 0);
      printf_ndc("successfully closed %s", datasetname);
    } else {
      printf_ndc("could not find that HDF5 dataset, but that doesn't matter");
    }
  }
  free(seq_containers);

  DMSLhdf5_get_container_vec_infos(store, vec_container_rows);
  int vec_row_num;
  for(vec_row_num=0; vec_row_num<vec_row_count; vec_row_num++) {
    printf_ndc("probing vector container row %d with handle %llx", vec_row_num, vec_container_rows[vec_row_num].handle);
    // try to open tag+handle

    char datasetname[256] = ""; // better hope this is big enough...
    char *tn = tagname(ms, index_tag);
    if (tn == NULL) {
	pop_ndc();
	return empty_container;
    }
#ifndef DAMSEL_SUBFILE
    snprintf(datasetname, 255, "%s-%llx", tn, vec_container_rows[vec_row_num].handle);
#else
    if(ms->num_subfiles == 0)
      snprintf(datasetname, 255, "%s-%llx", tn, vec_container_rows[vec_row_num].handle);
    else
    {
      assert(ms->num_subfiles > 1);
      snprintf(datasetname, 255, "%s-%llx-len", tn, vec_container_rows[vec_row_num].handle);
    }
#endif
    free(tn);

    printf_ndc("trying to open %s", datasetname);
	hid_t dataset = -1;
	if (H5Lexists(filestore->file_id, datasetname, H5P_DEFAULT)>0)
      dataset = H5Dopen(filestore->file_id, datasetname, H5P_DEFAULT);
    if(dataset > 0) {
      printf_ndc("successfully opened %s", datasetname);

      hid_t hc_dataspace = H5Dget_space(dataset);
      assert(hc_dataspace >= 0);
      int ret;

      hsize_t dims, maxdims;
      ret = H5Sget_simple_extent_dims(hc_dataspace, &dims, &maxdims);
      assert(ret >= 0);

      printf_ndc("this vector handle container has %d elements.",(int)dims);

      int dataset_pos;
      for(dataset_pos=0; dataset_pos < dims; dataset_pos++) {
        printf_ndc("copying handle at offset %d", dataset_pos);
        damsel_handle found_handle = vec_container_rows[vec_row_num].content[dataset_pos];

// TODO: this offset calculation needs to take into account the fact that
// we could be not in the first row of the container. If we're not in the
// right row, we can skip this for now and then we'll get the right handle
// when we are in the right row.

        assert(num_found_handles < found_handles_size);
        found_handles[num_found_handles] = found_handle;
        num_found_handles++;
        if(num_found_handles >= found_handles_size) {
          found_handles_size *= 2;
          found_handles = realloc(found_handles, found_handles_size * sizeof(damsel_handle));
          assert(found_handles != NULL);
        }
        assert(num_found_handles <= found_handles_size);
      }

      ret = H5Sclose(hc_dataspace);
      assert(ret >= 0);
      ret = H5Dclose(dataset);
      assert(ret >= 0);
      printf_ndc("successfully closed %s", datasetname);
    } else {
      printf_ndc("could not find that HDF5 dataset, but that doesn't matter");
    }
  }

  free(vec_container_rows);
  vec_container_rows=NULL;

  printf_ndc("creating vector handle container to store %d selected handles.",
    (int)num_found_handles);

  {
    push_ndc("Entries of vector container will be:");
    int z;
    for(z=0;z<num_found_handles;z++) {
      printf_ndc("position %d handle %llx", z, found_handles[z]);
    }
    pop_ndc();
  }

  damsel_container id;
  err = DMSLcontainer_create_vector((damsel_model)store, found_handles, num_found_handles, &id);
  free(found_handles);
  printf_ndc("Created vector container with id %p", id);

  pop_ndc();
  return id;
}

damsel_container DMSLselect_range(damsel_model model,
  damsel_handle index_tag, short int start, short int end) {

  damsel_err_t err;

  push_ndc("DMSLselect_range");

  assert(model!=NULL);

  damsel_model_internal *ms =(damsel_model_internal *)model;
  assert(ms->store.type == DAMSEL_STORETYPE_MEMORY);

  damsel_attached_file *store = (damsel_attached_file *)ms->attached_file;

  assert(store->store.type == DAMSEL_STORETYPE_HDF5);
  damsel_attached_file *filestore = (damsel_attached_file *)store;

  int found_handles_size = VECTOR_MAX;
  damsel_handle *found_handles = malloc(sizeof(damsel_handle) * found_handles_size);

  damsel_container empty_container;
  err = DMSLcontainer_create_vector((damsel_model)store, NULL, 0, &empty_container);

  int num_found_handles = 0;

  printf_ndc("selecting range from tag, values %d .. %d", (int) start, (int)end);

  int seq_container_count, vec_row_count;
  DMSLhdf5_get_container_counts(filestore, &seq_container_count, &vec_row_count);
  damsel_hdf5_infile_seq_handle_container *containers = malloc(sizeof(damsel_hdf5_infile_seq_handle_container) * seq_container_count);

// BUG: this only works for sequence containers
  DMSLhdf5_get_container_seq_infos(store, containers);

  short int i;
  for(i=start;i<=end;i++) {
    push_printf_ndc("searching for index tag entries with value %d", (int) i);
    // now need to scan over the entirety of the index tag, looking at the
    // values of each.

    // how can I find out the values of all tags? there is no mechanism for
    // that really at the moment... I could go over each handle container and
    // look for a corresponding (tag, container) hdf5 dataset, and if so,
    // look at the values in each of those?

    // get list of handle containers in file
    // for each handle container:
    int container_num;
    for(container_num = 0; container_num < seq_container_count;
      container_num++) {
      printf_ndc("probing container %d with handle %llx", container_num, containers[container_num].handle);
      // try to open tag+handle

      char datasetname[256] = ""; // better hope this is big enough...
      char *tn = tagname(ms, index_tag);
      if (tn == NULL) {
	  pop_ndc();
	  return empty_container;
      }
      snprintf(datasetname, 255, "%s-%llx", tn, containers[container_num].handle);
      free(tn);

      printf_ndc("trying to open %s", datasetname);
      hid_t dataset = H5Dopen(filestore->file_id, datasetname, H5P_DEFAULT);
      if(dataset >= 0) {
        printf_ndc("successfully opened %s", datasetname);
// now we can do something with this dataset that we've just opened.
// we need to scan it to look for the value that we are interested in.
// so hopefully its straightforward to read in the whole dataset.

        hid_t hc_dataspace = H5Dget_space(dataset);
        assert(hc_dataspace >= 0);
        int ret;

        hsize_t dims, maxdims;
        ret = H5Sget_simple_extent_dims(hc_dataspace, &dims, &maxdims);
        assert(ret >= 0);

        printf_ndc("this handle container has %d elements.",(int)dims);
        // note that this code only works for short ints, but that I'm
        // not making any attempt to check that here.
        short int *buffer = malloc(dims * sizeof(short int));
        ret = H5Dread(dataset, H5T_NATIVE_SHORT, H5S_ALL, H5S_ALL,
          H5P_DEFAULT, buffer);
        assert(ret != -1);
        printf_ndc("scanning...");

        int dataset_pos;
        for(dataset_pos=0; dataset_pos < dims; dataset_pos++) {
          if(buffer[dataset_pos] == i) {
            printf_ndc("found at offset %d", dataset_pos);
//   now, figure out which handle this was (by getting the
//   dataset_pos-th handle in the handle container), and add to
//   the list that will go in the result container.
            damsel_handle found_handle = containers[container_num].base 
              + dataset_pos * containers[container_num].step;
            found_handles[num_found_handles] = found_handle;
            num_found_handles++;
            if(num_found_handles >= found_handles_size) {
              found_handles_size *= 2;
              found_handles = realloc(found_handles, found_handles_size);
            }
            assert(num_found_handles <= found_handles_size);
          }
        }

        free(buffer);

        ret = H5Sclose(hc_dataspace);
        assert(ret >= 0);
        ret = H5Dclose(dataset);
        assert(ret >= 0);
        printf_ndc("successfully closed %s", datasetname);
      } else {
        printf_ndc("could not find that HDF5 dataset, but that doesn't matter");
      }
    }
    pop_ndc();
  }
  free(containers);

  printf_ndc("creating vector handle container to store %d selected handles.", (int)num_found_handles);

  {
    push_ndc("Entries of vector container will be:");
    int z;
    for(z=0;z<num_found_handles;z++) {
      printf_ndc("position %d handle %llx", z, found_handles[z]);
    }
    pop_ndc();
  }

  damsel_container id;
  err = DMSLcontainer_create_vector((damsel_model)store, found_handles, num_found_handles, &id);
  printf_ndc("Created vector container with id %p", id);
  free(found_handles);
  pop_ndc();
  return id;
}

damsel_err_t
DMSLselect_tag_by_name(damsel_model   model,
                       const char    *tagname,
                       damsel_handle *found_handle)  /* OUT */
{
    /* sanity check */
    assert(model!=NULL); /* if (model == NULL) return DMSL_EMODEL; */
    assert(tagname != NULL); /* if (tagname == NULL) return DMSL_EINVALTAG */

  push_ndc("DMSLselect_tag_by_name");

  damsel_model_internal *ms =(damsel_model_internal *)model;
  assert(ms->store.type == DAMSEL_STORETYPE_MEMORY);

  damsel_attached_file *store = (damsel_attached_file *) ms->attached_file;

  assert(store != NULL);
  assert(store->store.type == DAMSEL_STORETYPE_HDF5);

  printf_ndc("selecting tag by name: %s", tagname);

  int tag_count;
  DMSLmodel_get_tuple_count(model, NULL, &tag_count);

  printf_ndc("Tag count: %d", tag_count);

  damsel_hdf5_infile_tag_container *col =
    malloc(tag_count * sizeof(damsel_hdf5_infile_tag_container));

  DMSLhdf5_get_tag_tuples(store, col);

  int tag_num, found=0;
  *found_handle = 0;

  for(tag_num=0; tag_num<tag_count; tag_num++) {
    printf_ndc("Tag %d is %s:", tag_num, col[tag_num].name);
    if(!strcmp(col[tag_num].name, tagname)) {
      printf_ndc("FOUND!");
      found=1;
      *found_handle = col[tag_num].tag_handle;
      break;
    }
  }

  free(col);

  if(found == 0) {
      *found_handle = DAMSEL_HANDLE_TYPE_INVALID;
      return DMSL_ENOTFOUND;
  }

  printf_ndc("returning handle %llx", *found_handle);

  pop_ndc();

  return DMSL_OK;
}

/*! returns the name of the tag specified by the tag_handle.
    the caller needs to release the memory returned by this function. */
char *tagname(damsel_model_internal *mi, damsel_handle tag_handle) {

  push_ndc("tagname");
  assert(mi != NULL);

  damsel_attached_file *store = mi->attached_file;
  assert(store != NULL);

  assert(store->store.type == DAMSEL_STORETYPE_HDF5);

  int found=0;
  char *found_name=NULL;

  printf_ndc("selecting tag by handle: %llx", tag_handle);

  int tag_count;
  DMSLmodel_get_tuple_count((damsel_model)mi, NULL, &tag_count);

  printf_ndc("Tag count: %d", tag_count);

  damsel_hdf5_infile_tag_container *col =
    malloc(tag_count * sizeof(damsel_hdf5_infile_tag_container));

  DMSLhdf5_get_tag_tuples((damsel_attached_file *)store, col);

  int tag_num;
  for(tag_num=0; tag_num<tag_count; tag_num++) {
    printf_ndc("Tag %d is %s with handle %llx", tag_num, col[tag_num].name, col[tag_num].tag_handle);
    if(col[tag_num].tag_handle == tag_handle) {
      printf_ndc("This matches the tag handle that we are looking for.");
      found=1;
      found_name = malloc(sizeof(col[tag_num].name)+1);
      strcpy(found_name, col[tag_num].name);
    }
  }

  free(col);

  if(found == 0) {
    printf_ndc("Did not find a tag with handle %llx", tag_handle);
    pop_ndc();
    return NULL;
  }

  printf_ndc("returning tag name %s", found_name);

  assert(found == 1);
  pop_ndc();
  return(found_name);
}


// a DMSLmodel_attach that uses file stores underneath, as a transition
// away from stores.

damsel_err_t DMSLmodel_attach(damsel_model model_id, const char *filename, MPI_Comm comm, damsel_trait_bundle traits) {
  assert_is_initialized();
  assert(model_id != NULL);
  assert(filename != NULL);

  assert(((damsel_store_internal *)model_id)->type == DAMSEL_STORETYPE_MEMORY);

  damsel_model_internal *mi = (damsel_model_internal *)model_id;

  return damsel_internal_hdf5_open(mi, filename, comm, traits);
}

damsel_err_t DMSLmodel_transfer_sync(damsel_model model_id,
                                     damsel_transfer_type transfer_type) {
  damsel_request_t transfer_request;
  damsel_status_t status;
  DMSLmodel_transfer_async(model_id, transfer_type, &transfer_request);
  DMSLmodel_wait(transfer_request, &status);
  return DMSL_OK; // TODO should do something with status
}

// for transition, using damsel store* for representing model_ids
// user MPI_Request instead of damsel request
damsel_err_t DMSLmodel_transfer_async(
                                          damsel_model model_id,
                                            /**< model being queried */
                                          damsel_transfer_type transfer_type,
                                            /**< type of transfer */
                                          damsel_request_t *transfer_request
                                         /**< request object for the asynchro
nous operation */
                                          ) {
  damsel_execute_args *args;
  assert_is_initialized();
  damsel_err_t ret;
  assert(model_id != NULL);
  assert(transfer_request != NULL);
  damsel_model_internal *m_model = (damsel_model_internal *)model_id;
  assert(m_model->attached_file != NULL);
  assert(m_model->attached_file->store.type == DAMSEL_STORETYPE_HDF5);

// this checks that when we specify a direction here, we match up with the
// directions that we've already implicitly specified.
  if(m_model->direction == DAMSEL_TRANSFER_TYPE_INVALID) {
    m_model->direction = transfer_type;
  } else {
    assert(m_model->direction == transfer_type);
  }

  args = damsel_execute_args_create((damsel_model_internal *)model_id, &(transfer_request->mpi), transfer_type, m_model->attached_file);
  MPI_Grequest_start(damsel_query_fn, damsel_free_fn, damsel_cancel_fn, args, &(transfer_request->mpi));
  printf_ndc("MPI Grequest started, request is %lx", (unsigned long)transfer_request->mpi);
#ifdef DAMSEL_SINGLETHREAD
  damsel_transfer_thread(args);
#else
  pthread_create(&(transfer_request->posix), NULL, damsel_transfer_thread, args); // TODO use return value of this for success or not
  printf_ndc("after pthread creation, MPI request is %x", transfer_request->mpi);
#endif
  ret.id = 0;
  return ret;
}

damsel_err_t DMSLmodel_wait(damsel_request_t transfer_request,
                            damsel_status_t *request_status
                           ) {
  push_ndc("DMSLmodel_wait");
  assert_is_initialized();
  MPI_Status s;
  damsel_err_t err;
  printf_ndc("waiting for MPI completion: tranfer_request.mpi = %lx", (unsigned long)transfer_request.mpi);
  int r = MPI_Wait(&(transfer_request.mpi), &s);

  assert(r == MPI_SUCCESS);

// TODO BUG need to use MPI status to set return status and err code.
  request_status->id = 0;

// TODO: do a thread join now to tidy up the thread
// this looks like a double barrier for the same thing - the MPI wait
// and the posix thread...

#ifndef DAMSEL_SINGLETHREAD
  printf_ndc("waiting for pthread_join");
  pthread_join(transfer_request.posix, NULL);
#endif

  err.id = 0;
  pop_ndc();
  return err;
}


/* converts damsel types to a representation for storing in hdf5 wrapper
   which consists at the moment of arbitrary lines. by "internal"
   type here, do I mean an HDF5 specific representation or a 
   damsel-guts specific representation?

   now that the API exposed types are integers, I can use that number as
   an internal type number too, rather than doing mapping here.

 */

long damsel_type_to_serializable_type(damsel_data_type t) {
  assert_is_damsel_type(t);
  return (long)t; // BUG when there are more than a long-worth of types (!) this will wrap.
}

damsel_data_type damsel_serializable_type_to_damsel_type(long it) {
  damsel_data_type t = (damsel_data_type)it;

  if(!inner_assert_is_damsel_type(t)) {
    printf_ndc("serializable type %lx converts to damsel_data_type %llx: this is not a damsel type", it, t);
  }
  assert_is_damsel_type(t);
  return t;
}

damsel_model DAMSEL_MODEL_INVALID;
damsel_library DAMSEL_LIBRARY_INVALID;
damsel_container DAMSEL_CONTAINER_INVALID;

damsel_err_t DMSLlib_init(damsel_library *lib) {
  init_ndc();
  void *ctx = push_ndc("DMSLlib_init");
  printf_ndc("initializing damsel\n");
  dump_ndc();

  DAMSEL_MODEL_INVALID=NULL;
  DAMSEL_LIBRARY_INVALID=NULL;
  DAMSEL_CONTAINER_INVALID=NULL;

  int is_mpi;
  MPI_Initialized(&is_mpi);
  if(!is_mpi) {
    printf_ndc("MPI is not initialized.");
    *lib = DAMSEL_LIBRARY_INVALID;
    return DMSL_EINVALLIB;
  }
  /* not sure if this is the best place, but if we want to ensure the
   * apropriate level of thread support in the MPI library, this is what we'd
   * do */
#ifndef DAMSEL_SINGLETHREAD
  int threadlvl;
  MPI_Query_thread(&threadlvl);
  printf_ndc("Thread level check: thread level is %d; we are expecting MPI_THREAD_MULTIPLE = %d", threadlvl, MPI_THREAD_MULTIPLE);
  if(threadlvl != MPI_THREAD_MULTIPLE) {
    printf_ndc("Thread level is incorrect.");
    *lib = DAMSEL_LIBRARY_INVALID;
    return DMSL_EINVALLIB;
  }
#else
  printf_ndc("No thread level check: damsel was built in single threaded mode.");
#endif

  init_types();

  globals_initialized=1;
  assert_is_initialized();
  pop_ndc_context(ctx);
  *lib = (damsel_library)DMSLlib_init;
  return DMSL_OK;
}

damsel_err_t DMSLlib_finalize(damsel_library library_id) {
  assert_is_initialized();
  damsel_err_t err;
  if(library_id == (damsel_library)DMSLlib_init) {
    printf_ndc("finalising the only instance this implementation supports");
    release_types();
    H5close();
    err.id = 0;
  } else {
    printf_ndc("incorrect instance ID when finalizing");
    err.id = -1;
  }
  pop_ndc(); // the top level context created in initialize
  finalize_ndc();
  return err;
}

/* this impl will always return true... what I should do really is look up
   this_id in all the structures I know about. */
//int DMSLlib_ok(damsel_library library_id, damsel_id_t this_id) {
//   return (1==1);
//}

damsel_err_t DMSLmodel_create(damsel_handle_type  app_handle_type,
                                   /**< [in] Application-space Handle type to be used with this model */
                              damsel_model       *model
                                 ) {
  if (globals_initialized != 1)
      return DMSL_EUNINITIALIZED;

  if (app_handle_type != DAMSEL_HANDLE_TYPE_HANDLE64)
      return DMSL_EHANDLETYPE;
  // because that's all that is supported at the moment...

  printf_ndc("creating memory map store");
  damsel_model_internal *s = malloc(sizeof(damsel_model_internal));
  printf_ndc("store id is %p\n", s);
  if (s == NULL) return DMSL_EMALLOC;

  s->store.type = DAMSEL_STORETYPE_MEMORY;
//  s->store.first_container = NULL;
  s->store.refcount = 1;
  s->first_cached_data = NULL;
  s->first_entity_definition = NULL;
  s->attached_file = NULL;
#ifdef DAMSEL_SUBFILE
  s->attached_subfile = NULL;
  s->num_subfiles = 0;
#endif
  s->first_metadata_handle = NULL;
  s->mappings = NULL;
  s->more_mappings_allowed = 1;
  s->first_tag_definition = NULL;
  s->first_collection = NULL;
  s->global_mappings = NULL;
  s->direction = DAMSEL_TRANSFER_TYPE_INVALID;
  s->transaction_stack = NULL;
  s->first_chunk = NULL;

  *model = (damsel_model)s;
  return DMSL_OK;
}

int inner_assert_is_damsel_type(damsel_data_type t) {
  // this will check that we have a damsel_id / pointer style damsel id
  // which I'm using to make sure I don't get mixed up with a position-in-
  // list style type id.
  return (t <= linkedlist_length(defined_types) + DAMSEL_DATA_TYPE_PREDEFINED_WATERMARK);
}

int inner_assert_valid_tag_handle(damsel_handle h, damsel_model_internal *model) {
  // this will check that the specified tag handle has been defined in
  // the specified store
  assert(model != NULL);
  linkedlist *ll = model->first_tag_definition;
  while(ll!=NULL) {
    damsel_tag_internal *td = (damsel_tag_internal *)ll->head;
    if(td->handle == h) return (1==1);
    ll = ll->next;
  }
  printf("handle %llx is not a valid tag handle in model %p\n", h, model);
  // if we reach here, we didn't find the tag definition
  return (1==0);
}


/*! this is called when the app thinks there are maybe some tags it has
    mapped that have not been resolved yet.
*/
void DMSLmap_resolve_read_tags_internal(damsel_model model_id) {
  push_ndc("DMSLmap_resolve_read_tags");
  int tag_count;

  damsel_model_internal *ms = (damsel_model_internal *)model_id;

  damsel_attached_file *file = (damsel_attached_file *)ms->attached_file;

  DMSLmodel_get_tuple_count(model_id, NULL, &tag_count);

  damsel_hdf5_infile_tag_container *col =
    malloc(tag_count * sizeof(damsel_hdf5_infile_tag_container));

  DMSLhdf5_get_tag_tuples(file, col);

  int ix;
  for(ix=0;ix<tag_count;ix++) {
    // does this exist on memory side?
    // compare by names, I suppose? there's no other persistent ID(!)

    char *fileside_name=col[ix].name;
    printf_ndc("checking if infile tag %d, name=%s is defined on the application side", ix,fileside_name);
    int found=0;
    linkedlist *ll = ms->first_tag_definition;
    while(ll !=NULL && found == 0) {
      damsel_tag_internal *heretag = (damsel_tag_internal *)ll->head;
      if(!strcmp(heretag->name, fileside_name)) {
        printf_ndc("found that tag already defined on application side");
        found=1;
      }
      ll=ll->next;
    }

    if(found == 0) {
      printf_ndc("that tag was not found on application side. checking if its handle is mapped");
      // if not, then is its file-side handle mapped to a memory side handle?
      damsel_handle fileside_handle = col[ix].tag_handle;
      damsel_handle memside_handle;
      damsel_err_t ret;
      ret = handle_map(ms, DAMSEL_TRANSFER_TYPE_READ, fileside_handle,
        &memside_handle);
      if(ret.id == 0) {
        printf_ndc("handle maps - creating corresponding app-side tag");
        // if it maps, then create mem-side tag
        damsel_data_type type;
        type = damsel_serializable_type_to_damsel_type(col[ix].tag_datatype);
        DMSLtag_define_internal(model_id, &memside_handle, type, fileside_name, DAMSEL_WRITTEN);
      } else {
        printf_ndc("handle does not map - not creating an app-side tag");
      }
    }
  }
  free(col);
  pop_ndc();
}

void DMSLmodel_start_read(damsel_model m) {
  damsel_model_internal *mi = (damsel_model_internal *) m;
  linkedlist_prepend(&(mi->transaction_stack), mi->first_cached_data);
}

void DMSLmodel_stop_read(damsel_model m) {
  damsel_model_internal *mi = (damsel_model_internal *) m;
  assert(mi->transaction_stack != NULL);
  linkedlist *o = mi->transaction_stack;
  mi->transaction_stack = o->next;
  mi->first_cached_data = o->head;
  free(o);
}

void internal_record_pushed_container(damsel_attached_file *f,
  damsel_container app_container, damsel_container file_container,
  damsel_handle file_handle, int remote) {

  assert(f->container_push_allowed == 1);


  damsel_container_internal *app_internal = (damsel_container_internal *) app_container;
  damsel_container_internal *file_internal = (damsel_container_internal *) file_container;
  assert(app_internal == NULL || ((damsel_store_internal *)app_internal->model)->type == DAMSEL_STORETYPE_MEMORY);
  assert(file_internal == NULL || ((damsel_store_internal *)file_internal->model)->type == DAMSEL_STORETYPE_HDF5);


// I can probably make an assertion here based on the file handle
// because I know the encoding of rank into file handle
// but only at write time... at read time, I might be reading
// handles into any rank

/*
#ifndef NDEBUG
  int rank_in_mpi;
  rank_in_mpi = f->myrank;
  int rank_in_file_handle = file_handle >> 32;
  if(remote==0) {
    assert(rank_in_file_handle == rank_in_mpi);
  } else {
    assert(rank_in_file_handle != rank_in_mpi);
  }
#endif
*/

  assert(f->container_push_allowed == 1);

  push_ndc("internal_record_pushed_container");
  printf_ndc("recording pushed container, app_container = %p, file_container = %p, handle = %llx", app_container, file_container, file_handle);


  damsel_hdf5_internal_pushed_container *pc =
    malloc(sizeof(damsel_hdf5_internal_pushed_container));
  assert(pc != NULL);
  pc->handle = file_handle;
  pc->container = app_container;
  pc->file_container = file_container;
  pc->remote = remote;

  if(app_container) DMSLcontainer_take(app_container); // because its allow to use NULL for the app container - it might not exist
  DMSLcontainer_take(file_container);

#ifndef NDEBUG
  linkedlist *ll;
  for(ll = f->first_pushed_container; ll!= NULL; ll=ll->next) {
    damsel_hdf5_internal_pushed_container *ptest =
      (damsel_hdf5_internal_pushed_container *) ll->head;
    printf_ndc("checking pushed_container record %p for duplicate push", ptest);
    assert(ptest->container ==0 || ptest->container != pc->container);
    assert(ptest->handle != pc->handle);
    assert(ptest->file_container != pc->file_container);
  }
#endif

  linkedlist_prepend(&f->first_pushed_container, pc);
  pop_ndc();
}


// should have corresponding pull call, rather than expecting it to
// behave symmetrically?
// this one should push a container from the application side into the
// file side of things, and return the file-side handle to that container
// (or perhaps it should return a container structure? but for referring
// to things in the file, I want file side handles, or at least file-side
// container IDs that are not pointers, but instead persistent)

// this call is idempotent - if the same container is pushed multiple times,
// it will only end up in the push list once

damsel_err_t internal_push_container(damsel_container app_container_id, damsel_handle *handleptr) {
  damsel_err_t err;
  void *ctx = push_printf_ndc("internal_push_container: container %p", app_container_id);
  // TODO: assert this is valid handle container and that its on the memory
  //  side


  damsel_container_internal *app_container =
    (damsel_container_internal *) app_container_id;

  damsel_model_internal *model = (damsel_model_internal *) app_container->model;
  damsel_attached_file *file = (damsel_attached_file *) model->attached_file;

  // check if its pushed already
  linkedlist *ll;
  for(ll = file->first_pushed_container; ll!= NULL; ll=ll->next) {
    damsel_hdf5_internal_pushed_container *p = (damsel_hdf5_internal_pushed_container *)ll->head;
    if(p->container == app_container_id) {
      if(handleptr != NULL) *handleptr = p->handle;
      pop_ndc_context(ctx);
      return DMSL_OK;
    }
  }


  size_t qty = ((damsel_container_internal *)app_container_id)->quantity;

  damsel_handle *fileside_handles = malloc(sizeof(damsel_handle) * qty);

  int i;
  for(i=0; i<qty; i++) {
    damsel_handle source_handle = DMSLcontainer_handle_at_position(app_container_id, i);

    damsel_handle dest_handle;
    damsel_err_t ret;
    printf_ndc("mapping source handle %llx in store %p to store %p:", source_handle, model, file);
    ret = handle_map(model, DAMSEL_TRANSFER_TYPE_WRITE, source_handle,
      &dest_handle);

    assert_dmsl_ok(ret); // TODO: is it acceptable to not find a handle? i'm unsure... its failing sometimes.

    printf_ndc("mapped source handle %llx to destination handle %llx", source_handle, dest_handle);

    if(ret.id != 0) {
      printf_ndc("source handle %llx has no destination handle", source_handle);
      pop_ndc_context(ctx);
      return DMSL_ERROR_GENERAL;
    }

    assert(ret.id == 0);

    fileside_handles[i] = dest_handle;
  }

  damsel_container file_container;
  err = DMSLcontainer_create_vector((damsel_model)file, fileside_handles, qty, &file_container);
  free(fileside_handles);

  damsel_handle fh = get_new_handles(file, 1);

  internal_record_pushed_container(file, app_container_id,
    file_container, fh, 0);
  DMSLcontainer_release(file_container);

  pop_ndc_context(ctx);
  if(handleptr != NULL) *handleptr = fh;
  return DMSL_OK;
}

/*! returns a handle that has been previously allocated when pushing a container */
damsel_handle internal_push_container_lookup(damsel_container app_container_id) {
  void *ctx = push_printf_ndc("internal_push_container_lookup, container %p", app_container_id);
  assert(app_container_id != NULL);
  damsel_container_internal *app_container =
    (damsel_container_internal *) app_container_id;
  damsel_model_internal *model = (damsel_model_internal *) app_container->model;
  assert(model!=NULL);
  assert(model->store.type == DAMSEL_STORETYPE_MEMORY);
  damsel_attached_file *file = (damsel_attached_file *) model->attached_file;
  assert(file != NULL);
  assert(file->store.type == DAMSEL_STORETYPE_HDF5);
  linkedlist *l = file->first_pushed_container;
  while(l!=NULL) {
    damsel_hdf5_internal_pushed_container *pc = l->head;
    printf_ndc("checking container id = %p, handle = 0x%llx", pc->container, pc->handle);
    if(pc->container == app_container_id) {
      printf_ndc("found file-side handle %llx", pc->handle);
      pop_ndc_context(ctx);
      return pc->handle;
    }
    l=l->next;
  }
  pop_ndc_context(ctx);
  assert(1==0 && "tried to look up a container that has not been pushed");
  return DAMSEL_HANDLE_TYPE_INVALID;
}

/*! returns a handle that has been previously allocated when pushing a container, using a file-side container ID. (this will also look up remote pushed container
handles TODO) 

TODO should rename this to reflect the fact that its not only getting
file handles from the pushed container list, but also from the
remote mappings list

*/
damsel_handle internal_push_container_lookup_fileside(damsel_container file_container_id) {
  void *ctx = push_printf_ndc("internal_push_container_lookup_fileside, container %p", file_container_id);
  assert(file_container_id != NULL);
  damsel_container_internal *file_container =
    (damsel_container_internal *) file_container_id;
  damsel_attached_file *file = (damsel_attached_file *) file_container->model;
  assert(file != NULL);
  assert(file->store.type == DAMSEL_STORETYPE_HDF5);

  push_ndc("checking local pushed containers");

  linkedlist *l = file->first_pushed_container;
  while(l!=NULL) {
    damsel_hdf5_internal_pushed_container *pc = l->head;
    printf_ndc("checking container id = %p, fileside id = %p, handle = 0x%llx", pc->container, pc->file_container, pc->handle);
    if(pc->file_container == file_container_id) {
      printf_ndc("found file-side handle %llx", pc->handle);
      pop_ndc();
      pop_ndc_context(ctx);
      return pc->handle;
    }
    l=l->next;
  }

  pop_ndc();

  push_ndc("checking remote mapped containers");

  linkedlist *rl;
  for(rl = file->remote_mappings; rl != NULL ; rl = rl->next) {
    damsel_remote_mapping *rm = (damsel_remote_mapping *) rl->head;
    if(rm->file_container == file_container_id) {
      printf_ndc("found file-side handle %llx from rank %d", rm->file_handle, rm->rank);
      pop_ndc();
      pop_ndc_context(ctx);
      return rm->file_handle;
    }
  }

  pop_ndc();

  pop_ndc_context(ctx);
  assert(1==0 && "tried to look up a container that has not been pushed");
  return DAMSEL_HANDLE_TYPE_INVALID;
}


damsel_err_t DMSLmodel_get_collection_infos(damsel_model model_id,
                                            damsel_collection_buf_type *collection_buf)
{
  push_ndc("DMSLmodel_get_collection_infos");
  assert(model_id != NULL);
  damsel_model_internal *mi = (damsel_model_internal *) model_id;
  damsel_attached_file *file = mi->attached_file;
  assert(file != NULL);

  assert(file->collection_dataset >= 0);

  int count = DMSLmodel_get_collection_count(model_id);
  damsel_hdf5_infile_collection *ibuf = malloc(count * sizeof(damsel_hdf5_infile_collection));

  hid_t colltype = get_collection_type(file);

  int ret = H5Dread(file->collection_dataset, colltype, H5S_ALL, H5S_ALL, H5P_DEFAULT, ibuf);
  assert(ret >= 0);

  int ix;
  for(ix = 0; ix < count; ix++) {
    collection_buf[ix].type = ibuf[ix].type;

    damsel_handle *newhandle = malloc(sizeof(damsel_handle));
    *newhandle = ibuf[ix].collection_handle;
    collection_buf[ix].collection_handle = newhandle;

    damsel_container container = (damsel_container) damsel_handle_to_handle_container(file, ibuf[ix].container_handle);
    assert_valid_handle_container_id(container);

    collection_buf[ix].tag_handle_container = NULL;
    collection_buf[ix].contents = NULL;
    collection_buf[ix].tag_count = 0;

    if (container != NULL) {
      collection_buf[ix].tag_handle_container = container_of_tags_which_are_set_on(model_id, container);
      collection_buf[ix].tag_count = DMSLcontainer_count(collection_buf[ix].tag_handle_container);
      collection_buf[ix].contents = container;
    }
  }

  free(ibuf);

  pop_ndc();
  return DMSL_OK;
}

damsel_err_t DMSLmodel_close_collection_infos(damsel_model model_id,
  damsel_collection_buf_type *collection_buf) {
  int ix;
  int count = DMSLmodel_get_collection_count(model_id);
  for(ix = 0 ; ix < count ; ix ++) {
    free(collection_buf[ix].collection_handle);
    if(collection_buf[ix].tag_handle_container != NULL) {
      DMSLcontainer_release(collection_buf[ix].tag_handle_container);
    }
  }
  return DMSL_OK;
}

int DMSLmodel_get_collection_count(damsel_model model_id) {

  push_ndc("DMSLmodel_get_collection_count");

  hid_t spaceid;
  hssize_t hcsize;
  assert(model_id != NULL);
  damsel_model_internal *mi = (damsel_model_internal *) model_id;
  damsel_attached_file *file = mi->attached_file;
  assert(file != NULL);

  assert(file->collection_dataset >= 0);
  spaceid = H5Dget_space(file->collection_dataset);
  assert(spaceid >= 0);
  hcsize= H5Sget_simple_extent_npoints(spaceid);
  printf_ndc("collection dataset has %d points",
    (int) hcsize);
  H5Sclose(spaceid);

  pop_ndc();
  return (int) hcsize;
}

damsel_container container_of_tags_which_are_set_on(damsel_model m, damsel_container c) {
  damsel_err_t err;
  push_ndc("container_of_tags_which_are_set_on");

  assert_valid_handle_container_id(c);

  damsel_container return_container;
  int tag_count;

  damsel_model_internal *ms = (damsel_model_internal *)m;
  damsel_attached_file *file = ms->attached_file;


  DMSLmodel_get_tuple_count(m, NULL, &tag_count);

  printf_ndc("Tag count is %d", tag_count);

  damsel_hdf5_infile_tag_container *col =
    malloc(tag_count * sizeof(damsel_hdf5_infile_tag_container));

  DMSLhdf5_get_tag_tuples(file, col);

  damsel_handle *handlebuf = malloc(sizeof(damsel_handle) * tag_count);

  int handleix = 0;

  int ix;
  for(ix=0; ix<tag_count; ix++) {
    push_printf_ndc("checking overlap with tag number %d", ix);
    damsel_handle fileside_tag_handle = col[ix].tag_handle;
    damsel_container fc =
      DMSLselect_handles_with_values(m, fileside_tag_handle);
    int overlap = handle_containers_overlap(fc, c);
    DMSLcontainer_release(fc);
    if(overlap == 1) {
      printf_ndc("this tag has some overlap with the target container");
      handlebuf[handleix] = col[ix].tag_handle;
      handleix++;
    } else {
      printf_ndc("this tag has no value overlap with the target container");
    }
    pop_ndc();
  }
  printf_ndc("found a total of %d tags associated with the target container", handleix);
  if(handleix == 0) {
    return_container = NULL;
  } else {
    err = DMSLcontainer_create_vector((damsel_model)file, handlebuf, handleix, &return_container);
  }
  free(handlebuf); 
  free(col);
  pop_ndc();
  return return_container;
}

damsel_err_t DMSLmodel_populate(damsel_model model_id)  {
  damsel_err_t err;
  push_ndc("DMSLmodel_populate");

  damsel_model_internal *ms = (damsel_model_internal *)model_id;
  damsel_attached_file *file = ms->attached_file;
#ifdef DAMSEL_SUBFILE
  damsel_attached_file *subfile = ms->attached_subfile;
#endif

  if (!file) return DMSL_ERROR_GENERAL;
#ifdef DAMSEL_SUBFILE
  if(ms->num_subfiles > 1 && !subfile) return DMSL_ERROR_GENERAL;
#endif
    
  if(file->model_populated == NULL) {
    int handle_seq_container_count, handle_vec_row_count;
    push_ndc("populating model because it hasn't been populated yet");
    file->model_populated=malloc(sizeof(struct populated_info));
    file->model_populated->seq_info = NULL;
    file->model_populated->vec_info = NULL;
    populate_hdf5_get_container_counts(file,
      &handle_seq_container_count, &handle_vec_row_count);
    printf_ndc("handle_seq_container_count=%d, handle_vec_row_count=%d\n",
               handle_seq_container_count, handle_vec_row_count);

    file->model_populated->seq_count = handle_seq_container_count;
    file->model_populated->vec_rows = handle_vec_row_count;
#ifdef DAMSEL_SUBFILE
    if(ms->num_subfiles > 1) {
      printf_ndc("populating the attached subfile too");
#if 0 /* NOTE: no need to populate these for subfiles */
      subfile->model_populated=malloc(sizeof(struct populated_info));
      subfile->model_populated->seq_info = NULL;
      subfile->model_populated->vec_info = NULL;
#endif
    }
#endif
    damsel_hdf5_infile_seq_handle_container *hcseqrows =
      malloc(handle_seq_container_count *
      sizeof(damsel_hdf5_infile_seq_handle_container));

    damsel_hdf5_infile_vector_handle_container_row *hcvecrows =
      malloc(handle_vec_row_count *
      sizeof(damsel_hdf5_infile_vector_handle_container_row));

    populate_hdf5_get_container_seq_infos(file, hcseqrows);
    populate_hdf5_get_container_vec_infos(file, hcvecrows);

    file->model_populated->seq_info = hcseqrows;
    file->model_populated->vec_info = hcvecrows;

    int i;
    push_ndc("restoring sequence containers");
    for(i=0; i<handle_seq_container_count; i++) {
      printf_ndc("restoring sequence container row %d", i);
      damsel_handle collh=hcseqrows[i].handle;
      unsigned quantity=hcseqrows[i].quantity;
      damsel_handle baseh=hcseqrows[i].base;
      size_t step=hcseqrows[i].step;
      damsel_container id;
      err = DMSLcontainer_create_sequence((damsel_model) file, baseh, quantity, step, &id);
      set_written_container(id);
      internal_record_pushed_container(file, NULL, id, collh, 0);
      DMSLcontainer_release(id);
    }
    pop_ndc();
    push_ndc("restoring vector containers");
    for(i=0; i<handle_vec_row_count; i++) {
      printf_ndc("restoring vector container row %d", i);
// TODO: this needs to deal with rows not corresponding to an entire
// container - otherwise this code will break on a container larger than
// VECTOR_MAX.
// I need to assemble a single container from it...
      if(hcvecrows[i].blocknumber != 0) {
        printf_ndc("container row %d is a follow-on block - skipping", i);
      } else {
        printf_ndc("container row %d is an initial block - reading all blocks",
          i);
        damsel_handle collh=hcvecrows[i].handle;
        unsigned quantity=hcvecrows[i].quantity;
        damsel_handle *contents = malloc(sizeof(damsel_handle) * quantity);
        int block;
        for(block=0; block<quantity / VECTOR_MAX + 1; block++) {
          int o;
          int omax;
          if(block == quantity / VECTOR_MAX) { // last block
            omax = quantity % VECTOR_MAX;
          } else {
            omax = VECTOR_MAX;
          }
          for(o=0; o<omax; o++) {
            assert(hcvecrows[i+block].handle == collh);
            assert(hcvecrows[i+block].blocknumber == block);
            int off = o + block*VECTOR_MAX;
            assert(off <= quantity);
            contents[ off ] = hcvecrows[i+block].content[o];
// assumption here is that blocks must follow on in sequence from each other
// in the file
          }
        }
        damsel_container id;
        err = DMSLcontainer_create_vector((damsel_model)file, contents, quantity, &id);
	free(contents);
        internal_record_pushed_container(file, NULL, id, collh, 0);
        DMSLcontainer_release(id);
      }
    }
    pop_ndc();



    push_ndc("populating tag info");
    populate_hdf5_get_tag_count(file,
      &(file->model_populated->tag_count));

    damsel_hdf5_infile_tag_container *hctagrows =
      malloc(file->model_populated->tag_count *
      sizeof(damsel_hdf5_infile_tag_container));

    populate_hdf5_get_tag_tuples(file, hctagrows);

    file->model_populated->tag_info = hctagrows;

    pop_ndc();
    pop_ndc();
  } else {
    printf_ndc("model previously populated - skipping");
  }

  pop_ndc();
  return DMSL_OK;
}

void internal_record_remote_mapping(damsel_attached_file *fs, damsel_container ac, damsel_container fc, damsel_handle fh, int rank) {
  printf_ndc("internal_record_remote_mapping: recording remote mapping app container = %p, file container = %p, handle = %llx, rank = %d", ac, fc, fh, rank);
  damsel_remote_mapping *rm = malloc(sizeof(damsel_remote_mapping));
  assert(rm != NULL);
  rm->app_container = ac;
  rm->file_container = fc;
  rm->file_handle = fh;
  rm->rank = rank;
  linkedlist_prepend(&(fs->remote_mappings),rm);
}

damsel_err_t DMSLmodel_map_global_handles(damsel_container app_container,
                                              int *proc_ranks,
                                              damsel_handle_ptr remote_handles
                                              ) {
/*
typedef struct {
  int rank;
  damsel_container app_container;
  int *proc_ranks;
  damsel_handle_ptr remote_handles;
} damsel_internal_global_mapping;
*/
  push_printf_ndc("DMSLmodel_map_global_handles app_container=%p proc_ranks=%p remote_handles=%p", app_container, proc_ranks, remote_handles);
  damsel_container_internal *app_container_internal = (damsel_container_internal *)app_container;
  damsel_model_internal *model = (damsel_model_internal *) app_container_internal->model;

  damsel_internal_global_mapping *gm = malloc(sizeof(damsel_internal_global_mapping));

  int myrank, maxrank;

  assert(model != NULL);
  assert(model->attached_file != NULL);

  myrank = model->attached_file->myrank;
  maxrank = model->attached_file->maxrank;

  assert_valid_rank(myrank, maxrank);
  #ifndef NDEBUG
    int i;
    int imax = DMSLcontainer_count(app_container);
    for(i=0; i<imax; i++) {
      assert_valid_rank(proc_ranks[i], maxrank);
    }
  #endif

  gm->manage_memory_internally = 0;
  gm->magic = DAMSEL_GLOBAL_MAPPING_MAGIC;
  gm->rank = myrank;

  gm->app_container = app_container;
  DMSLcontainer_take(app_container);
  gm->proc_ranks = proc_ranks;
  gm->remote_handles = remote_handles;
  linkedlist_prepend(&model->global_mappings, gm);
  pop_ndc();
  return DMSL_OK;
}

