/*! \file damsel.h
 * \mainpage damsel Damsel: a Data Model Storage Library for Exascale Science
 *
 * As computational science uses more sophisticated data models to
 * depict physical phenomenon and more sophisticated computer science
 * to efficiently operated on those data models at large scale, the
 * need arises for an I/O library that can likewise easily store and
 * retrieve those models.
 *
 * The API version, not this prototype version, contains additional questions.
 *
 * This is benc's attempt at a working damsel API. When there exists a conflict
 * between ideal-but-underspecified behaviour in the documentaion, and
 * conflicting-but-working behaviour, the latter path should be chosen in this
 * codebase.
 *
 * Function documentation in this file should indicate whether that function is
 * collective or independent.
 * Documentation of handles should indicate the space in which the handle
 * is scoped (for example, the store in which the handle lives), and to the
 * type of damsel object to which that handle refers.
 */

#ifndef __DAMSEL_H
#define __DAMSEL_H

#include "mpi.h"

#ifndef DAMSEL_SINGLETHREAD
#include <pthread.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

#include <stdlib.h>
#include <stdint.h>


/** \name Cpp #defines
 * Cpp #defines for Damsel
 *//**@{*/

#define DAMSEL_HANDLE_CONTAINER_TYPE_SEQUENCE 89323
#define DAMSEL_HANDLE_CONTAINER_TYPE_VECTOR 89340
#define DAMSEL_HANDLE_CONTAINER_TYPE_TREE 89345
#define DAMSEL_STORETYPE_MEMORY 524
#define DAMSEL_STORETYPE_HDF5 1212

/**@}*/

/** \name typedefs
 * typedefs for Damsel
 *//**@{*/

/*! this is a handle to things in a damsel store - it remains consistent
  for the whole lifetime of the store.
  for now I need some handle space to be reserved for damsel internal
  purposes, because I want to use damsel storage mechanisms for some
  damsel internal structures: storing connectivity information as tags.
  not sure what the right way is to accomplish this, but for now, I'll
  say that all negative handles belong to damsel internal use. Those are
  allocated for now by the DAMSEL_HANDLE_INTERNAL_* #defines, and by
  taking a memory pointer and making it negative. both are ick.

  If the storage system is allocating all in-file handles, then this is
  not so much of a problem: it can allocate handles to other pieces of
  damsel just like it allocates them to application data.
*/
	typedef long long int damsel_handle;

/* A datatype identifier. This is not as opaque as other identifiers in the
   API. It must be an integer, and the predefined types must start at the
   bottom, 0, and progress upwards in a single block; runtime defined types
   must appear after the predefined watermark constant. */

	typedef long long int damsel_data_type;


    typedef struct damsel_internal_model_id* damsel_model;
    typedef struct damsel_internal_library_id* damsel_library;
    typedef struct damsel_internal_container_id* damsel_container;
    typedef struct damsel_internal_collection_id* damsel_collection;
    typedef struct damsel_internal_trait_bundle_id* damsel_trait_bundle;

      /* \struct damsel_err_t
       * \brief Type of error fields returned by Damsel.
       *
       * In this prototype, code assumes id=0 means OK and everything
       * else means an error.
       *
       * \ingroup Extern_Structs
       */
    typedef struct {int id;} damsel_err_t;

      /* \struct damsel_request_t
       * \brief Type of objects returned by asynchronous Damsel functions
       * \ingroup Extern_Structs
       */
    typedef struct {
      MPI_Request mpi;
#ifndef DAMSEL_SINGLETHREAD
      pthread_t posix;
#endif
    } damsel_request_t;

      /* \typedef damsel_status_t
       * \brief Type of objects created and returned by Damsel\
       * \ingroup Extern_Structs
       */
    typedef struct {int id;} damsel_status_t;


      /** \brief Private handle pointer type to avoid using void*, to enable type 
          checking 
      */
    typedef damsel_handle * damsel_handle_ptr;

    typedef damsel_container damsel_container_buf_type;
/**@}*/

/** \name Enumerations
 * Enumerations (also typedefs) for Damsel
 *//**@{*/

      /* \enum damsel_handle_type
       * \brief Enumeration defining handle types, used to pass back types as arguments.
       * The present implementation only permits use of HANDLE64.
       * \ingroup Enumerations
       */
    typedef enum {
        DAMSEL_HANDLE_TYPE_HANDLE32 = 0,
        DAMSEL_HANDLE_TYPE_HANDLE64,
        DAMSEL_HANDLE_TYPE_HANDLE128,
        DAMSEL_HANDLE_TYPE_INVALID
    } damsel_handle_type;

      /** \brief types of containers in damsel 
       */

    typedef enum {
        DAMSEL_HANDLE_COLLECTION_TYPE_VECTOR = 0,
        DAMSEL_HANDLE_COLLECTION_TYPE_SET
    } damsel_collection_type;


      /** \brief Enumeration defining entity types 
       */
    typedef enum damsel_entity_type {
        DAMSEL_ENTITY_TYPE_VERTEX = 0,
        DAMSEL_ENTITY_TYPE_EDGE,
        DAMSEL_ENTITY_TYPE_TRI,
        DAMSEL_ENTITY_TYPE_QUAD,
        DAMSEL_ENTITY_TYPE_POLYGON,
        DAMSEL_ENTITY_TYPE_TET,
        DAMSEL_ENTITY_TYPE_PRISM,
        DAMSEL_ENTITY_TYPE_PYRAMID,
        DAMSEL_ENTITY_TYPE_HEX,
        DAMSEL_ENTITY_TYPE_POLYHEDRON,
        DAMSEL_ENTITY_TYPE_UNDEFINED,
        DAMSEL_ENTITY_TYPE_ALL_TYPES
    } damsel_entity_type;

/* Damsel error codes */
    static const damsel_err_t
        DMSL_OK = {0},
        DMSL_ERROR_GENERAL   = {-1},
        DMSL_EINVALLIB       = {-2},   /* Invalid Damsel library */
        DMSL_EUNINITIALIZED  = {-3},   /* Damsel uninitialized */
        DMSL_EHANDLETYPE     = {-4},   /* Invalid Damsel handle type */
        DMSL_EMALLOC         = {-5},   /* Damsel internal malloc failed */
        DMSL_EMODEL          = {-6},   /* Invalid Damsel model */
        DMSL_ECMODE          = {-7},   /* Invalid file create mode */
        DMSL_EOMODE          = {-8},   /* Invalid file open mode */
        DMSL_EINVALCONTAINER = {-9},   /* Invalid container handle */
        DMSL_EINVALTAG       = {-10},  /* Invalid tag */
	DMSL_ENOTFOUND       = {-11}   /* Tag not found */
        ;


      /* \enum damsel_transfer_type
       * \brief types of transfers.
       * \ingroup Enumerations
       */
    typedef enum {
        DAMSEL_TRANSFER_TYPE_INVALID = 0,
        DAMSEL_TRANSFER_TYPE_READ,
        DAMSEL_TRANSFER_TYPE_WRITE
    } damsel_transfer_type;

/**@}*/


/** \name Buff types
 * Buff types for reading header-type information
 *//**@{*/

/*! \brief Tag info read from file. */
    typedef struct damsel_tag_buf_type {
      char *name;
        /**< tag name */
      damsel_data_type tag_datatype;
        /**< tag data type, e.g. double, int, etc. */
      damsel_handle_ptr tag_handle;
        /**< tag handle, in the file's handle space */
    } damsel_tag_buf_type;


/*! \brief Collection info read from file. */
    typedef struct damsel_collection_buf_type {
      damsel_collection_type type;        /**< type of this collection */
      damsel_handle_ptr collection_handle;
        /**< The handle to the collection, in the file's handle space. */
      size_t tag_count;
        /**< number of tags in tag_handle_container -
           redundant as tag_handle_container knows how many handles it
           contains. */
      damsel_container tag_handle_container;
        /**< A container of file-side tag handles that have values defined on
           this collection. See the discussion on damsel_entity_buf_type
           about the questionable utility and vagueness of this field. */
      damsel_container contents;
        /**< The handles stored in this container. */
    } damsel_collection_buf_type;

/** \brief Entity information as returned by DMSLmodel_get_entity_infos.
 */
    typedef struct damsel_entity_buf_type {

      enum damsel_entity_type entity_type;
        /**< entity type */

      size_t vertices_per_entity;
        /**< vertices_per_entity ( = 1 for vertex types) */

      size_t count;
        /**< number of entities in this entity definition. This is
           redundant wrt entity_container, as entity_container knows
           how many handles it contains. So perhaps it should be
           removed. */

      damsel_container entity_container;
        /**< A container of the file-side handles of the entities that this
           entity definition applies to. entity_container must have
           'count' handles in it. */

      damsel_container vertex_container;
        /**< A container of the vertex file-side handles of each entity
           in entity_container; there should be
           (vertices_per_entity * count) handles in
           this container.*/

      size_t tag_count;
        /**< number of tags in tag handle collection (i.e. tags that are on these entities).
           This is redundant as tag_handle_container knows its own size.
           So it should probably be removed. */

      damsel_container tag_handle_container;
        /**< handle container containing file-side handles to tags that have
           values defined on these entities. This is not implemented at the
           moment.

           It seems awkward: tags can be defined on arbitrary
           sets of handles and there is nothing that requires that if a tag
           is defined on one handle in the entity_container then it is
           defined on all handles in that container.

           So, for example, we can't feed that entity_container directly
           into a map call to get tag values.
        */
    } damsel_entity_buf_type;

	/* struct of vector data elements in memory */
    typedef struct {
      size_t len;  /* lenght of the data array             */
      void   *p;   /* Pointer to array of data elements    */
    } damsel_vector_t;

/**@}*/

/** \name externs
 * extern declarations
 *//**@{*/

    extern damsel_container DAMSEL_CONTAINER_INVALID;

    extern damsel_model DAMSEL_MODEL_INVALID;

      /* An invalid library identifier */
    extern damsel_library DAMSEL_LIBRARY_INVALID;

/**@}*/

/** \name DMSLlib
 * Library-level stuff
 *//**@{*/

      /* \brief initialize Damsel library.
       */
    damsel_err_t DMSLlib_init(damsel_library *lib);

      /* \brief terminate a Damsel library instance.
       */
    damsel_err_t DMSLlib_finalize(damsel_library library_id
                                    /**< library instance being terminated */
                                  );

/**@}*/

/** \name DMSLmodel
    Creating and closing a model
    @{*/
    damsel_err_t DMSLmodel_create(damsel_handle_type  app_handle_type,
                                  damsel_model       *model
                                    /**< [in] Application-space Handle type to be used with this model */
                                  );

/* \brief attach model to existing file.
 * This implementation uses damsel_store *, rather than a model_id,
 * pending implementation of numeric model IDs later on.
 */
    damsel_err_t DMSLmodel_attach(damsel_model model_id,
                                    /**< model being attached to file */
                                  const char *filename,
                                    /**< file to which we're attaching */
                                  MPI_Comm comm,
                                    /**< MPI communicator */
                                  damsel_trait_bundle trait
                                    /**< traits, NULL if none */
                                  );

    damsel_err_t DMSLmodel_close(damsel_model);

      /** \brief populate the model with information from the file
       * populates the tuple information in the model description, metadata only, not "raw" data 
       */
    damsel_err_t DMSLmodel_populate(damsel_model model_id
                                      /**< model being attached to file */
                                    );

    /** maps a buffer of data to be written on some handles on a tag.

        Is it an error for this data to never be written out (for example, if
        it ends up not being further mapped into storage) or is it permitted
        (albeit wasteful?)

        this call will allocate memory which is released when the transfer takes place.
        the mapping will also be removed at the point that the transfer takes places.

        this is an independent operation (in the MPI sense)
   */
    damsel_err_t DMSLmodel_map_tag(void * const tag_buffer,
                                     /**< application memory storing tag values for app_handles */
                                   damsel_container app_handles,
                                     /**< application handle container for which tag is being mapped */
                                   damsel_handle_ptr tag_handle
                                   /**< application-side tag handle whose values are being mapped */);

/* I'm assuming here that the app_container handle space is what used to be
   "memory store" handle space - which the application can allocate as it
   chooses to, and which is sometimes in my examples a pointer, and sometimes
   an integer; and that model_container handle space is the same as
   what used to be file-side handle space...
   I'm not sure if that's the correct understanding, though.

   mappings created with this call (and the wrappers to this call) will exist
   until the model is closed.
*/
    damsel_err_t DMSLmodel_map_handles(damsel_container app_container, damsel_container model_container);

/*! this maps handles from memory side into the file, allowing damsel to choose its own file-side handles. */
    damsel_err_t DMSLmodel_map_handles_inventing_file_handles(damsel_container app_container);

/*! this maps handles from memory side into the file, allowing damsel to
    choose its own file-side handles, but with the permutation of app-side
    data chosen by the supplied callback.
*/
    damsel_err_t DMSLmodel_map_handles_inventing_file_handles_with_permute(damsel_container app_container, void *permute_callback);

/*! maps a single handle into the file-side */
    damsel_err_t DMSLmodel_map_single_handle_inventing_file_handle(damsel_model m, damsel_handle h);

      /** \brief map local application-space handles to application-space handles on other processors
       * A given entity can be in \ref app_container multiple times, for cases where it is shared with
       * multiple other processors.  Length of \ref remote_handles is number of handles in \ref app_container.
       *
       * The space referred to by proc_ranks and remote_handles must not
       * be released until much later - after the model has been closed,
       * in this implementation.
       *
       * mappings defined in this way are not transitive: if a handle on
       * rank A is mapped to a handle on rank B, and that handle on rank
       * rank B is mapped to a further handle on rank C, then operations
       * on the A handle will not be connected with operations on the B
       * handle. This comes from an implementation decision to keep the
       * resolution algorithm simple. It could be replaced by a fancier
       * implementation to remove this restriction, probably.
       */
    damsel_err_t DMSLmodel_map_global_handles(damsel_container app_container,
                                                /**< container with application-space handles for entities being mapped */
                                              int *proc_ranks,
                                                /**< processor ranks of handles in remote_handles */
                                              damsel_handle_ptr remote_handles
                                                /**< vector of application-space handles on remote procs */
                                              );


      /* \brief "transfer" from file to application 
       *
       * The present implementation has a read/write distinction which applies
       * to the whole model when opened, so its a bit late to specify it here.
       * Reading and writing in the same model needs to be better specified
       * some time, and the implementation changed to reflect that better
       * specification.
       */
    damsel_err_t DMSLmodel_transfer_async(
        damsel_model model_id,
          /**< model being queried */
        damsel_transfer_type transfer_type,
          /**< type of transfer */
        damsel_request_t *transfer_request
          /**< request object for the asynchronous operation */
                                          );


      /* \brief wait for a request to complete.
       */
    damsel_err_t DMSLmodel_wait(damsel_request_t transfer_request,
                                  /**< request to wait for */
                                damsel_status_t *request_status
                                  /**< status of the request */
                                );


    damsel_err_t DMSLmodel_transfer_sync(damsel_model model_id,
                                         damsel_transfer_type transfer_type);

/** \name Read transactions
    In some places, it has been necessary to read some data from the file in
    order to decide what else should be read from the file. These 'read
    transactions' facilitate that.*/

/* start a new read transaction */
    void DMSLmodel_start_read(damsel_model m);

/* stop the current read transaction, dropping us back into the transaction
   before */
    void DMSLmodel_stop_read(damsel_model m);

/*! \brief returns information about all containers known in the attached file.

  the number of entries in container_buf is returned as the container_count
  value from DMSLget_tuple_count. */
    damsel_err_t DMSLmodel_get_container_infos(damsel_model model_id,
                                               damsel_container_buf_type *container_buf);

/** \brief get tag information in model.

    tag_buf must be big enough to accomodate as many damsel_tag_buf_type
    structs as returned by get_tuple_count.
*/
    damsel_err_t DMSLmodel_get_tag_infos(damsel_model model_id,
                                           /**< model being queried */
                                         damsel_tag_buf_type *tag_buf
                                           /**< vector of tag buffers, of length
                                              retrieved by DMSLget_info_count */
                                         );

/*! populates the supplied buffer with information about entity-definitions.
  buf must be big enough to hold the number of entity-definitions returned
  by DMSLentity_get_count().

  Information about the information returned is documented on the
  damsel_entity_buf_type structure.
*/
    damsel_err_t DMSLmodel_get_entity_infos(damsel_model m,
                                            damsel_entity_buf_type *buf);

/*!
 * \brief returns the count of the number of entity-definitions in the model
 */
    int DMSLmodel_get_entity_count(damsel_model m);

/*! \brief returns count of collections known in the attached file */
    int DMSLmodel_get_collection_count(damsel_model model_id);

/*! \brief returns information about all collections known in the attached file. */
    damsel_err_t DMSLmodel_get_collection_infos(damsel_model model_id,
                                                damsel_collection_buf_type *collection_buf);

/*! closes a collection infos structure, so that damsel can release holds on
    internal memory used in the structure. */
    damsel_err_t DMSLmodel_close_collection_infos(damsel_model model_id, damsel_collection_buf_type *collection_buf);

/*! \brief returns a count of containers (not collections) and tags in the
  attached file.

  the API document specifies that this will return counts of entities and
  collections; at present those are implemented as other calls.

  if a value is not needed, then NULL can be passed in as the relevant
  pointer.
*/
    void DMSLmodel_get_tuple_count(damsel_model, int *container_count, int *tag_count);

/* @} */

/** \name DMSLcontainer
    Functions that create containers of handles.
    **@{*/

/* !This will register a handle sequence starting at baseh, and name it
   with handle collh.

   This is an independent operation, not a collective operation.

   The parameters are (in order):
   the store
   the quantity (the number of handles in this container)
   the base handle
   the step
*/

    damsel_err_t DMSLcontainer_create_sequence(damsel_model store, damsel_handle baseh, unsigned quantity, size_t step, damsel_container *container);

/* !This will register a vector of handles as a handle container.

   This is an independent operation, not a collective operation.
*/

    damsel_err_t DMSLcontainer_create_vector(damsel_model store, 
	    /**< model in which container being created */
	    damsel_handle *contents, 
	    /**< an array of handles. Damsel will copy the contents, so caller
	     * may free memory as soon as function returns.  create_vector is a
	     * special exemption to the "do not copy" rule due to the inability
	     * to represent an arbitrary list of handles algorithmicly  */
	    unsigned quantity,
	    /**< number of items in 'contents' array */
            damsel_container *container);

/* there seems to be some difference between a container and a container
   that I don't fully understand. so for now I'll implement a container as
   a specific type of container... */

    damsel_container DMSLcontainer_tree_create(damsel_model model_id,
                                                 /**< model being queried */
                                               damsel_collection_type coll_type
                                                 /**< container type requested */
                                               );

      /** \brief add entities to an entity container from a handle container
       */
    damsel_err_t DMSLcontainer_tree_add(damsel_container model_container_id,
                                          /**< collection */
                                        damsel_container app_container
                                          /**< container holding handles to be added to container, in app handle space */
                                        );

      /** \brief add entities to an entity container from a handle vector. The array of
          handles passed in must remain allocated until some time in the future when the
          container is no longer needed. (which is not properly defined in this API at the
          moment, but is at least after all transfers have happened)
       */
    damsel_err_t DMSLcontainer_tree_add_fast(damsel_container model_container_id,
                                               /**< container id, in model handle space */
                                             damsel_handle_ptr app_handles,
                                               /**< vector of handles to be added to container, in app handle space */
                                             size_t count
                                               /**< number of handles in vector */
                                             );

/*! \brief Return the type of container (sequence, vector, or tree) */
    int  DMSLcontainer_get_type(damsel_container c
                                  /**< container for which type is requested */
                                );

/*! \brief closes the specified container. exactly what this means is unclear to benc
    at the moment. only containers that have been explicitly created by the application
    using container_create_* calls should be closed in this way. there maybe needs to be
    a way for the app to indicate that it has released other containers returned by the
    library that might be shared elsewhere and so need to not always be released when
    the application has finished with them. */
damsel_err_t DMSLcontainer_close(damsel_container c);

/*! indicates that a reference to the container is no longer used. for internally
    created containers, this might trigger cleanup. */
damsel_err_t DMSLcontainer_release(damsel_container c);

/*! takes a reference on a container */
damsel_err_t DMSLcontainer_take(damsel_container c);

/*! \brief returns the handle at position i in container c */
    damsel_handle DMSLcontainer_handle_at_position(damsel_container c, int i);

/*! \brief returns the number of handles in a container */
    int DMSLcontainer_count(damsel_container c);

/*! \brief get contents of sequence container */
    damsel_err_t DMSLcontainer_sequence_get_contents(damsel_model model_id,
                                                     /**< model id */
                                                   damsel_container seq_container_id,
                                                     /**< sequence container id */
                                                   damsel_handle_ptr handle_ptr,
                                                     /**< returned start handle of container */
                                                   size_t *count,
                                                     /**< number of handles in container */
                                                   size_t *increment
                                                     /**< increment of container */
                                                   );
    
/*! \brief get contents of vector container */
    damsel_err_t DMSLcontainer_vector_get_contents(damsel_model model_id,
                                                     /**< model id */
                                                   damsel_container vec_container_id,
                                                     /**< vector container id */
                                                   damsel_handle_ptr * const handle_ptr_ptr,
                                                     /**< pointer to pointer to handles in this vector */
                                                   size_t *count
                                                     /**< number of handles in container */
                                                   );
    
/*! \brief get contents of tree container 
 * If NULL is passed in for current container, returns first container in tree; for subsequent
 * calls, returns container after currently pointed-to container; returns NULL if there are no
 * more containers in tree, or (if NULL is passed in) if tree is empty.
 */
    damsel_err_t DMSLcontainer_tree_get_contents(damsel_model model_id,
                                                   /**< model id */
                                                 damsel_container vec_container_id,
                                                   /**< vector container id */
                                                 void *hc_node,
                                                   /**< the node in the tree */
                                                 damsel_container *container_ptr
                                                   /**< pointer to current/next container in tree */
                                                 );
    
                                                   

/**@}*/

/** \name Tags
    @{ */

/*! define a tag. is this collective or independent? benc's implementation
  assumes that this is a collective operation.
  the parameters are: store to create the tag in; handle for the new tag;
  name of the tag; type of the tag (in the present impl, proxied by the
  size of the type); and storage mode, which takes values from the
  DAMSEL_STORAGE_MODE_* #defines.

  the tag_handle must be mapped into the file on each rank.

  it is unclear how the collectiveness of this operation should interact
  with each rank having its own handle space, and in future how that
  interacts with global handle mapping.

  must the tag handle be the same on all ranks? can
  it be different on each rank? can a tag handle used on one rank then be
  reused on another rank? does an explicit mapping need to be made using
  map_global_handles to tie the tag together across ranks?

  defined tags will occupy memory until the model is closed. they do not need
  any other specific closing.

  the store_tags code assumes that tags have been defined on all ranks, so
  DMSLtag_define needs to create the same tags on all ranks.

*/
damsel_err_t DMSLtag_define(damsel_model model_id, damsel_handle_ptr tag_handle, damsel_data_type type, const char *);


/* @} */

/** \name Mapping between handle spaces
    Functions for mapping between handles in the application and handles in
    the attached file.
    @{
*/

/*!it seems unclear whether this should be a description of an action to
  happen once, or a "permanent mapping" in the sense that 
  map(a->b), a=3; flush(); a=7; flush(); should write out that 7...
  if thats the case, then:
  i) do we write out all known maps every time? in which case, need to
  explicitly remove maps. [A]
  ii) have dirty flags (for what? a range of handles? a particular mapping?)
  [B]
  if that's not the case, then if we want to write out the same handles
  to the same handles again, we have to call store_map again. [C]
  benc's taste is for option C but its not a particularly strong preference

  do I have to define tags on both stores in order to get the tag values to
  be mapped? (so that I can choose which tags get stored or not - for example,
  I might not be interested in storing temperature in a particular file
  container even though I've got a tag for it in my memory store, because
  I only want to visualise pressure) for now, I'll assume yes.

  the purpose of DMSLmap_handles is to map from one handle to
  another handle, regardless of their membership in containers.  The
  results of calling this function once with two N-sized handle
  containers should be the same as calling N times with containers
  holding one handle each.

*/

// void DMSLstore_map(damsel_store *, damsel_handle srcoll,
//                      damsel_store *, damsel_handle destcoll);

/*@}*/

/** \name Execution of operations
    Functions that cause Damsel to execute its queued operations
    **@{*/


/*!question on the semantics of what we are flushing here? are we flushing
  the memory store? if so, does that mean we flush everything that was
  mapped to any other store? can there be more than one relationship between
  stores? (in general, yes because one file store over 64k ranks, and 64k
  memory stores) 

  but then we're callign damsel flush with different memory stores on each
  rank - rather than calling damsel flush with a single file store on all
  ranks. I guess that means that we should be passing in the file store here,
  *not* the memory store.

  that also means its more a pull than a push... (some one used 'push' on 
  the list)

  what about with reads? do they need to then be on the memory store or what?
  perhaps a flush on the memory store will flush the reads and writes on that
  particular memory store so not be a global operation?

  in the context of reads (where read means a map from a file store to a 
  memory store, "flush" seems a weird name too. perhaps 'perform' is
  better. I suppose I shouldn't spend too long deciding a name ;)

  execute was also suggested, and I guess i like that better.

  From discussion on the mailing list:

  If DMSLtransfer_async is invoked and the destination is a new file,
  and no explicit mappings have been defined, then all the contents
  of the source file should be stored there.

  DMSLexecute is a collective operation.

*/

      /**@}*/

/** \name Getting metadata
    Functions that will return metadata. These are only usable in read mode
    (where 'read mode' is not even properly defined by this API!).

    The functions in this section are a little fragmentary and could probably
    be made more coherent. The focus so far has been on providing information
    rather than keeping a tidy API.

    In general, information is available about containers, tags,
    collections, and entity-definitions; first as a 'get count' function and
    then as a 'get all the metadata' function.

    **@{*/

      /**@}*/

/** \name Defining entities
    Functions for defining entities
    @{ */

/*! declares that handles in a handle container are handles to geometric
  entities. it specifies the type. for now, the vertices-per-entity parameter
  is ignored; but it is intended that when the type has variable number of
  vertices, then this will be used. the API needs to define what happens
  to this parameter when the type fixes the number of vertices - for
  example, that this parameter is ignored.

  this is an independent operation, not a collective operation.

  entity types come from DMSL_ENTITY_TYPE_*

*/

    damsel_err_t DMSLentity_define(
        damsel_container app_container,
          /**< a container with the application-space handles for the entities */
        damsel_entity_type ent_type,
          /**< entity types of entities being defined */
        size_t vertices_per_entity,
          /**< number of vertices per entity */
        damsel_container vertex_container_id
          /**< container of application-space vertex handles */
                                   );


    damsel_err_t DMSLentity_define_fast(
        damsel_container app_container,
          /**< a container with the application-space handles for the entities */
        damsel_entity_type ent_type,
          /**< entity types of entities being defined */
        size_t vertices_per_entity,
          /**< number of vertices per entity */
        damsel_handle_ptr vertex_handles
          /**< array of application-space vertex handles */
                                        );

/* @} */

/** \name Selecting handles
    Functions that will select handles from an existing file based
    on various criteria.
    @{ */

/*! returns the handle of the specified tag. for now, I'm only going to
  implement this on the hdf5 file store, as memory store handles are
  application specified and so we can expect the user to already
  know them.

  how can I assert failure when there's no handle to return?
  need to return both a status code and a handle...
*/

    damsel_err_t DMSLselect_tag_by_name(damsel_model model, const char *tagname, damsel_handle *found_handle);

/*! creates a handle container that consists of the handles which have
  tag values on the specified tag in the specified range, in the same
  order as the range.
  At the moment, only one specific numeric type of tag value will be
  supported, I think, for ease of implementation: short int
  both start and end are included in the handle container.
*/


    damsel_container DMSLselect_range(damsel_model model,
                                      damsel_handle index_tag, short int start, short int end);


/*! creates a container of all the handles which have a value set on the
  specified tag. there is no specific order that these handles will be
  returned in, so some other mechanism must be used by the application if
  they are desired to be in some order. */

    damsel_container DMSLselect_handles_with_values(damsel_model m,
                                                    damsel_handle tag_handle);

      /**@}*/


/** \name Damsel data types
    All about datatypes
    **@{*/
	typedef enum {
		DAMSEL_DATA_TYPE_INVALID = 0,
        DAMSEL_DATA_TYPE_BYTES = 1,
        DAMSEL_DATA_TYPE_INTEGER = 2,
        DAMSEL_DATA_TYPE_INT64 = 3,
        DAMSEL_DATA_TYPE_FLOAT = 4,
        DAMSEL_DATA_TYPE_DOUBLE = 5,
        DAMSEL_DATA_TYPE_HANDLE = 6,
        DAMSEL_DATA_TYPE_SHORTINT = 7,
        DAMSEL_DATA_TYPE_PREDEFINED_WATERMARK = 9
	} damsel_data_type_predefined;

/*! creates an opaque type, of specified byte count. this
  must be run on all ranks at once to ensure internal ID numbering matches
  up across ranks. How does this interact with the fact that we can do stuff
  over lots of different communicators, so "all ranks" isn't clearly defined
  here? */
    damsel_data_type DMSLtype_opaque_create(size_t bytecount);

/** \brief Create array type
 */
    damsel_data_type DMSLtype_array_create(
        damsel_data_type base,              /**< Base type for array */
        unsigned ndims,                     /**< Number of dimensions for array */
        const size_t dims[/* ndims */]      /**< Size of array's dimensions */
                                           );


/**
Creates a vector data type.
    
A vector contains data elements of damsel_vector_t, which can be an array 
of any size. Each element in a vector may have different size.

@param  base -- the base type for the vector type
@param  max_len -- HDF5 parallel IO does not support writing variable-length
                      data. For a temporary solution, we will use fixed length
                      array type of size max_len.
					  
@return a datatype identifier if successful; otherwise returns a negative value.
*/
damsel_data_type DMSLtype_vector_create(damsel_data_type base, size_t max_len);

/**
Creates an enum data type.
    
An enum type contains members of damsel_enum_member_t, which defines the 
<name, value> pairs of enum members.

@param  nmembs -- number of members of the enum type
@param  names  -- names of the enum type
@param  values -- values of the enum type. If values is NULL, enum values start with 0.

@return a datatype identifier if successful; otherwise returns a negative value. 
*/
damsel_data_type DMSLtype_enum_create(unsigned nmembs, char **names, int *values);

/**
Creates a struct data type.
    
A struct type contains members of damsel_struct_member_t, which defines the 
the members of the struct. A struct member has three fields: name, offset, and type.

@param  nmembs  -- number of members of the struct type
@param  names   -- names of the struct type
@param  types   -- types of the struct type
@param  offsets -- offsets of the struct members
@param  tsize   -- total size of the structure type

@return a datatype identifier if successful; otherwise returns a negative value.
*/
damsel_data_type DMSLtype_struct_create(unsigned nmembs, char **names, 
    damsel_data_type *types, size_t *offsets, size_t tsize);
	
/**@}*/

/** \name DMSLcoll
    All about collections
    **@{*/

/*! creates a collection containing the handles in the specified container,
  and with the specified handle. */
    damsel_collection DMSLcoll_create(damsel_model model_id,
                                      damsel_handle_ptr app_handle_ptr,
                                      damsel_container container_id,
                                      damsel_collection_type coll_type);

/**@}*/

/* layout rules */
void damsel_permute_id(damsel_container appside, damsel_container fileside);
void damsel_permute_connectivity(damsel_container appside, damsel_container fileside);

/* traits */
damsel_err_t DMSLtrait_create(damsel_trait_bundle *tbp);
damsel_err_t DMSLtrait_release(damsel_trait_bundle);
damsel_err_t DMSLtrait_take(damsel_trait_bundle tb);
damsel_err_t DMSLtrait_put(damsel_trait_bundle tb, char *key, char *value);
damsel_err_t DMSLtrait_get(damsel_trait_bundle tb, char *key, char **value);

const char* DMSLerror_str(damsel_err_t err);

#ifdef __cplusplus
}
#endif


#endif
