
/* 2d temperature experiment */

#include <assert.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include "damsel.h"
#include "damsel-internal.h"
#include "damsel-debug.h"

#include "damsel-ndc.h"

#define FILENAME "output/test-proto2d.h5"

damsel_handle x_tag_handle = NULL;
damsel_handle y_tag_handle = NULL;
damsel_handle temperature_tag_handle = NULL;
damsel_handle sfc_tag_handle = NULL;
damsel_handle tag_container[4] = {NULL, NULL, NULL, NULL};
damsel_handle rank_zero_tag_container[4];

/* we'll have a sheet of magic substrate, XLEN wide and YLEN long. The
   x-axis will be divided by the number of ranks.
   XLEN must be a multiple of the number of ranks - 12 has nice factors
   There are no constraints on the factors of YLEN.
*/
#define XLEN 12
#define YLEN 7

/* so here's our application storage for the substrate - XLEN*YLEN doubles
   in C array order. We don't need all XLEN columns except when numranks=0;
   instead, we'll use the first cols_on_this_rank columns on each rank. */
double rod[XLEN][YLEN];

/* these need initialising right after MPI_Init */
int myrank;
int numranks;
int cols_on_this_rank; // if XLEN is not divisible by number of ranks,
                       //  this won't be the same on each rank.
int smallest_x; // this is the X value for the leftmost (smallest) column

int rank_zeros[4] = {0,0,0,0};

/* temperature is 50 everywhere except in the 0,0 corner. */
void init_sim() {
  int x,y;
  printf("initializing new simulation.\n");
  for(x=0; x<cols_on_this_rank; x++) {
    for(y=0; y<YLEN; y++) {
      assert(x<XLEN);
      assert(x<cols_on_this_rank);
      assert(y<YLEN);
      rod[x][y] = 50;
    }
  }
  rod[0][0] = 99;
}

/* this is a collective operation to print - everyone outputs their
   own bit and flushes */
void print_sim() {
  int x,y;
  printf("=====\n");
  for(y=0; y<YLEN; y++) {
    printf("* ");
    for(x=0; x<XLEN; x++) {
      if(x>=smallest_x && x<(smallest_x+cols_on_this_rank))
        printf("%.2f ", rod[x-smallest_x][y]);
    }
    printf("\n");
  }
  printf("=====\n");
}

void step_sim() {
  double newstate[XLEN][YLEN];
  int x,y;
  for(x=0;x<cols_on_this_rank;x++) {
    newstate[x][0] = rod[x][0];
    newstate[x][YLEN-1] = rod[x][YLEN-1];
  }
  for(y=0;y<YLEN;y++) {
    newstate[0][y] = rod[0][y];
    newstate[cols_on_this_rank-1][y] = rod[cols_on_this_rank-1][y];
  }
  for(x=1; x<cols_on_this_rank-1; x++) {
    for(y=1; y<YLEN-1; y++) {
      newstate[x][y] =(
                rod[x-1][y-1] + rod[x][y-1] + rod[x+1][y-1] +
                rod[x-1][y] +   rod[x][y]   + rod[x+1][y] +
                rod[x-1][y+1] + rod[x][y+1] + rod[x+1][y+1]
                      )/ 9;
    }
  }

  for(x=0; x<cols_on_this_rank; x++) {
    for(y=0; y<YLEN; y++) {
      rod[x][y] = newstate[x][y];
    }
  }
}

void dump_sim() {
  printf("dumping\n");

  // initialise in-memory damsel store that maps from other memory
  // but does not contain a copy of that memory (so no mutating of
  // input memory 'till flush)
  damsel_model d_model;
  damsel_err_t err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &d_model);
  damsel_store_internal *model = (damsel_store_internal*)d_model;
  damsel_model_internal *m_model = (damsel_model_internal *)model;

  // create 'LEN' number of handles, giving a handle container

  // this is DMSLcontainer_create_sequence in the API doc - but its not
  // creating a sequence as much as registering a sequence that was
  // created by the simulation, and making that sequence form a handle
  // container.
  // Also, I changed the parameters a bit.
  damsel_container vertex_container_id;
  err = DMSLcontainer_create_sequence((damsel_model)model,
    (damsel_handle)rod, cols_on_this_rank*YLEN, sizeof(double),
    &vertex_container_id);

  // create a tag for co-ordinates
  // the API has a handle for this tag - so do we use that or the handle?
  // I'm using pointers for my handles. what handle can I use in this case,
  // where the tag doesn't have an in-memory representation in my simulation
  // for me to key off?
  // for now, I can get one easily like this: (ICK!)


  /* I'd like to store these as a single tag where each value is a pair
     of doubles. but I'd need to implement stuff with fancier
     datatypes first */
  assert_is_damsel_type(DAMSEL_DATA_TYPE_SHORTINT);
  DMSLtag_define((damsel_model)model, &x_tag_handle, DAMSEL_DATA_TYPE_SHORTINT, "xcoords");
  DMSLtag_define((damsel_model)model, &y_tag_handle, DAMSEL_DATA_TYPE_SHORTINT, "ycoords");

  // create a tag for the 1-d index / space filling curve that we will use
  // for partitioning at read time.
  DMSLtag_define((damsel_model)model, &sfc_tag_handle, DAMSEL_DATA_TYPE_SHORTINT, "sfc");

  short int xposns[XLEN][YLEN];
  short int yposns[XLEN][YLEN];
  short int sfc[XLEN*YLEN];

  int x,y;
  for(x=0;x<cols_on_this_rank; x++) {
    for(y=0;y<YLEN; y++) {
      xposns[x][y] = (short int) (100 * ((float)x)/((float)XLEN));
      yposns[x][y] = (short int) (100 * ((float)y)/((float)YLEN));
      sfc[x*YLEN+y] = smallest_x*YLEN + x*YLEN+y;
    }
  }

  assert_valid_handle_container_id(vertex_container_id);
  DMSLmodel_map_tag(xposns, vertex_container_id, &x_tag_handle);
  DMSLmodel_map_tag(yposns, vertex_container_id, &y_tag_handle);
  DMSLmodel_map_tag(sfc, vertex_container_id, &sfc_tag_handle);


  DMSLtag_define((damsel_model)model, &temperature_tag_handle, DAMSEL_DATA_TYPE_DOUBLE, "temperature");

  // set the handle values for temperature tag to temperatures.


/* this tag_set and the layout of handles needs to be aware of the
   topology/geometry of the simulation.
   I want to have file-side handles that are labelled as per the X/Y
   structure of the simulation as a whole, scanning across X as the least
   significant component. (there are various encoding schemes).
   So when I write out the rod array, I need to be writing out rows of it
   to separate chunks of handle (essentially one row at a time) - so
   'strided'.
   What API description will I have for that? Will I make a fancy file-side
   handle container, that is more than a straight sequence of handles?
   Then the geometry is encoded in that - the full geometry doesn't need to
   be declared for successful storage - just the striding pattern. Although
   it might be useful elsewhere to know the real geometry.

   actually, not so in this case - all the columns in a rank are stored
   contiguously in memory... I can make a similarly laid out in-file
   handle structure...
*/
  DMSLmodel_map_tag(rod, vertex_container_id, &temperature_tag_handle);

/*
ok this is the more vague bit...

  // map memory store to file store -- some kind of mapping of handles?
  // sync, flush, close etc.
*/

  // damsel_store *filestore;
  unlink(FILENAME); // maybe it doesn't exist, so we'll just silently
  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "WRITE");
#ifdef DAMSEL_SUBFILE
  //DMSLtrait_put(tb, "dmsl_num_subfiles", "2");
#endif

  DMSLmodel_attach((damsel_model)model, FILENAME,MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);
  assert(m_model->attached_file != NULL);

// TODO: these tag definitions need to be replaced by some mapping/pushing
// of the equivalently named tags above.
// how does mapping need to change to accomodate this?
// should I make a container of tags, and then map that container?
// which means when I am going through mapped stuff in the write routine,
// I need to not just look for values set in tags on those handles, but
// also for the list of tags to map?
// (note that some handles might have values set on them and also still be
// tags - because its possible to set a tag value on a tag.
// so this is not either/or behaviour - we need to find out which tags
// are mapped; and separately keep the value preserving behaviour
// )

/* I'll make a handle container of the three tag handles and map
   that. That will be the indication that I want that tag to appear
   in the destination store.
   Now, what does that mean when the tag is defined in multiple
   places? For example, its defined on each rank...
   Maybe I should be defining tags collectively in the global store?
   But that makes tag handles global space, which means my tag
   storage call will take  a global handle for one param, and a
   non-global call for other params...
   or I make tag define have both a global handle and a local handle,
   (where the local handle can be different everywhere, but the
   global handle must be consistent)
   hmm.

   I'm going to map the tags to the same handle in the filestore, I think.
   (is that what I did before? I think so... which is out of line with other
   ways in which I created handles, as small-magnitude integers. but I think I
   ended up not storing tag handles in the file anyway)
*/

  damsel_container tag_container_id;
  err = DMSLcontainer_create_vector(
    (damsel_model)model, tag_container, 4, &tag_container_id);

  damsel_container file_tag_container_id = DAMSEL_CONTAINER_INVALID;
  if(myrank == 0) {
    err = DMSLcontainer_create_vector((damsel_model)m_model->attached_file, tag_container, /* count = */ 4, &file_tag_container_id);

    DMSLmodel_map_handles(tag_container_id, file_tag_container_id);
  }

// TODO: I think I need to do global mapping, to tie up all the tag handles across ranks
// - will that cause all ranks to then write out their data? maybe? if not the library
// probably should issue a warning at least, that some data is unused/unwritten.
// might need to establish global handle mapping for all the data points too? hopefully
// not... in this case that would mean allocating memory for all the data points on one
// rank in order to get them memory (because that's how I'm doing things).

// each rank allocates its tag handles dynamically, so I can't statically map eg
// x_tag_handle to the same handle on each rank.

// instead, rank 0 will broadcast out its 4 tag handles, in the tag_container array, and
// all the other ranks will receive and map appropriately.

  if(myrank == 0) {
    // TODO: braodcast tag_container
    MPI_Bcast(tag_container, 4, MPI_LONG_LONG, 0, MPI_COMM_WORLD);
  } else {
    MPI_Bcast(rank_zero_tag_container, 4, MPI_LONG_LONG, 0, MPI_COMM_WORLD);
    DMSLmodel_map_global_handles(tag_container_id, rank_zeros, rank_zero_tag_container);
  }


/* so lets make this handle container more fancy - rather than a single
   container, I'm going to instead register a new container for every
   row fragment, and then map each row fragment separately.
   This is not the way to do it in the longer term but it will let me flesh
   out the structures I have already rather than building new ones.
*/

/* how am I going to lay out file-side handle space?
   i need handles for the vertex containers. instead of one per rank, I now
   need YLEN per rank.

   start (arbitrarily at 0x20) for filestore handles

   then start handles of points at 0x20 + total_ranks
     with rows first, so point x,y has handle:
                                   0x20 + total_ranks + x*YLEN + y

*/

  //damsel_handle filestore_vertex_container = 0x20 + myrank;

// total ranks is number of MPI ranks
// smallest_x is column number of the first column of our simulation that 
// we have on this rank
// it can't be sequential after the 0x20 handles above because thats based
// on the number of executing ranks, which can vary... so I assume 0x9000 here
// as being big enough
  // damsel_handle filestore_vertex_base = 0x9000 + smallest_x*YLEN;

  //damsel_container filestore_vertex_container_id;
  //err = DMSLcontainer_create_sequence_with_handle((damsel_model)m_model->attached_file, filestore_vertex_container,
  //  YLEN*(cols_on_this_rank), filestore_vertex_base, 1, &filestore_vertex_container_id);

  DMSLmodel_map_handles_inventing_file_handles(vertex_container_id);

  damsel_request_t req;
  DMSLmodel_transfer_async((damsel_model)model, DAMSEL_TRANSFER_TYPE_WRITE, &req);

  printf("waiting for damsel MPI request to complete\n");
  damsel_status_t status;
  DMSLmodel_wait(req, &status);
  printf("Cmpleted damsel MPI request\n");

  printf("rank %d closing...\n", myrank);

  DMSLmodel_close((damsel_model)model);
  DMSLcontainer_close(vertex_container_id);

  if(myrank == 0) {

    DMSLcontainer_close(tag_container_id);
    DMSLcontainer_close(file_tag_container_id);
  }

  printf("rank %d closed\n", myrank);

  printf("done dumping\n");
}

void reload_sim() {
  damsel_model model;
  damsel_err_t err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "READ");

  DMSLmodel_attach(model,FILENAME, MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);

  DMSLmodel_populate(model);

  /* my assumption is that tags are already defined in the filestore, as
     will be handle containers later; but that I need to create them in
     the memory store. */


  /* the following two define a tag and a vertex container
     as in the dump routine. the handles don't need to match up. */
  damsel_handle temperature_tag_handle = (damsel_handle) malloc(1);
  DMSLtag_define(model, &temperature_tag_handle, DAMSEL_DATA_TYPE_DOUBLE, "temperature");

  damsel_container vertex_container_id;
  err = DMSLcontainer_create_sequence(model,
    (damsel_handle)rod, cols_on_this_rank * YLEN, sizeof(double), &vertex_container_id);

  /* now I have a get call, which mirrors set.
     at the time of making this call, I don't have anything that specifies
     *where* to get the data from - just "the memory store" which never
     actually holds anything. */
  DMSLmodel_map_tag(rod, vertex_container_id, &temperature_tag_handle);

/*
 now, when I read back a different pattern of handles than I wrote,
 do I define a vertex container that persists in the file?
 probably not - but then I need to be clearer about what gets stored
 in the file to indicate handle ranges (for example, in the tag+NNN
 dataset.
 then, if I'm not storing them in the file, they don't need file-scoped
 handles.
 the present implementation needs handles of containers to be unique for
 the lifetime of a file in order to distinguish between different bits of
 tag values. It will probably also need it in order to store the range
 information associated with each tag value datastore.
*/

// this code computes handles to the existing containers that came into
// existence on write. but that's not generally what I should be using here:
// I need to make a new handle container that contains just the handles
// that I do want...
// so I'm deliberately using *different* handles to the containers used on
// writing, in order that they don't match up.
// that's handle space management that seems icky and I'd prefer to not have
// to have this behave like this

/* ----- START ---- */
/* This block that defines filestore_vertex_container is the code that
   should be replaced by something that selects based on something that isn't
   in terms of file-side handles.
*/

  damsel_handle sfc_tag;
  err = DMSLselect_tag_by_name(model, "sfc", &sfc_tag);

  damsel_container filestore_vertex_container_id = DMSLselect_range(model,
  sfc_tag, smallest_x*YLEN, (smallest_x+cols_on_this_rank)*YLEN-1);

/* ----- END ----- */
// these calculations are the same for both reading and writing, so factor?

// but I haven't declared the filestore_vertex_container to actually
// be anything... so how can damsel know which range of values to bring
// in 
// i need to register it... but as I'm not writing, it isn't going to get
// written into the real store. which I guess is fine. but then there's a
// second cache layer between the memory store bit of cache and the
// inside-hdf5-store cache.

  err = DMSLmodel_map_handles(vertex_container_id, filestore_vertex_container_id);
  assert_dmsl_ok(err);
  DMSLcontainer_release(filestore_vertex_container_id);

  damsel_request_t req;
  DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_READ, &req);

  damsel_status_t status;
  DMSLmodel_wait(req, &status);

  DMSLmodel_close(model);
  DMSLcontainer_close(vertex_container_id);
  free((void *)temperature_tag_handle);
}

int main(int argc, char **argv) {
  printf("start proto2d\n");

  x_tag_handle = &x_tag_handle;
  y_tag_handle = &y_tag_handle;
  temperature_tag_handle = &temperature_tag_handle;
  sfc_tag_handle = &sfc_tag_handle;
  tag_container[0] = x_tag_handle;
  tag_container[1] = y_tag_handle;
  tag_container[2] = temperature_tag_handle;
  tag_container[3] = sfc_tag_handle;

#ifndef DAMSEL_SINGLETHREAD
  int res;
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res>=MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif

  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);

  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
  MPI_Comm_size(MPI_COMM_WORLD, &numranks);

  cols_on_this_rank = XLEN / numranks;
  assert(XLEN % numranks == 0);
  assert(cols_on_this_rank <= XLEN);

  smallest_x = cols_on_this_rank * myrank;
  assert(smallest_x < XLEN);

  printf("I am rank %d\n", myrank);

  if(argc == 2 && !strcmp("continue",argv[1])) {
    reload_sim();
  } else {
    init_sim();
  }


  print_sim();

  int step = 1;
  while(step%10 != 0) {
    printf("simulation step %d\n", step);
    step_sim();
    print_sim();
    step++;
  }

  printf("done simulating. now dumping.\n");

  dump_sim();

  if(myrank == 0) dump_ndc();
  printf("finalizing damsel\n");
  DMSLlib_finalize(lib);
  printf("finalizing MPI\n");
  MPI_Finalize();
  printf("end proto2\n");
  return 0;
}

