/* test create_sequence and writing to disk */

#include <assert.h>
#include <mpi.h>
#include <unistd.h>
#include <stdio.h>

#include "damsel.h"

#define FILENAME "output/test-create-tree-addfast.h5"

int main(int argc, char **argv) {
  short int data[] = {3, 1, 4, 1, 5}; // some arbitrary data to write

  printf("%s\n", argv[0]);

#ifndef DAMSEL_SINGLETHREAD
  int res;
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res == MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif

  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);

  damsel_model model;
  err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
  assert(model != NULL);

  /* define a tag which will store our data. the tag will be
     named 'somedata' */

  damsel_handle data_tag_handle = 0x10203;
  DMSLtag_define(model, &data_tag_handle, DAMSEL_DATA_TYPE_SHORTINT, "somedata");

  /* define a container of the handles on which we will store data. we
     will store the data on handles 20000 .. upwards and identify the
     container of these handles by another handle, 10020. */

  /* we'll create a sequence handle container with 5 handles in it. */

  damsel_handle vector_of_handles[] = { 0x20000, 0x21000, 0x22000,
                                        0x23000, 0x24000 };


  /* now wrap this inside a COLL container */
  damsel_container vertex_container_id = DMSLcontainer_tree_create(model, DAMSEL_HANDLE_COLLECTION_TYPE_VECTOR);
  DMSLcontainer_tree_add_fast(vertex_container_id, vector_of_handles, 5);

  /* now set the values of the somedata tag on the handles */
  DMSLmodel_map_tag(data, vertex_container_id, &data_tag_handle);

  /* at this point, the in-memory store knows about all the data that
     we have. next we will write the data out to disk.
     we need to invent some handles for the file store too. They can overlap
     with memstore handles, though here I make sure that they do not.
  */

  unlink(FILENAME); // maybe it doesn't exist, so we'll just silently ignore any error
  // filestore = DMSLstore_create_hdf5(FILENAME);
  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "WRITE");

  DMSLmodel_attach(model, FILENAME, MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);

  DMSLmodel_map_handles_inventing_file_handles(vertex_container_id);

  /* we also need to map the handles for the tags which contain values
     that we want to store.
     This happens by creating a container containing the tag to map, and
     mapping that into the file store. */

  damsel_container tag_container_id;
  err = DMSLcontainer_create_sequence(model, 
    data_tag_handle, 1, 1, &tag_container_id);

  DMSLmodel_map_handles_inventing_file_handles(tag_container_id);

  /* Now we can tell damsel to execute everything. */

  //MPI_Request req;
  damsel_request_t req;
  // DMSLexecute(memstore, m_memstore->attached_file_store, &req);
  DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_WRITE, &req);
  // MPI_Status status;
  damsel_status_t status;
  DMSLmodel_wait(req, &status);
  //MPI_Wait(&req, &status);

  DMSLmodel_close(model);
  DMSLcontainer_close(tag_container_id);
  DMSLcontainer_close(vertex_container_id);

  DMSLlib_finalize(lib);

  MPI_Finalize();
  return 0;
}

