
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "damsel.h"
#include "damsel-ndc.h"
#include "damsel-debug.h"

// for getting access to the file-side store pointer when reading.
#include "damsel-internal.h"

/* test writing a full mesh in parallel, consisting of:
 * entities: 2 hexahedra, 8 vertices per proc, arranged in 1D array
 * collections: 2 collections (1 for partition, 1 for material set)
 * tags: MATERIAL_SET (on collection), X_COORDS/Y_COORDS/Z_COORDS (on vertices), GLOBAL_ID (on entities/collection/tags).
 *
 * Vertices are located (0-1, 0-1, 2P-(2P+2)), where P = number of procs
 *
 * HANDLE SPACES
 * All local (application-side) handles are long unsigned integers.  Local handles for these things are as follows:
 * vertices: 1-12  (NVERTS)
 * hexes: 100, 101
 * collection: 200 (material set), 201 (partition set)
 * tags: 500-506 (GLOBAL_ID, XCOORDS, YCOORDS, ZCOORDS, COLL_FLAGS, MATERIAL_SET, PARALLEL_PARTITION)

 * 
 */

#define CHECK_ERROR(err) { assert(err.id == DMSL_OK.id); if (err.id != DMSL_OK.id) return err; }
#define RC { assert(err.id == DMSL_OK.id); if (err.id != DMSL_OK.id) return 1; }

#define FILENAME "output/test-fullmesh-parallel.h5"

/* in this app, handles are pointer-sized so we can use pointers for local handles */
typedef size_t app_handle;

struct tag_rep
{
  long tag_id;
  char *tag_name;
  damsel_data_type tag_datatype;
  damsel_handle file_handle;
};

struct tag_struct
{
  void *data;
  int tag_num;
};

struct ent_rep
{
  long start_id;
  damsel_handle *dhandles;
  damsel_entity_type ent_type;
  int num_ents;
  struct tag_struct *tag_structs;
  int num_tag_structs;
    // if vertex
  double *xyz;
    // if entity
  int verts_per_entity;
  long *connect;
};

struct coll_rep 
{
  long coll_id;
  damsel_handle coll_dhandle;
  int coll_flag;
  damsel_collection_type coll_type;
  long *contents;
  int num_ents;
  struct tag_struct *tag_structs;
  int num_tag_structs;
};

#define NVERTS 12
#define NHEXES  2
#define NCOLLS  2

struct tag_rep *read_tags = NULL;
struct ent_rep *read_ents = NULL;
struct coll_rep *read_colls = NULL;
long num_read_ents = 0, num_read_tags = 0, num_read_colls = 0;
long max_id = -1;
damsel_handle xcoord_ahandle = 0, ycoord_ahandle = 0, zcoord_ahandle = 0;
damsel_handle xcoord_app_handle=501, ycoord_app_handle=502, zcoord_app_handle=503; // these match up with the entries in tag_handles below.
/* the following are LOCAL handles */
damsel_handle vert_handles[] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12}, read_vert_handles[NVERTS];
damsel_handle hex_handles[2] = {100, 101}, read_tet_handles[2];
damsel_handle coll_handles[2] = {200, 201}, read_coll_handle;
damsel_handle tag_handles[] = {500, 501, 502, 503, 504, 505, 506};
damsel_handle read_tag_fhandles[7];
double xcoords[] = {0.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0, 0.0}, read_xcoords[NVERTS];
double ycoords[] = {0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0}, read_ycoords[NVERTS];
double zcoords[NVERTS], read_zcoords[NVERTS];
damsel_handle hex_connect[] = {1, 2, 3, 4, 5, 6, 7, 8,
                               5, 6, 7, 8, 9, 10, 11, 12};

int gids[NVERTS + NHEXES + NCOLLS], read_gids[NVERTS + NHEXES + NCOLLS];
damsel_handle coll_contents[] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 100, 101, 100, 101};
damsel_handle read_coll_contents[16];
int coll_flags[] = {0x1 << 4, 0x1 << 5}, read_coll_flag;

damsel_err_t define_tags(damsel_model model, damsel_handle *app_taghs);
damsel_err_t write_vertices(int rank, int nprocs, damsel_model model, damsel_handle *handles, int *gids);
damsel_err_t write_entities(damsel_model model, damsel_handle *apptags, int *gids);
damsel_err_t write_collections(int rank, int nprocs, damsel_model model, damsel_handle *apptags, int *gids);
damsel_err_t read_infos(damsel_model model, 
                        damsel_tag_buf_type **tag_infos, int *num_tag_infos, 
                        damsel_entity_buf_type **ent_infos, int *num_ent_infos, 
                        damsel_collection_buf_type **coll_infos, int *num_coll_infos);
damsel_err_t process_tags(damsel_model model, damsel_tag_buf_type *tag_infos, int num_tag_infos);
damsel_err_t process_ent_info(damsel_model model, damsel_entity_buf_type ent_info);
damsel_err_t process_entity_tags(int num_tags, damsel_container tagh_cont, damsel_container ent_cont,
                                 struct tag_struct *tag_structs_ptr, int num_tag_structs);
damsel_handle find_id(damsel_handle handle);
damsel_err_t process_colls(damsel_model model, damsel_collection_buf_type *coll_infos, int num_read_colls);
damsel_err_t print_mesh();
damsel_err_t print_ent_tags(struct tag_struct *tag_structs, int num_tag_structs, int ent_index);

int main(int argc, char **argv) {
#ifndef DAMSEL_SINGLETHREAD  
  int res;
#endif
  fprintf(stderr, "test-fullmesh\n");

#ifndef DAMSEL_SINGLETHREAD
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res == MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif

  int i, j, rank, nprocs;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

    /* compute some data based on P */

    /* z coords */
  zcoords[0] = zcoords[1] = zcoords[2] = zcoords[3] = 2*rank;
  zcoords[4] = zcoords[5] = zcoords[6] = zcoords[7] = 2*rank+1;
  zcoords[8] = zcoords[9] = zcoords[10] = zcoords[11] = 2*rank+2;

    /* global ids */
  int base_gid = 8*rank;
  /* vertices */
  for (i = 0; i < NVERTS; i++) gids[i] = base_gid++;  
  /* hexes */
  for (i = NVERTS, j=100+rank; i< NVERTS + NHEXES; i++, j++) {
      gids[i] = j;
  }
  gids[NVERTS + NHEXES] = 200;                          /* material set (shared) */
  gids[NVERTS + NHEXES + 1] = 300+rank;                 /* partition set (non-shared) */
  

    /* ============== WRITING =============== */

    /* init */
  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);
  damsel_model model;
  err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
    /* attach the file */
  unlink(FILENAME);
  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "WRITE");

  DMSLmodel_attach(model, FILENAME, MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);
    /* specify/write the data */
  err = define_tags(model, tag_handles); RC;
  err = write_vertices(rank, nprocs, model, tag_handles, gids); RC;
  err = write_entities(model, tag_handles, gids); RC;
  err = write_collections(rank, nprocs, model, tag_handles, gids); RC;
    /* write/finalize */
  DMSLmodel_transfer_sync(model, DAMSEL_TRANSFER_TYPE_WRITE);
  err = DMSLmodel_close(model); RC;
  
  DMSLlib_finalize(lib);

/*
  lib = DMSLlib_init();
  model = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64);
  damsel_trait_bundle tb2;
  err = DMSLtrait_create(&tb2);
  DMSLtrait_put(tb2, "mode", "READ");
  DMSLmodel_attach(model, FILENAME, MPI_COMM_WORLD, tb2);
  DMSLtrait_release(tb2);
  int num_tag_infos = 0, num_ent_infos = 0, num_coll_infos = 0;
  damsel_tag_buf_type *tag_infos = NULL;
  damsel_entity_buf_type *ent_infos = NULL;
  damsel_collection_buf_type *coll_infos = NULL;
  err = read_infos(model, &tag_infos, &num_tag_infos, &ent_infos, &num_ent_infos,
                   &coll_infos, &num_coll_infos); RC;

  for(i=0;i<num_ent_infos;i++) {
    assert_valid_handle_container_id(ent_infos[i].tag_handle_container);
  }

  read_tags = (struct tag_rep*)malloc(num_tag_infos * sizeof(struct tag_rep));
  assert(read_tags != NULL);
  err = process_tags(model, tag_infos, num_tag_infos); RC;

  read_ents = (struct ent_rep*)malloc(num_ent_infos * sizeof(struct ent_rep));
  for (i = 0; i < num_ent_infos; i++) {
      if (ent_infos[i].entity_type == DAMSEL_ENTITY_TYPE_VERTEX) {
        err = process_ent_info(model, ent_infos[i]); RC;
      }
  }
  for (i = 0; i < num_ent_infos; i++) {
      if (ent_infos[i].entity_type != DAMSEL_ENTITY_TYPE_VERTEX) {
        err = process_ent_info(model, ent_infos[i]); RC;
      }
  }

  read_colls = (struct coll_rep*)malloc(num_coll_infos * sizeof(struct coll_rep));
  err = process_colls(model, coll_infos, num_coll_infos); RC;

  DMSLmodel_transfer_sync(model, DAMSEL_TRANSFER_TYPE_READ);
  err = DMSLmodel_close(model); RC;
  DMSLlib_finalize(lib);
  
  if (coll_infos) 
      free(coll_infos);
	  
  if (tag_infos) 
      free(tag_infos);
	  
  if (ent_infos)
      free (ent_infos);
  */
  MPI_Finalize();
  fprintf(stderr, "Printing mesh:\n");
  err = print_mesh(); RC;
  
  fprintf(stderr, "done!\n");
  return 0;
}

damsel_err_t define_tags(damsel_model model, damsel_handle *tag_handles)
{
  int i, n=7;
  damsel_data_type types[] = {DAMSEL_DATA_TYPE_INTEGER, DAMSEL_DATA_TYPE_DOUBLE, 
                              DAMSEL_DATA_TYPE_DOUBLE, DAMSEL_DATA_TYPE_DOUBLE, DAMSEL_DATA_TYPE_INTEGER, 
                              DAMSEL_DATA_TYPE_INTEGER, DAMSEL_DATA_TYPE_INTEGER};
  char *names[] = {"GLOBAL_ID", "XCOORDS", "YCOORDS", "ZCOORDS", "COLL_FLAGS", "MATERIAL_SET", "PARTITION_SET"};

  for (i=0; i<n; i++) {
    damsel_err_t err = DMSLtag_define(model, (damsel_handle_ptr)(tag_handles+i), types[i], names[i]);
	if (err.id != 0) return DMSL_ERROR_GENERAL;
    DMSLmodel_map_single_handle_inventing_file_handle(model, tag_handles[i]);
  }

  return DMSL_OK;
}

damsel_err_t write_vertices(int rank, int nprocs, damsel_model model, damsel_handle *tag_handles, int *gids) 
{
  size_t count = NVERTS;
  damsel_err_t err;

  damsel_container vertex_cont;
  err = DMSLcontainer_create_vector(model, vert_handles, count, &vertex_cont);
  
  if (DAMSEL_CONTAINER_INVALID == vertex_cont) return DMSL_ERROR_GENERAL;
  
  err = DMSLmodel_map_handles_inventing_file_handles(vertex_cont);
  CHECK_ERROR(err);

    /* define the entities to damsel */
  err = DMSLentity_define(vertex_cont, DAMSEL_ENTITY_TYPE_VERTEX, 1, vertex_cont);
  CHECK_ERROR(err);
  
  err = DMSLmodel_map_tag(gids, vertex_cont, (damsel_handle_ptr)(tag_handles));
  CHECK_ERROR(err);

  /* map the coordinate to damsel */
  err = DMSLmodel_map_tag(xcoords, vertex_cont, (damsel_handle_ptr)(tag_handles+1));
  CHECK_ERROR(err);

  err = DMSLmodel_map_tag(ycoords, vertex_cont, (damsel_handle_ptr)(tag_handles+2));
  CHECK_ERROR(err);

  err = DMSLmodel_map_tag(zcoords, vertex_cont, (damsel_handle_ptr)(tag_handles+3));
  CHECK_ERROR(err);

    /* map the shared vertex handles */
  if (rank > 0) {
      /* left side */
    damsel_handle rhandles[] = {8, 9, 10, 11};
    damsel_container lhandles;
    err = DMSLcontainer_create_sequence(model, 0, 4, 1, &lhandles);
    int ranks[] = {rank-1, rank-1, rank-1, rank-1};
    err = DMSLmodel_map_global_handles(lhandles, ranks, (damsel_handle_ptr)rhandles);
    CHECK_ERROR(err);
    DMSLcontainer_release(lhandles);
  }
  if (rank < nprocs-1) {
      /* right side */
    damsel_handle rhandles[] = {0, 1, 2, 3};
    damsel_container lhandles;
    err = DMSLcontainer_create_sequence(model, 8, 4, 1, &lhandles);
    int ranks[] = {rank+1, rank+1, rank+1, rank+1};
    err = DMSLmodel_map_global_handles(lhandles, ranks, (damsel_handle_ptr)rhandles);
    CHECK_ERROR(err);
    DMSLcontainer_release(lhandles);
  }

  err = DMSLcontainer_release(vertex_cont);
  CHECK_ERROR(err);

  return DMSL_OK;
}
 
damsel_err_t write_entities(damsel_model model, damsel_handle *tag_handles, int *gids) 
{
  damsel_err_t err;
  damsel_container ent_cont;
  err = DMSLcontainer_create_vector(model, (damsel_handle_ptr)hex_handles, 2, &ent_cont);
  if (DAMSEL_CONTAINER_INVALID == ent_cont) return DMSL_ERROR_GENERAL;
  
  err = DMSLmodel_map_handles_inventing_file_handles(ent_cont);
  CHECK_ERROR(err);

  /* define the entities to damsel */
  err = DMSLentity_define_fast(ent_cont, DAMSEL_ENTITY_TYPE_HEX, 8, (damsel_handle_ptr)hex_connect);
  CHECK_ERROR(err);
  
  err = DMSLmodel_map_tag(gids+NVERTS, ent_cont, (damsel_handle_ptr)(tag_handles));
  CHECK_ERROR(err);
  
  err = DMSLcontainer_release(ent_cont);
  CHECK_ERROR(err);

  return DMSL_OK;
}

damsel_err_t write_collections(int rank, int nprocs, damsel_model model, damsel_handle *tag_handles, int *gids) 
{
  damsel_err_t err;
  damsel_container part_cont;
  
    /* make/populate material and partition sets */
    /* material set */
  damsel_container mat_cont;
  err = DMSLcontainer_create_vector(model, (damsel_handle_ptr)coll_contents, 14, &mat_cont);
  if (DAMSEL_CONTAINER_INVALID == mat_cont) CHECK_ERROR(DMSL_ERROR_GENERAL);
  DMSLcoll_create(model, (damsel_handle_ptr)coll_handles, mat_cont, DAMSEL_HANDLE_COLLECTION_TYPE_SET);
  err = DMSLmodel_map_single_handle_inventing_file_handle(model, coll_handles[0]);
  CHECK_ERROR(err);  
  err = DMSLcontainer_release(mat_cont);
  CHECK_ERROR(err);  

    /* map the global handle; reuse mat_cont; handle on other proc is the same */
  err = DMSLcontainer_create_vector(model, (damsel_handle_ptr)coll_handles, 2, &mat_cont);
  if (rank > 0) {
    int other_ranks[] = {rank-1, rank-1};
    err = DMSLmodel_map_global_handles(mat_cont, other_ranks, (damsel_handle_ptr)coll_handles);
  }
  if (rank < nprocs-1) {
    int other_ranks[] = {rank+1, rank+1};
    err = DMSLmodel_map_global_handles(mat_cont, other_ranks, (damsel_handle_ptr)coll_handles);
  }
  err = DMSLcontainer_release(mat_cont);
  CHECK_ERROR(err);  

    /* partition set */
  err = DMSLcontainer_create_vector(model, (damsel_handle_ptr)(coll_contents+14), 2, &part_cont);
  if (DAMSEL_CONTAINER_INVALID == part_cont) CHECK_ERROR(DMSL_ERROR_GENERAL);
  DMSLcoll_create(model, (damsel_handle_ptr)(coll_handles+1), part_cont, DAMSEL_HANDLE_COLLECTION_TYPE_SET);
  err = DMSLmodel_map_single_handle_inventing_file_handle(model, coll_handles[1]);
  
  err = DMSLcontainer_release(part_cont);
  CHECK_ERROR(err);  
    /* map the global handle; reuse part_cont; handle on other proc is the same */
  err = DMSLcontainer_create_vector(model, (damsel_handle_ptr)coll_handles+1, 1, &part_cont);
  if (rank > 0) {
    int other_ranks[] = {rank-1, rank-1};
    err = DMSLmodel_map_global_handles(part_cont, other_ranks, (damsel_handle_ptr)(coll_handles+1));
  }
  if (rank < nprocs-1) {
    int other_ranks[] = {rank+1, rank+1};
    err = DMSLmodel_map_global_handles(part_cont, other_ranks, (damsel_handle_ptr)(coll_handles+1));
  }
  err = DMSLcontainer_release(part_cont);
  CHECK_ERROR(err);  
 
    /*  make a container containing the collection handles */
  damsel_container coll_cont;
  err = DMSLcontainer_create_vector(model, (damsel_handle_ptr)coll_handles, 2, &coll_cont);
  if (DAMSEL_CONTAINER_INVALID == coll_cont) return DMSL_ERROR_GENERAL;

    /* map the collection handles first */
  DMSLmodel_map_handles_inventing_file_handles(coll_cont);  

  /* assign the tags that go on both sets */
  /* collection flag */
  err = DMSLmodel_map_tag(coll_flags, coll_cont, (damsel_handle_ptr)(tag_handles+4));
  CHECK_ERROR(err); 

    /* gid */
  err = DMSLmodel_map_tag((gids+14), coll_cont, (damsel_handle_ptr)(tag_handles));
  CHECK_ERROR(err);

    /* release collection container */
  err = DMSLcontainer_release(coll_cont);
  CHECK_ERROR(err);

    /* matset flag for material collection; reuse coll_cont */
  err = DMSLcontainer_create_vector(model, (damsel_handle_ptr)coll_handles, 1, &coll_cont);
  err = DMSLmodel_map_tag(gids+14, coll_cont, (damsel_handle_ptr)(tag_handles+5));
  CHECK_ERROR(err);  
  err = DMSLcontainer_release(coll_cont);
  CHECK_ERROR(err);  

    /* partition flag for partition collection; reuse coll_cont */
  err = DMSLcontainer_create_vector(model, (damsel_handle_ptr)(coll_handles+1), 1, &coll_cont);
  err = DMSLmodel_map_tag(&rank, coll_cont, (damsel_handle_ptr)(tag_handles+6));
  CHECK_ERROR(err);  
  err = DMSLcontainer_release(coll_cont);
  CHECK_ERROR(err);  

  return DMSL_OK;
}

damsel_err_t read_infos(damsel_model model, 
                        damsel_tag_buf_type **tag_infos, int *num_tag_infos, 
                        damsel_entity_buf_type **ent_infos, int *num_ent_infos, 
                        damsel_collection_buf_type **coll_infos, int *num_coll_infos) 
{
  damsel_err_t err = DMSLmodel_populate(model);
  CHECK_ERROR(err);
  
  int num_containers;
  DMSLmodel_get_tuple_count(model, &num_containers, num_tag_infos);
  if (num_tag_infos) {
    *tag_infos = (damsel_tag_buf_type*)malloc(*num_tag_infos*sizeof(damsel_tag_buf_type));
    err = DMSLmodel_get_tag_infos(model, *tag_infos);
    CHECK_ERROR(err);
  }
  
  *num_ent_infos = DMSLmodel_get_entity_count(model);
  if (*num_ent_infos) {
    *ent_infos = (damsel_entity_buf_type*)malloc(*num_ent_infos*sizeof(damsel_entity_buf_type));
    err = DMSLmodel_get_entity_infos(model, *ent_infos);
    CHECK_ERROR(err);
  }

  *num_coll_infos = DMSLmodel_get_collection_count(model);
  if (*num_coll_infos) {
    *coll_infos = (damsel_collection_buf_type*)malloc(*num_coll_infos*sizeof(damsel_collection_buf_type));
    err = DMSLmodel_get_collection_infos(model, *coll_infos);
    CHECK_ERROR(err);
  }

  return DMSL_OK;
}

damsel_err_t process_tags(damsel_model model, damsel_tag_buf_type *tag_infos, int num_tag_infos)
{
  long i;
  damsel_err_t err;
  damsel_container app_cont, dmsl_cont;
  damsel_model_internal *model_internal = (damsel_model_internal *)model;
  damsel_model file_model = (damsel_model) model_internal->attached_file;

  printf_ndc("BENC: num_tag_infos = %d", num_tag_infos);
  assert(read_tags != NULL);
  for (i = 0; i < num_tag_infos; i++) {
    read_tags[i].tag_datatype = tag_infos[i].tag_datatype;
    read_tags[i].tag_name = (char*)malloc(strlen(tag_infos[i].name) + 1);
    strcpy(read_tags[i].tag_name, tag_infos[i].name);
    read_tags[i].file_handle = *(tag_infos[i].tag_handle);
    
    num_read_tags++;

    printf_ndc("BENC: tag name is %s", tag_infos[i].name);
    printf_ndc("BENC: tag file-side handle is %llx", *(tag_infos[i].tag_handle));
    printf_ndc("BENC: tag app-side handle is %lx", (long)&read_tags[i]);

    if (!strcmp(tag_infos[i].name, "XCOORDS")) xcoord_ahandle = (damsel_handle) (read_tags+i);
    if (!strcmp(tag_infos[i].name, "YCOORDS")) ycoord_ahandle = (damsel_handle) (read_tags+i);
    if (!strcmp(tag_infos[i].name, "ZCOORDS")) zcoord_ahandle = (damsel_handle) (read_tags+i);
    
  }

    /* map the app tag handles to the damsel-provided ones */
  damsel_handle *thandles = (damsel_handle*)malloc(2*num_tag_infos*sizeof(damsel_handle));
  
  for (i = 0; i < num_tag_infos; i++) {
    thandles[i] = (damsel_handle)(read_tags+i);
    thandles[num_tag_infos+i] = (damsel_handle)(read_tags[i].file_handle);
  }
  err = DMSLcontainer_create_vector(model, (damsel_handle_ptr)thandles, num_tag_infos, &app_cont);
  err = DMSLcontainer_create_vector(file_model, (damsel_handle_ptr)(thandles+num_tag_infos), num_tag_infos, &dmsl_cont);
  err = DMSLmodel_map_handles(app_cont, dmsl_cont);
  CHECK_ERROR(err);

  free(thandles);
  DMSLcontainer_release(app_cont);
  DMSLcontainer_release(dmsl_cont);

  printf_ndc("BENC: process_tags. xcoord_ahandle = %p, ycoord_ahandle = %p, zcoord_ahandle = %p", 
             (void*)xcoord_ahandle, (void*)ycoord_ahandle, (void*)zcoord_ahandle);
  return DMSL_OK;
}

damsel_err_t process_ent_info(damsel_model model, damsel_entity_buf_type ent_info) 
{
  int i;
    /* create an app-side sequence and map to file-side container */
  unsigned long start_id = max_id+1;
  damsel_err_t err;
  max_id += ent_info.count;
  damsel_container ent_cont;
  err = DMSLcontainer_create_sequence(model, start_id, ent_info.count, 1, &ent_cont);
  err = DMSLmodel_map_handles(ent_cont, ent_info.entity_container);
  CHECK_ERROR(err);

  long ind = num_read_ents++;
  
  read_ents[ind].start_id = start_id;
  read_ents[ind].ent_type = ent_info.entity_type;
  read_ents[ind].num_ents = ent_info.count;
  read_ents[ind].verts_per_entity = ent_info.vertices_per_entity;

  printf_ndc("BENC: ent_info.entity_type = %d",ent_info.entity_type);

  if (ent_info.entity_type != DAMSEL_ENTITY_TYPE_VERTEX) {
    read_ents[ind].xyz = NULL;
    read_ents[ind].verts_per_entity = ent_info.vertices_per_entity;
    read_ents[ind].connect = (long*)malloc(read_ents[ind].num_ents * read_ents[ind].verts_per_entity * sizeof(long));
    read_ents[ind].dhandles = (damsel_handle*)malloc(read_ents[ind].num_ents * sizeof(damsel_handle));
    for (i = 0; i < read_ents[ind].num_ents; i++)
      read_ents[ind].dhandles[i] = DMSLcontainer_handle_at_position(ent_info.entity_container, i);
    
      /* map connectivity */
    int vcount = DMSLcontainer_count(ent_info.vertex_container);
    for (i = 0; i < vcount; i++) {
      read_ents[ind].connect[i] = find_id(DMSLcontainer_handle_at_position(ent_info.vertex_container, i));
      assert(-1 != read_ents[ind].connect[i]);
    }

      /* process tags */
    read_ents[ind].num_tag_structs = ent_info.tag_count;
    read_ents[ind].tag_structs = (struct tag_struct*)malloc(read_ents[ind].num_tag_structs*sizeof(struct tag_struct));
    err = process_entity_tags(ent_info.tag_count, ent_info.tag_handle_container, ent_cont, 
                              read_ents[ind].tag_structs, read_ents[ind].num_tag_structs);
    CHECK_ERROR(err);
  }
  else {
    assert(0 != xcoord_ahandle && 0 != ycoord_ahandle && 0 != zcoord_ahandle);
    read_ents[ind].connect = NULL;

      /* should have one vertex per entity */
    assert(ent_info.vertices_per_entity == 1);
    read_ents[ind].num_ents = ent_info.count;
    read_ents[ind].xyz = (double*)malloc(read_ents[ind].num_ents * 3 * sizeof(double));
    read_ents[ind].dhandles = (damsel_handle*)malloc(read_ents[ind].num_ents * sizeof(damsel_handle));
    for (i = 0; i < read_ents[ind].num_ents; i++)
      read_ents[ind].dhandles[i] = DMSLcontainer_handle_at_position(ent_info.entity_container, i);


    err = DMSLmodel_map_tag(read_ents[ind].xyz, ent_cont, (damsel_handle_ptr)&xcoord_ahandle);
    CHECK_ERROR(err);
    err = DMSLmodel_map_tag(read_ents[ind].xyz+read_ents[ind].num_ents, ent_cont, (damsel_handle_ptr)&ycoord_ahandle);
    CHECK_ERROR(err);
    err = DMSLmodel_map_tag(read_ents[ind].xyz+2*read_ents[ind].num_ents, ent_cont, (damsel_handle_ptr)&zcoord_ahandle);
    CHECK_ERROR(err);

      /* process tags */
    read_ents[ind].num_tag_structs = ent_info.tag_count;
    read_ents[ind].tag_structs = (struct tag_struct*)malloc(read_ents[ind].num_tag_structs*sizeof(struct tag_struct));
    assert_valid_handle_container_id(ent_info.tag_handle_container);
    err = process_entity_tags(ent_info.tag_count, ent_info.tag_handle_container, ent_cont, 
                              read_ents[ind].tag_structs, read_ents[ind].num_tag_structs);
    CHECK_ERROR(err);
  }

  err = DMSLcontainer_release(ent_cont);
  CHECK_ERROR(err);
  
  return DMSL_OK;
}

damsel_err_t process_entity_tags(int num_tags, damsel_container tagh_cont, damsel_container ent_cont,
                                 struct tag_struct *tag_structs_ptr, int num_tag_structs)
{
  int num_ents = DMSLcontainer_count(ent_cont);
  int i;

  for (i = 0; i < num_tags; i++) {
    
      /* find the local handle for the tag  */
    damsel_handle dh = DMSLcontainer_handle_at_position(tagh_cont, i);
    printf_ndc("BENC: dh = %llx", dh);
    damsel_handle ind = find_id(dh);
    assert(-1 != ind);
    if (-1 == ind) return DMSL_ERROR_GENERAL;
    tag_structs_ptr[i].tag_num = ind;
    assert(read_tags != NULL); 
    switch (read_tags[ind].tag_datatype) {
      case DAMSEL_DATA_TYPE_INTEGER:
          tag_structs_ptr[i].data = malloc(num_ents * sizeof(int));
          break;
      case DAMSEL_DATA_TYPE_INT64:
          tag_structs_ptr[i].data = malloc(num_ents * sizeof(long));
          break;
      case DAMSEL_DATA_TYPE_FLOAT:
          tag_structs_ptr[i].data = malloc(num_ents * sizeof(float));
          break;
      case DAMSEL_DATA_TYPE_DOUBLE:
          tag_structs_ptr[i].data = malloc(num_ents * sizeof(double));
          break;
      case DAMSEL_DATA_TYPE_HANDLE:
          tag_structs_ptr[i].data = malloc(num_ents * sizeof(damsel_handle));
          break;
      case DAMSEL_DATA_TYPE_SHORTINT:
          tag_structs_ptr[i].data = malloc(num_ents * sizeof(short));
          break;
      default:
            /* shouldn't be any other types of tags */
          printf_ndc("unknown tag data type %llx", read_tags[ind].tag_datatype);
          return DMSL_ERROR_GENERAL;
    }

    damsel_handle tmp_handle = (damsel_handle)(read_tags+ind);
    damsel_err_t err = DMSLmodel_map_tag(tag_structs_ptr[i].data, ent_cont, &tmp_handle);
    CHECK_ERROR(err);
  }
  
  return DMSL_OK;
}

damsel_handle find_id(damsel_handle handle) 
{
  long ind, i;
  push_printf_ndc("find_id looking for handle %llx", handle);
  for (ind = 0; ind < num_read_ents; ind++) {
    for (i = 0; i < read_ents[ind].num_ents; i++)
      if (handle == read_ents[ind].dhandles[i]) {
        pop_ndc();
        return read_ents[ind].start_id + i;
      }
  }
  printf_ndc("entity scan found nothing. scanning read_tags");
  for (i = 0; i < num_read_tags; i++) {
    printf_ndc("comparing handle %llx with (read_tags+i) = %llx", handle, (damsel_handle)(read_tags+i));
    damsel_handle thandle = read_tags[i].file_handle;
    printf_ndc("tag handle is %llx", thandle);
    if (handle == thandle) {
      printf_ndc("found handle, id = %lx", i);
      pop_ndc();
      return i;
    }
  }
  printf_ndc("nothing found");
  pop_ndc();
  return -1;
}

damsel_err_t process_colls(damsel_model model, damsel_collection_buf_type *coll_infos, int num_colls) 
{
  long start_id = max_id + 1, i, j;
  damsel_err_t err;
  damsel_container single_cont;
  damsel_container coll_acont;
  err = DMSLcontainer_create_sequence(model, start_id, num_colls, 1, &coll_acont);
  damsel_handle *dhandles = (damsel_handle*)malloc(num_colls*sizeof(damsel_handle));
  for (i = 0; i < num_colls; i++) {
    read_colls[i].coll_id = ++max_id;
    read_colls[i].coll_type = coll_infos[i].type;
    read_colls[i].coll_dhandle = *(coll_infos[i].collection_handle);
    dhandles[i] = read_colls[i].coll_dhandle;
    read_colls[i].num_ents = DMSLcontainer_count(coll_infos[i].contents);
    read_colls[i].contents = (long*)malloc(read_colls[i].num_ents * sizeof(long));
    for (j = 0; j < read_colls[i].num_ents; j++) {
      read_colls[i].contents[j] = find_id(DMSLcontainer_handle_at_position(coll_infos[i].contents, j));
      assert(-1 != read_colls[i].contents[j]);
    }
  }
  
    /* map the collection handles */
  damsel_model_internal *model_internal = (damsel_model_internal *)model;
  damsel_model file_model = (damsel_model) model_internal->attached_file;
  err = DMSLcontainer_create_vector(file_model, dhandles, num_colls, &single_cont);
  err = DMSLmodel_map_handles(coll_acont, single_cont);
  CHECK_ERROR(err);
  err = DMSLcontainer_release(coll_acont);
  CHECK_ERROR(err);
  err = DMSLcontainer_release(single_cont);
  CHECK_ERROR(err);
  free(dhandles);

  for (i = 0; i < num_colls; i++) {
      /* process tags */
    read_colls[i].num_tag_structs = coll_infos[i].tag_count;
    read_colls[i].tag_structs = (struct tag_struct*)malloc(read_colls[i].num_tag_structs*sizeof(struct tag_struct));
    err = DMSLcontainer_create_sequence(model, read_colls[i].coll_dhandle, 1, 1, &single_cont);
    err = process_entity_tags(coll_infos[i].tag_count, coll_infos[i].tag_handle_container, single_cont,
                              read_colls[i].tag_structs, read_colls[i].num_tag_structs);
    CHECK_ERROR(err);
    err = DMSLcontainer_release(single_cont);
    CHECK_ERROR(err);
  }

  num_read_colls = num_colls;
  
  return DMSL_OK;
}

damsel_err_t print_mesh() 
{
    /* print information about the whole mesh */
  fprintf(stderr, "Vertices:\n");
  int i, j, k;
  for (i = 0; i < num_read_ents; i++) {
    if (read_ents[i].ent_type != DAMSEL_ENTITY_TYPE_VERTEX) continue;
    for (j = 0; j < read_ents[i].num_ents; j++) {
      fprintf(stderr, "   %ld : x = %f, y = %f, z = %f.\n", read_ents[i].start_id+j, 
             read_ents[i].xyz[j], read_ents[i].xyz[read_ents[i].num_ents+j], read_ents[i].xyz[2*read_ents[i].num_ents+j]);
      print_ent_tags(read_ents[i].tag_structs, read_ents[i].num_tag_structs, j);
    }
  }
  
  for (i = 0; i < num_read_ents; i++) {
    if (read_ents[i].ent_type == DAMSEL_ENTITY_TYPE_VERTEX) continue;
    fprintf(stderr, "Entity type %d:\n", read_ents[i].ent_type);
    for (j = 0; j < read_ents[i].num_ents; j++) {
      fprintf(stderr, "   %ld : connect: ", read_ents[i].start_id+j);
      for (k = 0; k < read_ents[i].verts_per_entity; k++) 
        fprintf(stderr, "%ld ", read_ents[i].connect[j*read_ents[i].verts_per_entity+k]);
      fprintf(stderr, "\n");
      print_ent_tags(read_ents[i].tag_structs, read_ents[i].num_tag_structs, j);
    }
  }

  fprintf(stderr, "Collections:\n");
  for (i = 0; i < num_read_colls; i++) {
    fprintf(stderr, "   Collection %ld: type = %d, flag = %d\n", read_colls[i].coll_id, read_colls[i].coll_type, read_colls[i].coll_flag);
    fprintf(stderr, "      Contents:\n");
    for (j = 0; j < read_colls[i].num_ents/8; j++)
      fprintf(stderr, "         %ld %ld %ld %ld %ld %ld %ld %ld\n", 
             read_colls[i].contents[8*j  ], read_colls[i].contents[8*j+1], read_colls[i].contents[8*j+2], read_colls[i].contents[8*j+3], 
             read_colls[i].contents[8*j+4], read_colls[i].contents[8*j+5], read_colls[i].contents[8*j+6], read_colls[i].contents[8*j+7]);
    fprintf(stderr, "         ");
    for (j = 8*(read_colls[i].num_ents/8); j <  read_colls[i].num_ents; j++) printf("%ld ", read_colls[i].contents[j]);
    fprintf(stderr, "\n");

    print_ent_tags(read_colls[i].tag_structs, read_colls[i].num_tag_structs, i);
  }

  return DMSL_OK;
}

damsel_err_t print_ent_tags(struct tag_struct *tag_structs, int num_tag_structs, int ent_index) 
{
  int t;
  if (!num_tag_structs) {
    fprintf(stderr, "   Tags: (none)\n");
    return DMSL_OK;
  }

  for (t = 0; t < num_tag_structs; t++) {
    fprintf(stderr, "      %s: ", read_tags[tag_structs[t].tag_num].tag_name);
    if (DAMSEL_DATA_TYPE_INTEGER == read_tags[tag_structs[t].tag_num].tag_datatype ||
        DAMSEL_DATA_TYPE_INT64 == read_tags[tag_structs[t].tag_num].tag_datatype ||
        DAMSEL_DATA_TYPE_HANDLE == read_tags[tag_structs[t].tag_num].tag_datatype ||
        DAMSEL_DATA_TYPE_SHORTINT == read_tags[tag_structs[t].tag_num].tag_datatype) 
      fprintf(stderr, "%d\n", ((int*)tag_structs[t].data)[ent_index]);
    else if (DAMSEL_DATA_TYPE_DOUBLE == read_tags[tag_structs[t].tag_num].tag_datatype)
      fprintf(stderr, "%f\n", ((double*)tag_structs[t].data)[ent_index]);
    else if (DAMSEL_DATA_TYPE_FLOAT == read_tags[tag_structs[t].tag_num].tag_datatype) 
      fprintf(stderr, "%f\n", ((float*)tag_structs[t].data)[ent_index]);
    else fprintf(stderr, "(non-printable)\n");
  }

  return DMSL_OK;
}
