/* tests reading into a sequence */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <unistd.h>
#include <memory.h>

#include "damsel.h"
#include "damsel-hdf5.h"
#include "damsel-debug.h"
#include "damsel-connectivity.h"


int main(int argc, char **argv) {

  // we'll read the data into data[] and then check that
  // it matches up with target data check that
  // it matches up with target data
  short int targetdata[] = {3, 1, 4, 1, 5, 9};
  short int data[6];

  printf("%s\n", argv[0]);
  if(argc!=2) {
    printf("specify filename of .h5 dump file to check\n");
    return 2;
  }

#ifndef DAMSEL_SINGLETHREAD
  int res;
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res == MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif

  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);

  damsel_model model;
  err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
  assert(model != NULL);
  damsel_model_internal *ms = (damsel_model_internal *)model;
  assert(ms != NULL);

  assert_valid_store_id(model);
  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "READ");

  DMSLmodel_attach(model, argv[1], MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);
  DMSLmodel_populate(model);

// for now, need to call get_container_infos to cause containers to be
// defined in memory.
// maybe this happen at open time?

// BUG: this call needs to happen before defining any runtime containers
// or any other call that creates runtime containers internally, I think,
// because otherwise we'll end up returning runtime created containers too.
  int container_count;

  DMSLmodel_get_tuple_count((damsel_model)ms, &container_count, NULL);
  damsel_container_buf_type *container_buf =
    malloc(container_count * sizeof(damsel_container_buf_type));

  assert_valid_store_id(model);
  DMSLmodel_get_container_infos(model, container_buf);

  free(container_buf);


  damsel_handle data_tag_handle = 0x31000;

  /* we'll create a sequence handle collection with 6 handles in it. */

  assert_valid_store_id(model);
  damsel_container vertex_collection_id;
  err = DMSLcontainer_create_sequence(model, 0x80900, 6, 0x5, &vertex_collection_id);

  assert(vertex_collection_id != NULL);

  assert_valid_store_id(model);
  damsel_container tag_collection_id;
  err = DMSLcontainer_create_sequence(model, 
    data_tag_handle, 1, 1, &tag_collection_id);

  assert(tag_collection_id != NULL);

  assert_valid_store_id(model);
  damsel_handle fileside_data_tag_handle;
  err  = DMSLselect_tag_by_name(model, "pidigits", &fileside_data_tag_handle);
  damsel_container file_tag_collection_id;
  err = DMSLcontainer_create_sequence((damsel_model)ms->attached_file, fileside_data_tag_handle, 1, 1, &file_tag_collection_id);
  DMSLmodel_map_handles(tag_collection_id, file_tag_collection_id);

  assert_valid_store_id(model);

  /* now set the values of the somedata tag on the handles */
  DMSLmodel_map_tag(data, vertex_collection_id, &data_tag_handle);

  int entity_count;
  entity_count = DMSLmodel_get_entity_count(model);
  assert(entity_count == 1); // only going to allow data files with one
                             // entity definition
  damsel_entity_buf_type *entity_buffer = 
    malloc(entity_count * sizeof(damsel_entity_buf_type));
  assert(entity_buffer != NULL);
  assert_valid_store_id(model);
  DMSLmodel_get_entity_infos(model, entity_buffer);

  damsel_container vertex_container = entity_buffer[0].vertex_container;
  damsel_container entity_container = entity_buffer[0].entity_container;

  printf("entity container = %p\n", entity_container);
  printf("entity container size = %d\n", ((damsel_container_internal *)entity_container)->quantity);
  printf("vertex container = %p\n", vertex_container);
  printf("vertex container size = %d\n", ((damsel_container_internal *)vertex_container)->quantity);

  // each edge has 2 vertices.
  assert( ((damsel_container_internal *)entity_container)->quantity * 2
       == ((damsel_container_internal *)vertex_container)->quantity);

  assert( ((damsel_container_internal *)entity_container)->quantity == entity_buffer[0].count);

  assert(entity_buffer[0].count == 5);

  DMSLmodel_close_entity_infos(model, entity_buffer);
  free(entity_buffer);

  // pi should be stored on the point vertices that these edge entities connect.

  // pick an arbitrary edge. (the first one in the entity container, for example)
  // do I have public functions for introspection into containers?

  damsel_handle leftmost = DMSLcontainer_handle_at_position(vertex_container, 0);

  // now lets see if we can move left a hop

  int ix;
  int changed = 1;

  while(changed == 1) {
    changed=0;
    for(ix=1; ix<10; ix+=2) {
      if(DMSLcontainer_handle_at_position(vertex_container, ix) == leftmost) {
        leftmost = DMSLcontainer_handle_at_position(vertex_container, ix-1);
        changed=1;
      }
    }
  }

  // when we reach here, leftmost should be leftmost
  // so now we should be able to move right, digit at a time.

  damsel_handle digitpos[6];

  int dpix=0;
  digitpos[dpix] = leftmost;
  dpix++;

  while(dpix<6) {
    for(ix=0; ix<10; ix+=2) {
      if(DMSLcontainer_handle_at_position(vertex_container, ix) == digitpos[dpix-1]) {
        digitpos[dpix] = DMSLcontainer_handle_at_position(vertex_container, ix+1);
        dpix++;
        break;
      }
    }
  }

  // now, digitpos[] should contain the 6 file-side handles to use, properly
  // permuted.

  damsel_attached_file *fstore = ((damsel_model_internal *)model)->attached_file;
  damsel_container targetcontainer;
  err = DMSLcontainer_create_vector((damsel_model)fstore, digitpos, 6, &targetcontainer);

  // use the vertex thing to move left 'till we hit the start


 // damsel_container filestore_vertex_collection_id = DMSLselect_handles_with_values(model, fileside_data_tag_handle);

  DMSLmodel_map_handles(vertex_collection_id, targetcontainer);

  damsel_request_t req;
  DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_READ, &req);
  printf("DMSLmodel_transfer_async has MPI request %x\n", req.mpi);
  damsel_status_t status;
  DMSLmodel_wait(req, &status);

  for(ix=0; ix<6; ix++) {
    printf("A: Position %d value %d should be %d\n", ix, data[ix], targetdata[ix]);
  }
  for(ix=0; ix<6; ix++) {
    printf("B: Position %d value %d should be %d\n", ix, data[ix], targetdata[ix]);
  }
  int fail=0;
  for(ix=0; ix<6; ix++) {
    printf("C: Position %d value %d should be %d\n", ix, data[ix], targetdata[ix]);
    if(targetdata[ix] != data[ix]) {
      printf("fail at position %d\n", ix);
      printf("D: Position %d value %d should be %d\n", ix, data[ix], targetdata[ix]);
      fail = 1;
    }
  }
  if(fail == 1) return 1;

  DMSLmodel_close(model);
  DMSLcontainer_close(targetcontainer);
  DMSLcontainer_close(vertex_collection_id);
  DMSLcontainer_close(tag_collection_id);
  DMSLcontainer_close(file_tag_collection_id);
  DMSLlib_finalize(lib);

  MPI_Finalize();
  return 0;
}

